/*
 * Handle the .dderc file
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "../sedit.h"

static int ddercNoHome;

typedef struct ddePath
{
	struct ddePath *next;
	char *path;
} ddePath;

static ddePath *ddePaths;

void ddercParse();

void
freeDDEPaths()
{
	ddePath *ddep = ddePaths;

	while (ddep)
	{
		ddePath *next = ddep->next;
		Free(ddep->path);
		Free(ddep);
		ddep = next;
	}
}

/*
 * Append 'dir' to the list
 */

static void _path(dir)
char *dir;
{
	ddePath *ddep;

	ddep = Malloc(sizeof(*ddep));
	ddep->next = 0;
	ddep->path = roomfor(dir);
	if (!ddePaths)
		ddePaths = ddep;
	else
	{
		ddePath *temp = ddePaths;
		while (temp->next)
			temp = temp->next;
		temp->next = ddep;
	}
}

/*
 * Append 'dir' to the tagfile
 */

static void _tags(dir)
char *dir;
{
	addTagFile(dir);
}

void
ddercParse(oname)
char *oname;
{
	FILE *f;
	static int depth = 0;
	char *name;

	name = _patchenv(oname);

	if (depth++ > 30)
	{
		fprintf(stderr, "dde: recursive .dderc files\n");
		exit(1);
	}

	f = fopen(name, "r");
	if (f)
	{
		char buf[100000];

		while (fgets(buf, sizeof(buf), f))
		{
			char **subStrings = 0;
			int nSubStrings;
			int i;
			int buflen = strlen(buf);

			if (buf[0] && buf[strlen(buf) - 1] == '\n')
				buf[strlen(buf) - 1] = '\0';

			nSubStrings = sliceargs(buf, &subStrings, 0, 0);
			if (nSubStrings)
			{
				void (*func)() = 0;

				if (!strcmp(subStrings[0], "source"))
					func = ddercParse;
				else if (!strcmp(subStrings[0], "path"))
					func = _path;
				else if (!strcmp(subStrings[0], "tags"))
					func = _tags;

				if (func)
				{
					char *dirName = Malloc(strlen(name) + buflen + 1);
					strcpy(dirName, name);
					if (strrchr(dirName, '/'))
					{
						int endpos;
						*(strrchr(dirName, '/') + 1) = '\0';
						endpos = strlen(dirName);
						for (i = 1; i < nSubStrings; i++)
						{
							if (subStrings[i][0] == '/' ||
												subStrings[i][0] == '$')
								(*func)(subStrings[i]);
							else
							{
								strcat(dirName, subStrings[i]);
								(*func)(dirName);
								dirName[endpos] = '\0';
							}
						}
					}
					Free(dirName);
				}
				else if (!strcmp(subStrings[0], "nohome"))
				{
					ddercNoHome = 1;
				}
			}
			if (subStrings)
				cleanUpSlicedArgs(subStrings, nSubStrings);
		}
		fclose(f);
	}
	Free(name);
	depth--;
}

void
ddercInit()
{
	_path(".");					/* Put current directory at the start */

	/* Try to find a .dderc */
	{
		char *firstdderc = searchup(".", ".dderc");
		if (firstdderc)
		{
			ddercParse(firstdderc);
			Free(firstdderc);
		}
	}
	if (!ddercNoHome)
	{
		char *home = getenv("HOME");
		if (home)
		{
			char *nbuf = Malloc(strlen(home) + 10);
			sprintf(nbuf, "%s/.dderc", home);
			if (home)
				ddercParse(nbuf);
			Free(nbuf);
		}
	}
}

char *
ddercLookupFile(name)
char *name;
{
	char *ret = roomfor(name);

	if (!strchr(name, '/'))
	{
		ddePath *ddep;
		for (ddep = ddePaths; ddep && !signalled(); ddep = ddep->next)
		{
			char *thisPath = Malloc(strlen(ddep->path) + strlen(name) + 4);
			sprintf(thisPath, "%s/%s", ddep->path, name);
			if (access(thisPath, R_OK) == 0)
			{
				Free(ret);
				ret = thisPath;
				break;
			}
			else
			{
				Free(thisPath);
			}
		}
	}
	return nodotslash(ret);
}
