/*
 * Document mode editing functions
 */

#include "../sedit.h"
#include <stdio.h>

int rmargin;
int lmargin;
int justifymode;
int debug;
extern unsigned char *getnext(), *fwd_fnw();
unsigned char *format();

extern unsigned char *cur_lp;		/* Start of current line */
extern unsigned char *cur_ptr;		/* Current working pointer */
extern unsigned char *mem_buf;		/* The text buffer */
extern unsigned char *end_buf;
extern int cur_x, cur_y, width;
extern unsigned char *blkstart, *blkend;
extern unsigned char *newcur_ptr;
extern int save_coff;
extern int nondocmode;
extern unsigned char *blkstart;
extern unsigned char *blkend;
extern int blkset;

/*
 * Reformat the paragraph at the cursor, starting from the cursor position
 */

reformat(upddisp)
int upddisp;		/* If true, update screen display */
{
	if (nondocmode)
		nondocreformat(upddisp);
	else
		docreformat(upddisp);
}

int blank(c)
unsigned char c;
{
	return c == ' ' || c == '\t' || c == '\n' || c == '\r';
}

nondocreformat(upddisp)
int upddisp;
{
	unsigned char *para_start;
	unsigned char *para_end;
	unsigned char *work_buf;
	unsigned char *final_buf;
	unsigned char *in_p, *out_p;
	int col;
	int final_buf_len;
	unsigned char *last_blank;
	unsigned char *new_cur_ptr;
	int at_start;
	int rmargin_adjust;

	/* Find the start */
	para_start = cur_ptr;
	new_cur_ptr = cur_ptr;

	while (para_start > mem_buf)
	{
		if (para_start == mem_buf + 1 && para_start[-1] == '\n')
			break;
		if (para_start[-1] == '\n' && para_start[-2] == '\n')
			break;
		para_start--;
	}

	/* Find the end */
	para_end = cur_ptr;
	while (para_end < end_buf)
	{
		if (para_end[0] == '\n' && para_end == end_buf - 1)
			break;
		if (para_end [0] == '\n' && para_end[1] == '\n')
			break;
		para_end++;
	}

/*
	fprintf(stderr, "cur=%d, start=%d, end=%d\n",
		cur_ptr - mem_buf,
		para_start - mem_buf,
		para_end - mem_buf);
*/

	if (para_end == para_start)
		return;

	work_buf = Malloc((para_end - para_start) * 10 + 10);
	final_buf = Malloc((para_end - para_start) * 10 + 10);
	in_p = para_start;
	out_p = work_buf;

	while (blank(*in_p) && in_p < para_end)
	{
		*out_p++ = *in_p++;
	}

	if (in_p < para_end)
		*out_p++ = *in_p++;

	while (in_p < para_end)
	{
		if (blank(*in_p))
		{
			if (out_p[-1] != ' ')
			{
				*out_p++ = ' ';
				if (strchr(".?!", out_p[-2]))
					*out_p++ = ' ';
			}
			else
			{
				if (in_p < cur_ptr)
					new_cur_ptr--;
			}
		}
		else
		{
			*out_p++ = *in_p;
		}
		in_p++;
	}
	*out_p = 0;

/*	fprintf(stderr, "stripped: '%s'\n", work_buf); */

	/* Now add newlines */
	col = 0;
	in_p = work_buf;
	last_blank = 0;
	rmargin_adjust = rmargin - lmargin;
	if (rmargin_adjust < 10)
		rmargin_adjust = 10;

	while (*in_p)
	{
		if (*in_p == ' ')
		{
			if (col == rmargin_adjust)
			{
				*in_p = '\n';
				col = 0;
				in_p++;
				last_blank = 0;
			}
			else
			{
				if (col > rmargin_adjust)
				{
					if (last_blank)
						in_p = last_blank;
					*in_p = '\n';
					last_blank = 0;
					col = 0;
					in_p++;
				}
				else
				{
					last_blank = in_p;
					in_p++;
					col++;
				}
			}
		}
		else
		{
			in_p++;
			col++;
		}
	}

	if (col > rmargin_adjust && last_blank)
	{
		*last_blank = '\n';
	}

	/* Now add leading space */
	out_p = final_buf;
	in_p = work_buf;
	while (*in_p)
	{
		if (lmargin)
		{
			if (in_p > work_buf && in_p[-1] == '\n')
			{
				int i;
				for (i = 0; i < lmargin; i++)
				{
					*out_p++ = ' ';
				}
				if ((in_p - work_buf) < (cur_ptr - para_start))
				{
					new_cur_ptr += lmargin;
				}
			}
		}
		*out_p++ = *in_p++;
	}

/*	new_cur_ptr -= front;*/

/*	fprintf(stderr, "massaged: '%s'\n", work_buf);*/

	final_buf_len = out_p - final_buf;
	if (memcmp(para_start, final_buf, final_buf_len))
	{
		predelete(para_start, para_end - para_start);
		blkswap(para_start, para_end - para_start, final_buf, final_buf_len);
		if (upddisp)
		{
			blkstart = para_start;
			blkend = para_start + final_buf_len;
			blkset = 0;
		}
		cur_ptr = new_cur_ptr;
		redoptr();
		redhere();
	}
	Free(work_buf);
	Free(final_buf);
}

/*
 * Reformat the paragraph at the cursor, starting from the cursor position
 */

docreformat(upddisp)
int upddisp;		/* If true, update screen display */
{
	int length, l, offset;
	register unsigned char *cp, *buf, *cp2, ch;
	unsigned char *curspos;

	offset = cur_ptr - cur_lp;
	/* Find end of line: can't use getnext()! */
	cp = cur_ptr;
	while (*cp && *cp != '\n')
		cp++;

	length = cp - cur_lp;
	if (length < 3)
		return;
	buf = Malloc(length + 2);
	movmem(cur_lp, buf, length);
	buf[length] = 0;

	/* Convert soft spaces at start of paragraph into hard ones */
	cp = buf;
	while (MASK(*cp) == ' ')
		*cp++ = ' ';

	/* Strip soft spaces */
	cp = cp2 = buf;
	curspos = cp + offset;
	l = 0;
	while (*cp)
	{
		ch = *cp;
		l++;
		if (ch != 0xa0 && ch != 0x8d && ch != 0x8a)
			*cp2++ = ch;
		else if ((ch == 0x8a) && (l != length))
		{
			/* Replace soft NL with original hard space */
			*cp2++ = ' ';
		}
		/* May need to adjust offset for removed chars */
		if ((ch == 0xa0 || ch == 0x8d) && cp < curspos)
			offset--;
		cp++;
	}
	*cp2 = '\0';

	if (cp2 = format(buf, offset))
	{
		length -= cur_ptr - cur_lp;
		blkswap(cur_ptr, length, cp2, strlen(cp2));
		if (upddisp)
			redhere();
		set_param(cur_lp);
		if (upddisp)
			paraend();
		Free(cp2);
	}
	Free(buf);
}

/* Format all paragraphs which start in the block */
formatblock()
{
	register unsigned char *cp;

	newcur_ptr = cur_ptr;
	cp = blkstart;
	cur_x = 0;		/* Used by reformat() */
	while (cp < blkend)
	{
		if (cp == mem_buf || cp[-1] == '\n')
		{
			cur_ptr = cp;
			while (MASK(*cur_ptr) == ' ')
				cur_ptr++;
			redoptr();
			reformat(0);
		}
		cp++;
	}
	cur_ptr = newcur_ptr;
	redoptr();
	redhere();
}

/*
 * Format the paragraph. 'offset' is the number of bytes into 'in' at which
 * to start
 */

unsigned char *
format(in, offset)
register unsigned char *in;
{
	register unsigned char *incpy, *cp, *cp2;
	unsigned char *lastwordend, *cpstart;
	register int nlines, lineno;
	unsigned char *Lines[200];
	register int width;
	int i, j;
	int state;
	register unsigned char *opbuf;
	int margin;
	int dispcolumn, origdispcol;

	if (debug)
		Printf("\014offset = %d ", offset);

	/* Work out the current control-free display column */
	dispcolumn = 0;
	for (i = 0; i < offset; i++)
	{
		if (in[i] >= ' ')
			dispcolumn++;
	}
	dispcolumn -= lmargin;

	if (debug) Printf("dispcolumn = %d, offset = %d ",
		dispcolumn, offset);
	/* Chop the line into partial lines in lines[] */
	/* States:
		ATFRONT: waiting for first non-white
		INFIRST: in first word of a partial line
		ENDLOOK: waiting for end of partial line
	 */
#define ATFRONT 0
#define INFIRST 1
#define ENDLOOK 2
	incpy = Malloc(strlen(in) + 1);
	strcpy(incpy, in);
	width = rmargin - lmargin;
	lineno = 0;
	state = ATFRONT;
	cpstart = cp = lastwordend = incpy + offset;
	origdispcol = dispcolumn;
	if (debug)
		Printf("Original: <%s>\n<%s>\n", incpy, cp);
	do
	{
		if (*cp >= ' ')
			dispcolumn++;
		switch (state)
		{
		case ATFRONT:
			if (*cp != ' ')
				state = INFIRST;
			break;
		case INFIRST:
			if (*cp == ' ' || !*cp)
			{
				lastwordend = cp;
				state = ENDLOOK;
			}
			else
			{
				if (dispcolumn >= width)
				{
					int i;
					for (i = 0; i < lineno; i++)
						Free(Lines[i]);
					return 0;	/* Can't cope */
				}
			}
			break;
		case ENDLOOK:
			if (*cp == ' ' || !*cp)
			{
				if (dispcolumn >= width)
				{	/* Break it at last word end */
					if (debug)
					{
						Printf("dc=%d,width=%d\n", dispcolumn, width);
						Printf("lastwordend=0x%x, cp=0x%x\n", lastwordend, cp);
					}
					if (dispcolumn == width)
						lastwordend = cp;
					Lines[lineno] = Malloc(width * 2);
					movmem(cpstart, Lines[lineno],
						lastwordend - cpstart);
					Lines[lineno][lastwordend-cpstart] = 0;
					if (debug)
					{
						Printf("line %d:<%s>\n", lineno, Lines[lineno]);
						Printf("*cp=0x%x\n", *cp);
						rawgetchar();
					}
					lineno++;
					if (*cp)	/* More text to go */
					{
						cp = lastwordend + 1;
						cpstart = cp;
					}
					else
						cpstart = lastwordend;
					dispcolumn = 0;
					state = INFIRST;
				}
				else
					lastwordend = cp;
			}
			break;
		}
	}
	while (*cp++);
	cp--;

	if (cpstart != cp)
	{
		Lines[lineno] = Malloc(width * 2);
		while (*cpstart == ' ')
			cpstart++;
		movmem(cpstart, Lines[lineno], cp - cpstart + 1);
		if (debug)
		{
			Printf("last line %d:<%s> ", lineno,
				Lines[lineno]);
			rawgetchar();
		}
		lineno++;
	}

	nlines = lineno;
	if (nlines == 0)
		return 0;

	if (debug)
	{
		Printf("origdispcol=%d");
		rawgetchar();
	}
	if (justifymode == J_RIGHT || justifymode == J_CENTRE)
		rjust(Lines, nlines, origdispcol);

	/* Now concatenate them */
	j = 0;
	for (i = 0; i < nlines; i++)
		j += strlen(Lines[i]) + 3 + lmargin;

	opbuf = Malloc(j);
	cp = opbuf;

	/* Number of spaces to insert in front line */
	margin = lmargin;
	if (justifymode == J_CENTRE)
		margin = 0;

	for (i = 0; i < nlines; i++)
	{
		cp2 = Lines[i];
		for (j = margin; j < lmargin; j++)
			*cp++ = 0xa0;
		margin = 0;
		while (*cp2)
			*cp++ = *cp2++;
		if (i < nlines - 1)
		{
			*cp++ = '\n' + 0x80;
		}
		Free(Lines[i]);
	}
	*cp = '\0';
	Free(incpy);
	if (debug)
		Printf("Output: <%s>\n", opbuf);
	return opbuf;
}

/*
 * Passed an array of strings.
 *
 * If we are right justifying, pad them out to the right margin,
 * else centre them
 */

rjust(Lines, nlines, offset)
register unsigned char **Lines;
register int nlines;
{
	register int width, len;
	unsigned char *Inline, *outline;
	register unsigned char *cpin, *cpout;
	int lineno;
	int linewidth;
	int lastline;
	char inisin;		    /* True if copying from Inline */

	linewidth = rmargin - lmargin;
	Inline = Malloc(linewidth * 3);
	outline = Malloc(linewidth * 3);
	lastline = (justifymode == J_CENTRE) ? nlines : nlines - 1;

	for (lineno = 0; lineno < lastline; lineno++)
	{
		if (justifymode == J_CENTRE)
		{
			cpin = Lines[lineno];
			/* Strip leading space */
			while ((*cpin & 0x7f) == ' ')
				cpin++;
			strcpy(Inline, cpin);
			width = strlen(Inline);
			/* Remove trailing space */
			while (width && ((Inline[width-1] & 0x7f) == ' '))
				width--;
			Inline[width] = '\0';
			/* Insert soft spaces */
			width = (linewidth - width) / 2;
			cpout = Lines[lineno];
			while (width--)
				*cpout++ = 0xa0;
			strcpy(cpout, Inline);
			continue;
		}

		/* Do right justification */
		width = linewidth - offset;
		offset = 0;
		if (!strchr(Lines[lineno], ' '))
		{
			if (debug) Printf("line %d:<%s>: no spaces!\n", lineno,
				Lines[lineno]);
			continue;
		}
		strcpy(Inline, Lines[lineno]);
		if (lineno & 1)
			reverse(Inline);
		/* Count length, ignoring control chars */
		len = 0;
		cpin = Inline;
		while (*cpin)
		{
			if (*cpin >= 0x20)
				len++;
			cpin++;
		}
		inisin = 1;
		while (len < width)
		{
			cpin = inisin ? Inline : outline;
			cpout = inisin ? outline : Inline;
			while (*cpin)
			{
				if (len < width && (*cpin & 0x7f) == ' ')
				{       /* Another pad */
					*cpout++ = 0xa0;
					len++;
					while ((*cpin & 0x7f) == ' ')
						*cpout++ = *cpin++;
				}
				else
					*cpout++ = *cpin++;
			}
			*cpout = 0;
			inisin ^= 1;
		}
		strcpy(Lines[lineno], (inisin) ? Inline : outline);
		if (lineno & 1)
			reverse(Lines[lineno]);
	}
	Free(Inline);
	Free(outline);
}

/*
 * Reverse a string
 */

reverse(cp)
register unsigned char *cp;
{
	register unsigned char *cp2;
	register unsigned char ch;

	cp2 = cp;
	while (*cp2)
		cp2++;
	while (cp2 > cp)
	{
		ch = *cp;
		*cp++ = *--cp2;
		*cp2 = ch;
	}
} 

/* Set left or right margin */
setmargin(isright)
{
	char buf[SEDLINESIZE];
	unsigned val;

	*buf = 0;
	tgetsn(buf, isright ? "Right margin? " : "Left margin? ");
	if (!*buf)
	{
		unsigned char *cp;

		cp = cur_lp;
		val = 0;
		while (cp < cur_ptr)
		{
			if (*cp >= ' ')
				val++;
			cp++;
		}
		val %= width;
	}
	else
		val = atoi(buf) - 1;
	if (isright)
	{	/* Right */
		if (val <= lmargin + 6)
			goto bad;
		rmargin = val;
	}
	else
	{
		if (val >= rmargin - 6)
			goto bad;
		lmargin = val;
	}
	return;
bad:
	sstatmes("\007Bad margin setting");
}

/* Centre a line */
centre()
{
	register unsigned char *cp, *cp2, *copy;
	int length, width, i, remain, sublen;

	cp = cur_lp;
	cp2 = cp;

	length = 0;
	while (*cp2 && ((*cp2 & 0x7f) != '\n'))
	{
		if ((*cp2 & 0x7f) >= ' ')
			length++;		/* Control chars don't count */
		cp2++;
	}

	/* Make a copy of the whole line */
	copy = Malloc(length + 3);
	movmem(cur_lp, copy, length);
	copy[length] = 0;

	/* Chop off leading space */
	cp = copy;
	while (*cp && ((*cp & 0x7f) == ' '))
		cp++;
	if (!*cp)
	{
		Free(copy);
		return;
	}

	/* Chop off trailing space */
	cp2 = &copy[length - 1];
	while ((*cp2 & 0x7f) == ' ')
		*cp2-- = 0;

	sublen = strlen(cp);	/* The stripped heading length */
	width = rmargin - lmargin;
	if (sublen < 1 || sublen >= width)
	{
		Free(copy);
		return;
	}

	/* Make room for the centred string */
	cp2 = Malloc(width + 3);
	remain = lmargin + (width - sublen) / 2;
	for (i = 0; i < remain; i++)
		cp2[i] = ' ';
	cp2[i] = '\0';
	strcat(cp2, cp);
	blkswap(cur_lp, length, cp2, strlen(cp2));
	Free(cp2);
	Free(copy);
	cntrlx(1);
	save_coff = 0;
	redhere();
}

/* Go to start of next paragraph */
paraend()
{
	if (!nondocmode)
	{
		register unsigned char *lastcurptr;

		lastcurptr = cur_ptr = cur_lp;
		while (strchr(cur_ptr, '\n'))	/* Is there somewhere to go? */
		{
			cntrlx(1);
			if (cur_ptr == lastcurptr)
				break;
			lastcurptr = cur_ptr;
			cur_ptr = cur_lp;
			if (cur_ptr[-1] == '\n')
			{	/* Hard newline */
				/* See if it is empty */
				if ((*cur_ptr & 0x7f) != '\n')
					break;
			}
		}
		if ((*cur_ptr & 0x7f) == ' ')
			cur_ptr = fwd_fnw(cur_ptr);
		save_coff = 0;
		redoptr();
		newcurs();
	}
	else
	{
		unsigned char *p = cur_ptr;
		
		while (p < end_buf)
		{
			if (*p == '\n')
				cntrlx(0);
			p++;
			if (	*p != '\n' &&
					p[-1] == '\n' &&
					(p > mem_buf + 1 && p[-2] == '\n'))
			{
				break;
			}
		}
		save_coff = 0;
		redoptr();
		redhere();
	}
}

/*
 * Go to start of current paragraph. If we are currently at the start of one,
 * go to the start of the previous one
 */

parastart()
{
	if (!nondocmode)
	{
		register unsigned char *cp, *cp2;

		cp = cur_lp;

		/* Search backwards for a hard newline */
		while ((cp > mem_buf) && (*cp != '\n'))
			cp--;
		if (cp == mem_buf)
			goto gotptr;

		/* Is the cursor at the start of this paragraph? */
		cp2 = cp;
		cp += 1;	/* Past newline */
		if (cp >= end_buf)
			goto backup;
		while (cp < cur_ptr)
		{
			if ((*cp & 0x7f) != ' ')
			{	/* cur_ptr is not at start of paragraph */
				goto gotptr;
			}
			cp++;
		}

		/* Need to back up another paragraph */
backup:	cp = cp2 - 1;
		while ((cp > mem_buf) && (*cp == '\n'))
			cp -= 1;		/* Skip any empty Lines */

		if (cp < mem_buf)
			cp = mem_buf;
		while ((cp > mem_buf) && (*cp != '\n'))
			cp--;
		if (cp == mem_buf)
			goto gotptr;

		cp += 1;	/* Past newline */
gotptr:
		while ((*cp & 0x7f) == ' ')
			cp++;
		cur_ptr = cp;
		save_coff = 0;
		redoptr();
		newcurs();
	}
	else
	{
		unsigned char *p = cur_ptr;
		while (p > mem_buf)
		{
			if (*p == '\n')
				cntrle(0);
			--p;
			if (p[0] != '\n' &&
				(p > mem_buf && p[-1] == '\n') &&
				(p > (mem_buf + 1) && p[-2] == '\n'))
			{
				break;
			}
		}
		save_coff = 0;
		redoptr();
		redhere();
		if (cur_y == 1)
			sendtome(CTL('W'));
	}
}

/*
 * Read margin settings from line under cursor
 */

readmargins()
{
	register unsigned char *cp;
	register int lmar, rmar;

	cp = cur_lp;
	while ((*cp & 0x7f) == ' ')
		cp++;
	lmar = cp - cur_lp;
	while ((*cp & 0x7f) != '\n' && (cp < (end_buf - 1)))
		cp++;
	/* Search back */
	while ((cp[-1] & 0x7f) == ' ')
		cp--;
	rmar = cp - cur_lp;

	if ((rmar - lmar) > 6)
	{
		rmargin = rmar;
		lmargin = lmar;
	}
	else
		sstatmes("\007Can't set margins");
}

/* strlen, ignoring control chars */

prstrlen(p)
register unsigned char *p;
{
	register int i;

	i = 0;
	while (*p)
	{
		if (*p >= ' ')
			i++;
		p++;
	}
	return i;
}
