/*	TABS4 ex: set ts=4:
 * I/O and diddly stuff
 */

#include <stdio.h>
#include <fcntl.h>
#include "../sedit.h"
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/errno.h>

#define NUNDOS	10	/* Number of undo buffers */

unsigned char *udbufs[NUNDOS];

extern unsigned char *mem_buf;	/* Start of text */
extern unsigned char *end_buf;	/* Points to null after last line */
extern int top_lin;	/* Line number for line at top of screen */
extern int bot_lin;	/* Number of line beyond last displayed */
extern int cur_lin;	/* Current */
extern unsigned char *top_lp;	/* Points to source for top line */
extern unsigned char *bot_lp;	/* Points to the string beyond the last displayed */
extern unsigned char *cur_lp;	/* Start of current line */
extern unsigned char *cur_ptr;	/* Current working pointer */
extern int cur_off;	/* cur_ptr - cur_lp (column offset of cursor) */
extern int save_coff;	/* Flag for preserving cur_off */
extern int bot_y;	/* Display line at which no text is displayed */
extern int top_y;	/* Display line of top displayed line */
extern int cur_y;	/* Current Y cursor pos */
extern int cur_x;	/* Current X cursor pos */
extern int lin_siz;	/* Number of screen lines occupied by a string */
extern int lin_dlen;	/* Number of screen unsigned chars occupied by a string */
extern int lin_len;	/* Number of real unsigned chars in a string */
extern int width;	/* Screen width */
extern int height;	/* Screen height */
extern int scrnleft;	/* Used for preventing printing from going off the bottom */
extern int max_mem;	/* Maximum allowed buffer size */
extern int ibindex;	/* Index for insert buffer */
extern int nreps;	/* Number of repeats for search/substitute */
extern int dlscren;	/* Flag: Enable scrolling in draw_lin() */
extern int modflag;	/* Flag: file modified */
extern unsigned char *blkstart;	/* Start of block */
extern unsigned char *blkend;	/* End of block */
extern unsigned char *newcur_ptr;/* Used for keeping track of where we are */
extern unsigned char **markers;	/* Markers */
extern int blkset;	/* Flag: block is marked */
extern int hilit;	/* Flag: highlight mode is set */
extern int scr_top;	/* Current screen top */
int curscrtop;
extern int frozen;	/* Flag: screen frozen */
extern int tabwidth;

extern unsigned char *statmes;	/* Top left status message */

extern int debug, rmargin, lmargin, justifymode;
extern int attrnow, attrwant;
extern int dispcontrol;
extern int overstrike;		/* Chars overtype current ones */

extern long efilestat;		/* Status bits of edited file */
extern int euserid;		/* User ID of edited file */

static unsigned char ftoolarge[] = "\007File too large";

char fileHasCR;

read_file(name, mode, isnewfile)
unsigned char *name;
int mode;		/* 0 = whole file, 1 = insert */
{
	int fd, ret;
	char *pname;

	pname = (char *)patchenv(name);
	fd = open(pname, O_RDONLY);

	if (unlink_input)
	{
		unlink(name);
		unlink_input = 0;
	}

	Free(pname);
	if (fd < 0)
	{
		statsyserr(fd);
		ret = -1;
	}
	else
	{
		if (mode == 0)
			addRollFile(name);
		ret = read_fd(fd, name, mode, 0, isnewfile);
	}

	if (isnewfile && ret < 0)
	{	/* Do we want carriage returns? */
		char *c;

		c = getenv("EDITOR_USE_CR");
		if (c)
			fileHasCR = 1;
		else
			fileHasCR = 0;
	}
	return ret;
}

statsyserr(err)
{
	extern int errno;

	strncpy(statmes, strerror(errno), 50);
	statmes[49] = '\0';
}

int read_fd(long fd, unsigned char *name, int mode, int poll, int isnewfile)
{
	unsigned char *tmp;
	register int ch;
	register int bufi;
	int nbr;
	unsigned char buf[1025];

	if (mode == 0)
	{
		struct stat statbuf;

		if (fstat(fd, &statbuf) < 0) {
			statsyserr(errno);
			return -1;
		}
		if (statbuf.st_size > max_mem) {
			sstatmes(ftoolarge);
			return -1;
		}
		nbr = read(fd, mem_buf, max_mem);
		if (nbr < 0)
		{
			statsyserr(nbr);
			return -1;
		}
		if (!nbr)
		{
			close(fd);
			return -1;
		}
		end_buf = mem_buf + nbr;
		/* Smash the carriage-returns */
		{
			register unsigned char *icp, *ocp, *ecp;

			if (isnewfile)
				fileHasCR = 0;

			icp = ocp = mem_buf;
			ecp = mem_buf + nbr;
			while (icp < ecp)
			{
				if (*icp != '\r')
					*ocp++ = *icp;
				else if (isnewfile)
					fileHasCR = 1;
				icp++;
			}
			end_buf = ocp;
		}
	}
	else
	{
		tmp = cur_ptr;
		bufi = 0;

		if (isnewfile)
			fileHasCR = 0;

		do
		{
			char ccc;

			if (mode == 1)
			{	/* Read from fd */
				if (read(fd, &ccc, 1) < 1)
					ch = '\0';
				else
					ch = ccc;
			}
			else
			{	/* Read from pipe */
				ch = getc((FILE *)fd);
				if (ch == EOF)
					ch = '\0';
			}

			if (ch == '\r')
			{
				if (isnewfile)
					fileHasCR = 1;
				continue;
			}

			if (ch)
				buf[bufi++] = ch;
			if ((bufi > 1023) || !ch)
			{		/* Move the block */
				if ((end_buf + bufi) > (mem_buf + max_mem))
				{
					sstatmes(ftoolarge);
					goto rfexit;
				}
				blkswap(tmp, 0, buf, bufi);
				tmp += bufi;
				bufi = 0;
			}
		}
		while (ch);
rfexit:		;
	}
	close(fd);
	*end_buf = 0;
	modflag++;
	return (end_buf == mem_buf) ? -1 : 0;
}

#define WBSIZE	16384

write_file(name, first, last)
unsigned char *name;
unsigned char *first, *last;
{
	register int ofd, ec;
	register unsigned char *nname;
	int ret = 0;

	if (*name == '>')
	{
		char *nn2;
		nname = (unsigned char *)patchenv(name + 1);
		nn2 = (char *)nname;
		while(iswhite(*nn2))
			nn2++;
		ofd = open(nn2, O_WRONLY|O_APPEND|O_CREAT);
	}
	else
	{
		char *nn2;
		nname = (unsigned char *)patchenv(name);
		nn2 = (char *)nname;
		while(*nn2 != '/' && iswhite(*nn2))
			nn2++;
		ofd = creat(nn2, (efilestat == -1) ? 0666 : efilestat);
	}

	if (ofd < 0)
	{
		statsyserr(ofd);
		ret = -1;
	}
	else
	{
		if (fileHasCR)
		{
			unsigned char _buf[512];
			unsigned char *buf;
			register unsigned char *icp, *ocp, *ecp;

			buf = (unsigned char *)malloc(WBSIZE);
			if (buf == 0)
			{
				buf = _buf;
				ecp = buf + sizeof(_buf);
			}
			else
				ecp = buf + WBSIZE;

			icp = first;
			ecp--;				/* ocp may overrun by 1 */
			while (icp < last)
			{
				ocp = buf;
				while (ocp < ecp && icp < last)
				{
					if (*icp == '\n')
						*ocp++ = '\r';
					*ocp++ = *icp++;
				}
				if ((ec = write(ofd, buf, ocp - buf)) < 0)
				{
					statsyserr(ec);
					ret = -1;
					break;
				}
			}
			if (buf != _buf)
				free(buf);
		}
		else if ((ec = write(ofd, first, last - first)) < 0)
		{
			statsyserr(ec);
			ret = -1;
		}

		if (ofd >= 0)
			ec = close(ofd);

		if (ret >= 0)
		{
			if (ec < 0)
			{
				statsyserr(ec);
				ret = -1;
			}
		}
	}
	Free(nname);
	return ret;
}

/*
 * Replace the bytes between (to) and (to + oldlen - 1) with the bytes between
 * (from) and (from + newlen - 1)
 */

blkswap(to, oldlen, from, newlen)
register unsigned char *to, *from;
int oldlen, newlen;
{
	register int distance;		/* How far we are moving */
	register int i;
	unsigned char *ptr;
	register unsigned char **pptr;
	register unsigned char *source, *dest;
	extern unsigned char *hilitestart, *hiliteend;

	distance = newlen - oldlen;
	source = to + oldlen;
	dest = to + newlen;

	if ((end_buf + distance) > (mem_buf + max_mem))
	{
		sstatmes(ftoolarge);
		return -1;
	}

	for (i = 0; i < 14; i++)
	{	/* Scan the markers, updating any which have moved */
		switch (i)
		{
		case 10:
			pptr = &blkstart;
			break;
		case 11:
			pptr = &blkend;
			break;
		case 12:
			pptr = &newcur_ptr;
			break;
		case 13:
			pptr = &hilitestart;
			break;
		case 14:
			pptr = &hiliteend;
			break;
		default:
			pptr = &markers[i];
			break;
		}

		if (ptr = *pptr)
		{
			if (ptr >= source)
				ptr += distance;	/* Within moved block */
			else
			{
				if ((distance < 0) && (ptr >= dest))
					ptr = dest;	/* Marker in deleted block */
			}
			*pptr = ptr;	/* New value */
		}
	}
	blkset = (blkset && (blkstart < blkend));

	/* Fix the ^QP pointers */
	doqpreloc(source, dest, distance);

	movmem(source, dest, end_buf - to - oldlen);
	end_buf += distance;
	movmem(from, to, newlen);
	*end_buf = 0;
	modflag++;
	return 0;
}

/*
 * Delete n chars from *cur_ptr. Only allow a delete up to end of line.
 * If *cur_ptr == '\n', join the two lines.
 * Redraw screen, update bottom stuff
 */

delete(n)
register int n;
{
	register int c, l;
	register int nlinnext;	/* Number of displayed lines on next line */
	register int ntodel;	/* Number of lines to delete */
	register unsigned char *tempptr;

	if (cur_ptr >= end_buf - 1)
		return;
	l = lin_siz;
	if ((tempptr = INDEX(cur_ptr, '\n')) == 0)	/* No CR at end */
		tempptr = end_buf;
	c = tempptr - cur_ptr;
	if (n > c)
		n = c;		/* Don't go beyond end of line */
	if (n == 0)
		n = 1;
	if ((c = MASK(*cur_ptr)) == '\n')
	{
		set_param(getnext(cur_lp));
		nlinnext = lin_siz;
	}
	if (n > 1)
		predelete(cur_ptr, n);
	blkswap(cur_ptr, n, cur_ptr, 0);
	bot_lp -= n;
	set_param(cur_lp);
	if ((lin_siz + cur_y) >= height)
	{	/* Joined line won't fit on screen */
		bot_lp = cur_lp;
		bot_y = cur_y;
		bot_lin = cur_lin;
		scrl_up();
	}
	else
	{
		if (c == '\n')
			bot_lin--;
		else
/* Adjust so that ntodel does not reflect length of next line */
			nlinnext = 0;	
		draw_lin(cur_ptr, cur_x, cur_y);
		if (cur_ptr == end_buf) return;
		if ((c == '\n') || (l != lin_siz))
		{	/* Need to move the rest of the screen up */
			ntodel = l + nlinnext - lin_siz;
			dellines(cur_y + lin_siz, ntodel);
			bot_y -= ntodel;
			draw(bot_lp, bot_lin, bot_y);
		} 
	}
}

/*
 * Prepare for a deletion: place the block to be deleted in the buffer
 */

predelete(ptr, nbytes)
unsigned char *ptr;
{
	register unsigned char *ret;

	if (nbytes == 1 && ptr[0] == '\n')
		return;
	if (udbufs[NUNDOS-1])
		Free(udbufs[NUNDOS-1]);	/* Dump the old one */
	ret = Malloc(nbytes + 1);
	if ((long)ret < 0)
		ret = 0;
	else
	{
		movmem(ptr, ret, nbytes);
		ret[nbytes] = 0;
	}
	/* Make room */
	movmem(udbufs, udbufs + 1, (NUNDOS - 1) * sizeof(*udbufs));
	udbufs[0] = ret;
}

/*
 * Insert the contents of an undo buffer
 */

putundobuf(bufnum)
unsigned bufnum;
{
	int ret;

	if ((bufnum >= NUNDOS) || (!udbufs[bufnum]))
		return -1;
	ret = blkswap(cur_ptr, 0, udbufs[bufnum], strlen(udbufs[bufnum]));
	draw(cur_lp, cur_lin, cur_y);
	return ret;
}

/* Print out undo buffer contents */

undoreview()
{
	int bufindex;
	register unsigned char *cp, ch;
	register short column;

	column = height - NUNDOS - 2;
	if (column < 1)
		column = 1;
	gotoxy(0, column);
	equalsigns();
	for (bufindex = 0; bufindex < NUNDOS; bufindex++)
	{
		Printf("\n%d: ", bufindex);
		if (cp = udbufs[bufindex])
		{
			column = 3;
			while ((ch = MASK(*cp++)) && (column < width - 2))
			{
				if (ch < 0x20)
				{
					if (ch == '\n')
						opchar('/');
					else if (ch == '\t')
						opchar(' ');
					else
						column--;	/* Predec */
				}
				else
					opchar(ch);
				column++;
			}
			if (ch)
				opchar('+');
		}
		clear_eol();
	}
	hil(2);
	Printf("\nHit a key ");
	resthil();
	clear_eol();
	rawgetchar();
	redwhole();
}

equalsigns()
{
	int i;

	hil(2);
	for (i = 0; i < width; i++)
		opchar('=');
	resthil();
}

bad_cmd()
{
	sstatmes("?? ^QH for help");
}

sstatmes(str, p1, p2, p3)
char *str;
{
	sprintf((char *)statmes, str, p1, p2, p3);
}

tgetsfile(buf, prompt)
unsigned char *buf, *prompt;
{
	tgets(buf, prompt);
	exwild(buf);
}

/* Call tgetsn, clearing string first */
tgets(buf, prompt)
unsigned char *buf, *prompt;
{
	*buf = '\0';
	tgetsn(buf, prompt);
}

tgetsnfile(buf, prompt)
unsigned char *buf, *prompt;
{
	tgetsn(buf, prompt);
	exwild(buf);
}

/*
 * Print a string at the top and get response.
 */

tgetsn(buf, prompt)
unsigned char *buf, *prompt;
{
	unsigned char lbuf[SEDLINESIZE];
	int len;

	gotoxy(curscrtop ? TGETPOS : 0, STATUS);
	hil(1);
	Printf(" %s", prompt);
	hil(0);
	clear_eol();
	strcpy(lbuf, buf);
	len = width - TGETPOS - strlen(prompt);
	if (len <= 15)
		len = 15;
	nledit(lbuf, len);
	resthil();
	clearmem(buf, SEDLINESIZE);
	strcpy(buf, lbuf);
	if (!curscrtop)
	{
		if (top_y != 0)
			redhere();	/* Partial line at top: messy */
		else
			draw_lin(top_lp, 0, 0);
	}
}

/*
 * Print a string and get a number. Return zero if bad number
 */

topgetnum(prompt)
char *prompt;
{
	unsigned char buf[SEDLINESIZE];

	tgets(buf, prompt);
	return atoi((char *)buf);
}

parse_sr_options(sdp, buf)
struct sr_data *sdp;
char *buf;
{
	sdp->ignorecase = 0;
	sdp->reverse = 0;
	sdp->fromtop = 0;
	sdp->ntimes = 0;
	sdp->wrap = 0;
	sdp->initialised = 1;

	while (*buf)
	{
		switch (*buf)
		{
		case 'i':
		case 'u':
		case 'c':
			sdp->ignorecase = 1;
			buf++;
			break;
		case 't':
		case 'a':
			sdp->fromtop = 1;
			buf++;
			break;
		case 'r':
		case 'b':
			sdp->reverse = 1;
			buf++;
			break;
		case 'w':
			sdp->wrap = 1;
			buf++;
			break;
		case 'g':
			sdp->ntimes = 0x7fffffff;
			buf++;
			break;
		default:
			if (*buf >= '0' && *buf <= '9')
			{
				sdp->ntimes = atoi(buf);
				while (*buf >= '0' && *buf <= '9')
					buf++;
			}
			else
				buf++;
			break;
		}
	}
	if (sdp->ntimes == 0)
		sdp->ntimes = 1;
	return;
}

/*
 * Ask for and evaluate search/replace options
 */

char *
getrepstring(srd)
struct sr_data *srd;
{
	char buf[SEDLINESIZE];

	tgets(buf, "Options? [(g)lobal (i)gnore case (t)op down (r)everse (w)rap (NN)] ");
	parse_sr_options(srd, buf);
}

/*
 * Return true if the last line in the file is terminated by a LF
 */

endatcr()
{
	return (MASK(*(end_buf - 1)) == '\n');
}

/*
 * Once-off initialisation
 */

sedinit()
{
	terminit();

	/* Enable status line */
	curscrtop = 1;

	scr_top = curscrtop;
	newcur_ptr = blkend = blkstart = 0;
	frozen = blkset = hilit = dlscren = 0;
	statmes[0] = 0;
/* Look! */
	clearmarkers();
	clearmem(udbufs, sizeof(udbufs));
	debug = 0;
}

void sedterm()
{
	int i;

	for (i = 0; i < NELEM(udbufs); i++)
	{
		if (udbufs[i])
			Free(udbufs[i]);
	}
}

clearmarkers()
{
	clearmem(markers, 40);
	blkstart = (unsigned char *)0;
	blkend = (unsigned char *)0;
}

/*
 * Search backwards for a non-white after a white or for a stoppable char
 */

unsigned char *rev_fnw(unsigned char *ptr)
{
	register int inwhite;	/* 0 = in non-white, 1 = in white */
	register int iss;

	if ((ptr == mem_buf) || ((inwhite = issorwh(*--ptr)) == 2))
		return ptr;;
	for ( ; ; )
	{
		if ((ptr == mem_buf) || (MASK(*ptr) == '\r') ||
			((iss = issorwh(*ptr)) == 2))
			return ptr;
		if (inwhite)	/* In white, waiting for non-white */
			inwhite = iss;
		else		/* In non-white, waiting for white */
			if (iss) return ptr + 1;
		ptr--;
	}
}

/*
 * Search forwards for a non-white after a white or for a stoppable unsigned char
 */

unsigned char *fwd_fnw(unsigned char *ptr)
{
	register int inwhite;	/* 0 = in non-white, 1 = in white */
	register int iss;

	inwhite = issorwh(*ptr);
	for ( ; ; )
	{
		if (ptr++ == end_buf)
			return end_buf;
		iss = issorwh(*ptr);
		if ((MASK(*ptr) == '\n') || (*ptr == '\0') || (iss == 2))
			return ptr;
		if (inwhite == 1)	/* In white space, waiting for non-white */
		{
			if (!iss)
				return ptr;
		}
		else		/* In non-white, waiting for white */
			inwhite = iss;
	}
}

/*
 * Return 1 if white, 2 if stoppable unsigned char, else 0
 */

issorwh(ch)
register unsigned ch;
{
	ch &= 0x7f;	/* With ch unsigned char, the compiler generated bad code here! */
	if ((ch == ' ') || (ch == '\t') || (ch == '\n'))
		return 1;
	return (INDEX("./!,:;?()[]{}-", ch)) ? 2 : 0;
}

/* Return true if block is set up ok */
blkok()
{
	if (blkset && blkstart && blkend && (blkstart < blkend))
		return 1;
	else
	{
		sstatmes("\007No block");
		return 0;
	}
}

/*
 * Do a block copy. The pointers are ok. Return 0 if error
 */

blkcpy(c)
char c;		/* 'C': copy, 'V': move */
{
	if
	(
		(cur_ptr >= blkstart) &&
		(
			(c == 'V' && cur_ptr <= blkend) ||
			(c == 'C' && cur_ptr < blkend)
		)
	)
	{
		sstatmes("\007In block");
		return 0;
	}
/* Note: May have to move from blkend to precompensate for the first move in blkswap() */
	blkswap(cur_ptr, 0, (cur_ptr < blkstart) ? blkend : blkstart,
		blkend - blkstart);
	return 1;
}

/*
 * Do a block delete. The pointers are OK. Return 0 if error
 */

blkdel()
{
	predelete(blkstart, blkend - blkstart);
	blkswap(blkstart, blkend - blkstart, blkend, 0);
	return 1;
}


/* Do an index, ignoring top bit */
unsigned char *
maskindex(cp, ch)
register unsigned char *cp, ch;
{
	ch &= 0x7f;
	while (*cp)
	{
		if (MASK(*cp) == ch)
			return cp;
		cp++;
	}
	return (unsigned char *)0;
}

maskstrncmp(s1, s2, n)
register unsigned char *s1, *s2;
register int n;
{
	while (n--)
	{
		if (MASK(*s1++) != MASK(*s2++))
			return 1;
	}
	return 0;
}

/* Kludge end of file */
chk_end()
{
	if (!endatcr())
		blkswap(end_buf, 0, "\n", 1);
}

/* Get a return char */
getret()
{
	char c;

	hil(2);
	Printf("edit: hit ENTER ");
	while ((c = rawgetchar()) != '\r' && (c != '\n'))
		;
	resthil();
	unfreeze(1);
}

char *
roomfor(s)
char *s;
{
	char *ret;

	ret = Malloc(strlen(s) + 1);
	strcpy(ret, s);
	return ret;
}
