/* TABS4
 * Line editor.
 */
#include <string.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <stdio.h>

#ifdef Linux
#include <dirent.h>
#define direct dirent
#define d_namlen d_reclen
#endif
#include "../sedit.h"

#define MAXIPLINE	512
#define DNLASTLINES	40	/* Default number of last lines */
#define EOFCHAR	CTL('D')

/* Editor externals */
static char *newbuf;			/* Pointer to new string */
static char *oldbuf;			/* Pointer to last displayed string */
static unsigned cursor;			/* Where the cursor is in the string */
static unsigned cursscrnpos;		/* Where the cursor is on the screen */
static unsigned oldlen;			/* Length of previously displayed string */
static unsigned len;
static unsigned nlastlines;		/* Number of last lines for recall */
static char **lastlines;		/* Most recent commands */

int inleditor;

char *roomfor();

/* Initialise the line editor stuff */
static
ledit_init()
{
	register int i;

	nlastlines = DNLASTLINES;
	lastlines = Malloc(sizeof(*lastlines) * nlastlines);
	for (i = 0; i < nlastlines; i++)
	{
		lastlines[i] = Malloc(2);
		*lastlines[i] = '\0';
	}
}

void leditEnd()
{
	int i;

	if (lastlines)
	{
		for (i = 0; i < nlastlines; i++)
		{
			if (lastlines[i])
				Free(lastlines[i]);
		}
		Free(lastlines);
		lastlines = 0;
	}
}

issep(ch)
char ch;
{
	return (ch == ' ') || (ch == '\t');
}

iswhite(ch)
char ch;
{
	return ((ch == ' ') || (ch == '\t') || (ch == '/'));
}

/*
 * Remove the char at the cursor
 */

static
remch()
{
	if (newbuf[cursor])
		strcpy(newbuf + cursor, newbuf + cursor + 1);
}

/*
 * Insert a char in the buffer
 */

static insert_ch(char ch, int length)
{
	register char *ptr;

	if (len < (length - 1))
	{
		ptr = newbuf + cursor++;
		movmem(ptr, ptr + 1, strlen(ptr) + 1);
		*ptr = ch;
		len++;
	}
}

/*
 * Restore len
 */

static
getlen()
{
	len = strlen(newbuf);
}

/*
 * Redisplay buffer
 */

static void redisp()
{
	register unsigned short cp;		/* Current buffer position */
	register unsigned short newspos;	/* Current screen position */
	register unsigned short newcurspos;	/* New cursor screen position */
	unsigned olsave;		/* Temp save for oldlen */
	register char pe;		/* Print enable */
	register unsigned char ch;		/* Current char */

	newspos = 0;
	cp = 0;
	pe = 0;

/* Scan the whole new buffer */
	do
	{
		ch = newbuf[cp];

		if (newspos == cursscrnpos)
			pe = 1;	/* Start printing when the cursor is picked up */

		if (!pe && (ch != oldbuf[cp]))
		{	/* Move the cursor back to where they differ */
			while (cursscrnpos > newspos)
			{
				rawputchar('\b');
				--cursscrnpos;
			}
			pe = 1;
		}

		if (cp == cursor)
			newcurspos = newspos;	/* New cursor position */

		if (ch == '\t')
		{
			do
				if (pe)
					rawputchar(' ');
			while (++newspos & 7);
		}
		else
		{
			if (ch)
			{
				if (ch < 0x20)
				{	/* Control char */
					if (pe)
					{
						rawputchar('^');
						rawputchar(ch | 0x40);
					}
					newspos += 2;
				}
				else
				{
					if (pe)
						rawputchar(ch);
					newspos++;
				}
			}
		}

		oldbuf[cp] = ch;
		cp++;
	}
	while (ch);

	olsave = newspos;

/* New string is out. Print any necessary spaces */
	while (newspos < oldlen)
	{
		rawputchar(' ');
		newspos++;
	}

/* Back up to correct cursor position */
	while (newspos > newcurspos)
	{
		rawputchar('\b');
		--newspos;
	}

	oldlen = olsave;
	cursscrnpos = newspos;
}

static
countwords( line )
char *line;
{
	register int count = 0;
	register char *cp;
	register int inword = 0;

	cp = line;
	while( *cp )
	{
		if( *cp != ' ' && *cp != '\t' && !inword )
		{
			count++;
			inword = 1;
		}
		else if( *cp == ' ' || *cp == '\t' ) /* End of word */
			inword = 0;

		++cp;
	}
	return( count );
}

static char wbuf[128];

static char *
getword( line, posn )
char *line;
int posn;
{
	register int cw = 0;
	register char *cp, *cp1;
	int	inword = 0;

	cp = line;
	while( *cp )
	{
		if( *cp != ' ' && *cp != '\t' && !inword )
		{
			cw++;
			if( cw == posn )
				break;
			inword = 1;
		}
		else if( *cp == ' ' || *cp == '\t' ) /* End of word */
			inword = 0;

		++cp;
	}

	/* At the beginning of the word */
	cp1 = wbuf;
	if( *cp && cw == posn )
	{
		while( *cp && *cp != ' ' && *cp != '\t' )
			*cp1++ = *cp++;

		*cp1++ = ' ';
		*cp1 = (char)0;
	}
	else
		return( (char *)0 );

	return( wbuf );
}

/* Sorting comparison function */
int
ppcmp(s1, s2)
char **s1, **s2;
{
	return strcmp(*s1, *s2);
}

static char *__nledit(char *lbuf, int length)
{
	register int ch;
	register int i;
	register unsigned llindex;	/* Index into lastlines[] */
	char *ob;		/* Copy of previously displayed string */
	char arrow;
	int nxtch = 0;
	char *oldlbuf;
	static char initted = 0;

	if (!initted)
	{
		ledit_init();
		initted = 1;
	}

	if (length >= MAXIPLINE)
		length = MAXIPLINE-1;
	ob = Malloc(MAXIPLINE);
	oldlbuf = Malloc(MAXIPLINE);
	*oldlbuf = '\0';

	llindex = nlastlines - 1;
	newbuf = lbuf;
	oldbuf = ob;
	len = strlen(lbuf);
	cursor = 0;
	oldlen = 0;
	cursscrnpos = 0;

	redisp();
leditloop:
	if (nxtch)
	{
		ch = nxtch;
		nxtch = 0;
	}
	else
	{
		if ((ch = rawgetchar()) == -1)
			ch = EOFCHAR;	/* Force end of input */
	}
	if (ch == EOFCHAR && len == 0)
		goto inp_eof;	/* EOFCHAR first entered */
	if ((ch >= ' ') || strchr("[WEFGTVXYP", ch + 0x40))
	{	/* Will alter */
		strcpy(oldlbuf, lbuf);
	}
	if (ch == CTL('U'))
	{
		strcpy(lbuf, oldlbuf);
		getlen();
		if (cursor > len)
			cursor = len;
		redisp();
		goto leditloop;
	}

	switch (ch)
	{
	case CTL('A'):	/* ^A Backward word */
		if (cursor)
		{
			do
				cursor--;
			while (iswhite(lbuf[cursor]) && cursor);

			while (cursor && (!iswhite(lbuf[cursor-1])))
				cursor--;
		}
		break;
	case CTL('B'):	/* ^B Goto end */
		cursor = cursor ? 0 : len;
		break;
	case CTL('D')|0x80:	/* ^D Forward char */
	case CTL('D'):	/* ^D Forward char */
		if (cursor < len)
			cursor++;
		break;
	case '\033':	/* ESC - search back for history */
		{
			int i, olen;
			olen = len;
			for (i = nlastlines - 1; i > 0; i--)
			{
				if (!strncmp(lbuf, lastlines[i], olen))
				{
					int ch2;
					strcpy(lbuf, lastlines[i]);
					getlen();
					cursor = len;
					redisp();
					ch2 = rawgetchar();
					if (ch2 != '\033')
					{
						nxtch = ch2;
						goto got;
					}
				}
			}
		}
got:
		break;

	case CTL('I'):		/* Filename completion */
		{
			char dirname[MAXIPLINE];
			char filename[MAXIPLINE];
			char **filenames = 0;
			int nfilenames = 0;
			int namelen;
			char *cp, *ocp, *cp2;

			cp = lbuf + cursor;
			while (cp > lbuf && !issep(*cp))
				cp--;
			if (issep(*cp))
				cp++;
			ocp = dirname;
			while (*cp && !issep(*cp))
				*ocp++ = *cp++;
			*ocp = '\0';
#if 0
			printf("dirname=%s\n", dirname);
#endif
			/* Split up the directory and filename bit */
			cp = dirname;
			while (cp2 = strchr(cp, '/'))
				cp = cp2 + 1;
			if (cp == dirname)
			{	/* No directories specified */
				strcpy(filename, dirname);
				strcpy(dirname, ".");
			}
			else
			{
				strcpy(filename, cp);
				*cp = 0;
			}
#if 0
			printf("dirname=%s, filename=%s\n", dirname, filename);
#endif
			/* Read all allowable files */
			{
				DIR *dir;
				struct direct *direct;
				int pass, Index;

				namelen = strlen(filename);
				if ((dir = opendir(dirname)) == (DIR *)0)
				{
					beep();
					break;
				}

				for (pass = 0; pass < 2; pass++)
				{
					if (pass)
						rewinddir(dir);
					if (pass == 0)
						nfilenames = 0;
					Index = 0;
					while (	(direct = readdir(dir)) &&
							(pass == 0 || (Index < nfilenames)))
					{
						if (*filename == '\0' ||
								(direct->d_namlen >= namelen &&
									!strncmp(filename, direct->d_name, namelen)))
						{	/* Acceptable name */
							if (pass == 0)
								nfilenames++;
							else
							{
								filenames[Index] =
											Malloc(direct->d_namlen + 1);
								strncpy(	filenames[Index],
											direct->d_name,
											direct->d_namlen);
								filenames[Index][direct->d_namlen] = '\0';
								Index++;
							}
						}
					}
					if (pass == 0)
						filenames = Malloc(nfilenames * sizeof(*filenames));
				}
				closedir(dir);
			}
#if 0
			{
				int ii;

				printf("%d names match:\n", nfilenames);
				for (ii = 0; ii < nfilenames; ii++)
					printf("%s\n", filenames[ii]);
			}
#endif
			/* Now start matching */
			if (nfilenames)
			{
				int fnindex = 0;
				int ii;
				int ch2;
				int olen = cursor;

				/* Alpha-sort them */
				qsort(filenames, nfilenames, sizeof(*filenames), ppcmp);

				for ( ; ; )
				{
					int isadir = 0;
					/* Insert the filename */
					for (ii = namelen; ii < strlen(filenames[fnindex]); ii++)
					{
						insert_ch(filenames[fnindex][ii], length);
					}

					{
						char *fullname;
						struct stat statbuf;
						fullname =	Malloc(strlen(dirname) +
									strlen(filenames[fnindex]) + 2);
						sprintf(	fullname,
									"%s/%s",
									dirname,
									filenames[fnindex]);
						if (stat(fullname, &statbuf) >= 0)
						{
							if (S_ISDIR(statbuf.st_mode))
							{
								insert_ch('/', length);
								isadir = 1;
							}
						}
						Free(fullname);
					}

					getlen();
					redisp();
					ch2 = rawgetchar();
					if (ch2 != CTL('I'))
					{
						if (ch2 == '\033')
						{	/* Restore old buffer */
							cursor = olen;
							for (	ii = namelen;
									ii < strlen(filenames[fnindex]);
									ii++)
								remch();
							if (isadir)
								remch();
						}
						else
							nxtch = ch2;
						goto allgot;
					}
					cursor = olen;
					for (ii = namelen; ii < strlen(filenames[fnindex]); ii++)
						remch();
					if (isadir)
						remch();
					fnindex++;
					if (fnindex >= nfilenames)
						fnindex = 0;
				}
			}
allgot:
			if (filenames)
			{
				int ii;

				for (ii = 0; ii < nfilenames; ii++)
					Free(filenames[ii]);
				Free(filenames);
			}
			break;
		}

	case CTL('O'): /* Cycle back through history argument at a time */
		{
			char *getword();
			int olen, i, j, w, k;
			char *word;

			olen = cursor;
			for( i = nlastlines - 1; i > 0; i-- )
			{
				/* Split line up into words */
				w = countwords( lastlines[i] );
				for( j = w; j > 0; j-- )
				{
					int ch2, l = 0;
					word = getword( lastlines[i], j );

					while( ch = word[l++] )
						insert_ch( ch, length );

					getlen();
					redisp();
					ch2 = rawgetchar();
					if( ch2 != CTL('O') )
					{
						nxtch = ch2;
						goto ggot;
					}
					cursor = olen;
					for( k = 0; k < strlen( word ); k++ )
						remch();
				}
			}
		}
ggot:
		break;

	case CTL('W'):	/* ^W */
		arrow = 2;
		goto udarrow;
	case CTL('E'):	/* ^E Previous line */
		arrow = 1;		/* Flag: up arrow */
		goto udarrow;
	case CTL('F'):	/* ^F Forward word */
		if (cursor != len)
		{
			do
				cursor++;
			while (!iswhite(lbuf[cursor]) && (cursor != len));
			if (cursor != len)
			{
				do
					cursor++;
				while (iswhite(lbuf[cursor]) && (cursor != len));
			}
		}
		break;
	case CTL('G'):	/* ^G, DEL Delete char forward */
		remch();
		getlen();
		break;
	case CTL('K'):	/* ^K: insert marked text into buffer */
		if (blkok())
		{
			extern unsigned char *blkstart;	/* Start of block */
			extern unsigned char *blkend;	/* End of block */
			unsigned char *p;
			for (p = blkstart; p < blkend; p++)
				insert_ch(*p, length);
		}
		break;
	case 0x7f:
	case '\b':	/* ^H, Delete char backward */
		if (cursor)
		{
			cursor--;
			remch();
			len--;
		}
		break;
	case CTL('S'):	/* ^S Backward char */
		if (cursor)
		{
			cursor--;
			redisp();
		}
		break;
	case CTL('T'):	/* ^T Delete word forward */
		if (!iswhite(lbuf[cursor]))
			while ((lbuf[cursor]) && (!iswhite(lbuf[cursor])))
				remch();
		else
			while (iswhite(lbuf[cursor]))
				remch();
		getlen();
		break;
	case CTL('V'):	/* ^V Delete line backward */
		strcpy(lbuf, lbuf + cursor);
		cursor = 0;
		getlen();
		break;
	case CTL('X'):	/* ^X Previous line recall */
		arrow = 0;
/*
 * If arrow == 0, scroll forward.
 * If arrow == 1, scroll to previous.
 * If arrow == 2, insert previous line.
 */
udarrow:	if (arrow != 2)
			lbuf[cursor = len = 0] = '\0';		/* Joke */
		if ((!arrow) && (++llindex >= nlastlines))
			llindex = 0;		/* Wrap */
		i = 0;
		while (ch = lastlines[llindex][i++])
			insert_ch(ch, length);
		if ((arrow) && (--llindex >= nlastlines))
			llindex = nlastlines - 1;
		break;
	case CTL('Y'):	/* ^Y Delete line forward */
		lbuf[cursor] = '\0';
		len = cursor;
		break;
	case '\r':	/* Terminator */
	case '\n':
inp_eof:
		cursor = len;
		redisp();	/* Move cursor to end */
		/* Don't copy empty or same string */
		if (len && strcmp(lbuf, lastlines[nlastlines - 1]))
		{
			Free(lastlines[0]);
			movmem(&lastlines[1], lastlines,
				(nlastlines - 1) * sizeof(*lastlines));
			lastlines[nlastlines - 1] = roomfor(lbuf);
		}
		rawputchar('\n');
		Free(ob);
		Free(oldlbuf);
		return ((unsigned)ch == EOFCHAR) ? 0 : lbuf;
	case CTL('P'):	/* ^P Insert control char */
		ch = rawgetchar();
		goto Default;
	case ALT('p'):	/* Insert the current filename */
		{
			char *filename = currentFileName();
			while (*filename)
				insert_ch(*filename++, length);
		}
		break;
	case ALT('P'):	/* Insert the current pathname */
		{
			char *pathname = currentPathName();
			while (*pathname)
				insert_ch(*pathname++, length);
		}
		break;
	default:	/* Normal char */
Default:
		insert_ch(ch, length);
		break;
	}
	redisp();
	goto leditloop;
}

char *nledit(char *lbuf, int length)
{
	char *ret;

	inleditor = 1;
	ret = __nledit(lbuf, length);
	inleditor = 0;
	maybegotsigwinch();
	return ret;
}

