/*	TABS4 ex: set ts=4:
 * Screen editor
 */

#include "../sedit.h"
#include <stdio.h>
#include <stdlib.h>

extern unsigned char *mem_buf;	/* Start of text */
extern unsigned char *end_buf;	/* Points to null after last line */
extern int top_lin;	/* Line number for line at top of screen */
extern int bot_lin;	/* Number of line beyond last displayed */
extern int cur_lin;				/* Current */
extern unsigned char *top_lp;	/* Points to source for top line */
extern unsigned char *bot_lp;	/* Points to the string beyond the last displayed */
extern unsigned char *cur_lp;	/* Start of current line */
extern unsigned char *cur_ptr;	/* Current working pointer */
extern int bot_y;	/* Display line at which no text is displayed */
extern int top_y;	/* Display line of top displayed line */
extern int cur_y;	/* Current Y cursor pos */
extern int cur_x;	/* Current X cursor pos */
extern int lin_siz;	/* Number of screen lines occupied by a string */
extern int lin_dlen;	/* Number of screen unsigned chars occupied by a string */
extern int width;	/* Screen width */
extern int height;	/* Screen height */
extern int nfound;	/* Number of successful searches/substitutes */
extern int modflag;	/* Flag: file modified */
extern unsigned char *blkstart;	/* Start of block */
extern unsigned char *blkend;	/* End of block */
extern unsigned char *newcur_ptr;/* Used for keeping track of where we are */
extern unsigned char **markers;	/* Markers */
extern int blkset;	/* Flag: block is marked */
extern int hilit;	/* Flag: highlight mode is set */
extern int scr_top;	/* Current screen top */
extern int curscrtop;
extern int frozen;	/* Flag: screen frozen */
extern int quitflag;	/* Flag: time to exit */
extern int tabwidth;	/* Optional tab size */
extern int auto_format;

extern int dispcontrol;
extern int overstrike;	/* Characters overtype current ones */
extern int nondocmode;

extern unsigned char *file_name;
extern unsigned char *statmes;	/* Top left status message */

extern char *newfilename;

extern unsigned char *getnext(), *get_prev();
extern unsigned char *fwd_fnw(), *rev_fnw();
extern char *roomfor();

extern char *ourname;
extern unsigned char prevfilename[];

extern int autotabs, romver;

extern int optNoviceMode;

extern int zapBeingEdited;

int nosavename = 0;
struct sr_data search_sr_data, subs_sr_data;

/*
 * Do a control-X
 */

cntrlx(dofix)
{
	if (getnext(cur_ptr) != end_buf)
	{
		cur_ptr = cur_lp = getnext(cur_ptr);
		cur_lin++;
		cur_y += lin_siz;
		if (cur_lin == bot_lin)
			scrl_up();
		set_param(cur_lp);
	}
	if (dofix)
		fix_cpos();
}

/*
 * Do a control-E command
 */

cntrle()
{
	if (cur_lp != mem_buf)
	{
		cur_ptr = cur_lp = get_prev(cur_lp);
		set_param(cur_lp);
		cur_y -= lin_siz;
		cur_lin--;
		if (cur_lin < top_lin)
			scrl_dn();
		set_param(cur_lp);
	}
	fix_cpos();
}

/*
 * Control-K and Control-Q commands
 */

cntrlkq(korq)
unsigned char korq;
{
	register int i;
	register unsigned char *tempptr;
	register unsigned char ch, origch;
	register int wantline;
	register int blksize;
	static char filenotsaved[] = "File not saved";
	extern int _Hbss, _Lbss;
	unsigned char rfile[SEDLINESIZE];

	sprintf((char *)statmes, "^%c ", korq);
	statline(cur_x, cur_y);
	ch = noFkeyRawgetchar(0xffffffff);
	if (ch != 0x1b)
		fkeyRx(ch);
	origch = ch;
	if (ch >= CTL('A') && ch <= CTL('Z'))
	{
		ch |= 0x40;
	}
	ch = toupper(ch);
	if ((ch == 0x1b) || (ch == CTL(']')))
	{
		statmes[2] = '^';
		statmes[3] = '[';
		statmes[4] = 0;
	}
	else
	{
		statmes[2] = ch;
		statmes[3] = 0;
	}
	statline(cur_x, cur_y);

	if (ch == '?')
	{
		help(korq);
		return;
	}

	if (korq == 'Q')
	{
		switch (ch)
		{
		case '%':	/* Parenthesis match */
			parenmatch(0, 0, 0);
			break;
		case 'A':	/* Substitute */
			{
				char rexp = (origch == 'A');
				if (invertRegexp())
					rexp = !rexp;
				tgets(subs_sr_data.old, rexp ? "regexp Replace? " : "Replace? ");
				if (!*subs_sr_data.old)
					break;
				tgets(subs_sr_data.new, "  With? ");
				getrepstring(&subs_sr_data);
				subs_sr_data.useregexp = rexp;
				substitute(&subs_sr_data);
			}
			break;
		case 'B':
			tempptr = blkstart;
			goto tomarker;
		case 'C':	/* To end of file */
			wantline = -1;
			goto toline;
		case 'D':	/* Forward to end of line */
			cur_ptr = getnext(cur_lp) - 1;
			break;
		case ALT('d'):
			if ((cur_ptr += 80) > (end_buf - 1))
				cur_ptr = end_buf - 1;
			if (cur_ptr > (tempptr = getnext(cur_lp) - 1))
				cur_ptr = tempptr;
			break;
		case 'E':	/* To top of screen */
			cur_lp = top_lp;
			cur_lin = top_lin;
			cur_y = top_y;
			set_param(cur_lp);
			fix_cpos();
			break;
		case 'F':	/* Find */
			{
				char rexp = (origch == 'F');
				if (invertRegexp())
					rexp = !rexp;
				tgets(search_sr_data.old, rexp ? "regexp Locate? " : "Locate? ");
				if (!*search_sr_data.old)
					break;
				getrepstring(&search_sr_data);
				search_sr_data.useregexp = rexp;
				search(&search_sr_data);
			}
			break;
		case 'G':	/* Go to line */
			wantline = topgetnum("Line? ");
toline:
			gotoLine(wantline);
			break;
		case 'H':	/* Help screen */
			helpscreen();
			break;
		case 'K':	/* Set block end marker */
			tempptr = blkend;
			goto tomarker;
		case 'L':   /* CPW: find text under cursor */
			{
				unsigned char *txtp;
#define isqltxt(c)  (isalnum(c) || (c) == '_' || (c) == '$')

				if (!isqltxt(*cur_ptr))
				{
					beep();
					break;
				}
				sendtome(CTL('Q'));
				sendtome('f');
				for (txtp = cur_ptr; isqltxt(*txtp); --txtp)
					;
				for (++txtp; isqltxt(*txtp); ++txtp)
					sendtome(*txtp);
				sendtome('\r');
				sendtome('\r');
			}
			break;
		case 'P':	/* Goto previous place */
			{
				unsigned char *qpp, *popqpsp();
				extern int wasqp;

				wasqp = 1;
				if (qpp = popqpsp())
				{
					if (qpp < mem_buf)
						qpp = mem_buf;
					if (qpp >= end_buf)
						qpp = end_buf - 1;
					cur_ptr = qpp;
					redoptr();
					redwhole();
				}
			}
			break;
		case 'R':	/* Start of file */
			wantline = 1;
			goto toline;
		case 'S':	/* Back to start of line */
			cur_ptr = cur_lp;
			break;
		case ALT('s'):
			cur_ptr -= 80;
			if (cur_ptr < cur_lp)
				cur_ptr = cur_lp;
			break;
		case 'T':		/* Set tab size */
			{
				int tw;
				char buf[40];

				tw = topgetnum(" Tabwidth? ");
				if (tw > 0 && tw < 100)
				{
					tabwidth = tw;
					redhere();
				}
				sprintf(buf, "Tabwidth is %d", tabwidth);
				sstatmes(buf);
			}
			break;
		case 'U':	/* Review undo buffers */
			undoreview();
			break;
		case 'V':
			gototag((char *)1);
			break;
		case 'X':	/* Bottom of screen */
			cur_lin = bot_lin - 1;
			cur_lp = get_prev(bot_lp);
			set_param(cur_lp);
			cur_y = bot_y - lin_siz;
			fix_cpos();
			break;
		case 'Y':
			if (MASK(*cur_ptr) != '\n')
				delete(getnext(cur_ptr) - cur_ptr);
			break;
		case 0x1b:
		case '[':	/* Goto start of this C function */
			if (cur_ptr > mem_buf)
			{
				tempptr = cur_ptr - 1;
				while ((tempptr > mem_buf) && !(*tempptr == '{' && /* } */
						tempptr[-1] == '\n'))
				tempptr--;
				if (tempptr != cur_ptr)
				{	/* Found it */
					cur_ptr = tempptr;
					redoptr();
					newcurs();
				}
			}
			break;
		case CTL(']'):
		case ']':	/* Goto start of next C function */
			tempptr = cur_ptr + 1;
			while ((tempptr < end_buf) &&
				!(*tempptr == '{' && tempptr[-1] == '\n'))	/* } */
				tempptr++;
			if (tempptr != (cur_ptr + 1) && (tempptr != end_buf))
			{	/* Found it */
				cur_ptr = tempptr;
				redoptr();
				newcurs();
			}
			break;
		default:
			if ((ch >= '0') && (ch <= '9'))
				tempptr = markers[ch - '0'];
			else
			{
				bad_cmd();
				goto out;
			}
tomarker:	if (!tempptr)
				sstatmes("\007No marker");
			else
			{
				cur_ptr = tempptr;
				redoptr();
				newcurs();
				set_param(cur_lp);
			}				
out:			break;
		}
	}
	else if (korq == 'K')
	{	/* Control-K command */
		blksize = blkend - blkstart;
		newcur_ptr = cur_ptr;
		switch (ch)
		{
		case '<':	/* Move block left a tab */
			if (blkok())
			{
				int oldSize = blkend - blkstart;
				unsigned char *buf = Malloc(oldSize);
				unsigned char *to = buf;
				unsigned char *from = blkstart;

				while (from < blkend)
				{
					if (*from == '\t')
						from++;
					/* Advance to next line */
					if (from < blkend)
					{
						do
						{
							*to++ = *from++;
						}
						while (from < blkend && from[-1] != '\n');
					}
				}
				blkswap(blkstart, oldSize, buf, to - buf);
				cur_ptr -= (oldSize - (to - buf));
				Free(buf);
				redoptr();
				redhere();
			}
			break;
		case '>':	/* Move block right a tab */
			if (blkok())
			{
				int oldSize = blkend - blkstart;
				unsigned char *buf = Malloc(oldSize * 2);
				unsigned char *to = buf;
				unsigned char *from = blkstart;

				while (from < blkend)
				{
					if (*from != '\n')
						*to++ = '\t';
					/* Advance to next line */
					if (from < blkend)
					{
						do
						{
							*to++ = *from++;
						}
						while (from < blkend && from[-1] != '\n');
					}
				}
				blkswap(blkstart, oldSize, buf, to - buf);
				cur_ptr -= (oldSize - (to - buf));
				Free(buf);
				redoptr();
				redhere();
			}
			break;
		case '!':	/* Read from piped command */
			tgets(rfile, "Command? ");
			if (rfile[0])
			{
				FILE *f;
				f = popen((char *)rfile, "r");
				if (f)
				{
					read_fd((long)f, rfile, 2, 0, 0);
					draw(cur_lp, cur_lin, cur_y);
					pclose(f);
				}
			}
			break;
		/* CPW: new command ^K' */
		case '|':	/* Do command on block */
			if (!blkok())
				break;
			tgets(rfile, "Command? ");
			if (rfile[0])
			{
				FILE *f;
				char tmp[L_tmpnam];

				sprintf(tmp, "dde1%d", getpid());

				/* write block to temp file */
				if (write_file(tmp, blkstart, blkend) == 0)
				{
					/* set cmd stdin from it */
					strcat(strcat(rfile, " <"), tmp);

					f = popen((char *)rfile, "r");
					if (f)
					{
						read_fd((long)f, rfile, 2, 0, 0);
						pclose(f);
					}
				}
				unlink(tmp);
			}
			break;
		case '.':	/* Replay capture buffer */
			replaycapture();
			break;
		case '=':	/* Edit previous file */
			if (modflag && writecurrentfile(	(char *)0,
												mem_buf,
												end_buf))
				break;		/* Write problem */
			modflag = 0;
			donewfile("#");
			break;
		case 'B':	/* Mark block start */
			if (blkset)
			{
				if (cur_ptr < blkstart)
				{	/* Grown block: redraw up to old start */
					blkstart = cur_ptr;
					redblock();
					blkset = 1;
				}
				else if (cur_ptr < blkend)
				{	/* Shrunk block */
					blkstart = cur_ptr;
					redhere();
					blkset = 1;
				}
				else
				{	/* Block not displayable */
					blkset = 0;
					redblock();
					blkstart = cur_ptr;
				}
			}
			else
			{
				blkstart = cur_ptr;
				if (blkend && blkend > blkstart)
				{
					blkset = 1;
					redblock();
				}
			}
			break;
		case 'C':	/* Block copy */
			if (blkok() && blkcpy('C'))
				goto movend;
			break;
		case 'D':	/* File write */
			if (!writecurrentfile(file_name, mem_buf, end_buf))
				modflag = 0;
			break;
		case 'F':	/* Freeze screen */
			if (frozen)
				unfreeze(1);
			else
			{
				if ((i = cur_y + cur_x / width) > (curscrtop))
				{
					scr_top = i + 1;
					frozen = 1;
					if (scr_top > (height - 10))
						scr_top = (height - 10);
					/* Draw frozen screen separator */
					gotoxy(0, scr_top - 1);
					equalsigns();
					redraw(cur_lp, cur_lin, (scr_top + height)/2);
				}
			}
			break;
		case 'G':	/* Edit file named by marked block */
			if (!blkok())
			{
				beep();
			}
			else
			{
				sendtome(CTL('K'));
				sendtome(CTL('N'));
				sendtome(CTL('K'));
				sendtome('\r');
			}
			break;
		case 'H':	/* Set / clear block markers */
			if (blkset)
			{
				blkset = 0;
				redblock();
			}
			else if (blkstart && blkend && (blkstart < blkend))
			{
				blkset = 1;
				redblock();
			}
			break;
		case 'I':	/* Indefinite shell escape */
			gotoxy(0, height - 2);
			printf("Type '^D' to return to editor  ");
			if (modflag)
			{
				hil(1);
				printf(filenotsaved);
			}
			hil(0);
			clear_eol();
			flushop();
			savestate();
			termend();
			printf("\n");
			flushop();
			{
				char *shell;
				shell = (char *)getenv("SHELL");
				if (shell == (char *)0)
					shell = "/bin/sh";
				system(shell);
			}
			termstart();
			reststate();
			unfreeze(1);
			break;
		case 'K':	/* Block end */
			if (blkset)
			{
				if (cur_ptr > blkend)
				{	/* Grown */
					blkend = cur_ptr;
					redblock();
				}
				else if (cur_ptr > blkstart)
				{	/* Shrunk */
					blkend = cur_ptr;
					redhere();
				}
				else
				{	/* Block not displayable */
					blkset = 0;
					redblock();
					blkend = cur_ptr;
				}
			}
			else
			{
				blkend = cur_ptr;
				if (blkstart && (blkstart < cur_ptr))
				{
					blkset = 1;
					redblock();
				}
			}
			break;	
		case 'N':	/* Switch to new file */
			donewfile(0);
			break;
		case 'O':
			if (blkok())
				formatblock();
			break;
		case 'P':	/* Put block in undo buffer */
			if (blkok())
				predelete(blkstart, blkend - blkstart);
			break;
		case 'Q':	/* Quit */
			doquit(origch);
			break;
		case 'R':	/* Read file */
			tgetsfile(rfile, "Filename? ");
			if (*rfile)
			{
				read_file(rfile, 1, 0);
				draw(cur_lp, cur_lin, cur_y);
			}
			break;
		case 'S':		/* Toggle status line display */
			if (curscrtop == 0)	/* It is currently off */
				curscrtop = 1;
			else
				curscrtop = 0;
			if (!frozen)
				scr_top = curscrtop;
			if (curscrtop == 0)
				redhere();
			break;
		case 'T':	/* Goto tag */
			if (origch == 'T')
			{	/* Goto next occurrence of previous tag */
				gototag((char *)1);
			}
			else
			{	/* Goto new tag */
				tgets(rfile, "Tag? ");
				if (*rfile)
					gototag(rfile);
			}
			break;
		case 'U':	/* Case conversion */
			if (blkok() && (origch != CTL('U')))
			{
				if (origch == 'U')
				{
					for (tempptr = blkstart; tempptr < blkend;
							tempptr++)
						if (*tempptr >= 'a' && *tempptr <= 'z')
						{
							*tempptr -= 'a' - 'A';
							modflag++;
						}
				}
				else
				{
					for (tempptr = blkstart; tempptr < blkend;
							tempptr++)
						if (*tempptr >= 'A' && *tempptr <= 'Z')
						{
							*tempptr += 'a' - 'A';
							modflag++;
						}
				}
				redblock();
			}
			break;
		case 'V':	/* Block move */
			if (blkok() && blkcpy('V') && blkdel())
			{
movend:
				cur_ptr = newcur_ptr - blksize;
				blkstart = cur_ptr;
				blkend = cur_ptr + blksize;
				blkset = 0;
				redoptr();
				redhere();
			}
			break;
		case 'W':	/* Block write */
			if (blkok())
			{
				nosavename = 1;
				*rfile = '\0';	/* Force new filename */
				writecurrentfile(rfile, blkstart, blkend);
				nosavename = 0;
			}
			break;
		case 'X':
			if (!writecurrentfile(file_name, mem_buf, end_buf))
			{
				modflag = 0;
				doquit(origch);
			}
			break;
		case 'Y':	/* Block delete */
			if (blkok() && blkdel())
			{
				cur_ptr = newcur_ptr;
				redoptr();
				redhere();
			}
			break;
		default:
			if ((ch >= '0') && (ch <= '9'))
			{
				markers[ch-'0'] = cur_ptr;
			}
			else
				bad_cmd();
			break;
		}	/* switch */
		newcur_ptr = 0;	/* Speed up block swapping code (a bit) */
	}		/* else */
	else
	{	/* ^O functions */
		extern int debug, justifymode;
		switch (ch)
		{
		case '@':
			clr_scrn();
			dumpRollFiles();
			gotoxy(0, height - 1);
			getaret();
			redhere();
			break;
		case 0x1b:		/* Function key mapping */
			mapFkey();
			break;
		case ' ':
			completion();
			break;
		case '(':		/* ) Go back a block */
			parenmatch("}", "{", BACKWARD);
			break;
		case ')':		/* ( Go fwd a block */
			parenmatch("{", "}", FORWARD);
			break;
		case 'A':		/* Toggle autoformat */
			auto_format = !auto_format;
			break;
		case 'B':
			if (atFirstRollFile())
				beep();
			else
				donewfile(roll(-1));
			break;
		case 'C':
			centre();
			break;
		case 'D':
			if (!nondocmode)
				dispcontrol = !dispcontrol;
			redwhole();
			break;
		case 'E':
			parastart();
			break;
		case 'F':
			if (atLastRollFile())
				beep();
			else
				donewfile(roll(1));
			break;
		case 'I':
			overstrike = !overstrike;
			break;
		case 'J':
			justifymode++;
			if (justifymode == J_MAX)
				justifymode = 0;
			break;
		case 'L':
			setmargin(0);
			break;
		case 'M':
			readmargins();
			break;
		case 'N':
			nondocmode = !nondocmode;
			if (nondocmode)
			{
				dispcontrol = 1;
				if (tabwidth == 5)
					tabwidth = 8;
			}
			redwhole();
			break;
		case 'O':
			reformat(1);
			break;
		case 'R':
			setmargin(1);
			break;
		case 'T':
			autotabs = !autotabs;
			break;
		case 'X':
			paraend();
			break;
		case 'Z':		/* Zap file state */
			zapBeingEdited = 1;
			saveFileState(file_name);
			zapBeingEdited = 0;
			break;
		default:
			bad_cmd();
			break;
		}
	}
	chk_end();
	return;
}

/*
 * Write out the current file: passed filename.
 * If name == 0, write out using current filename. Otherwise use 'name' as prompt.
 * Return 0 if OK
 */

writecurrentfile(name, first, last)
char *name;
unsigned char *first, *last;
{
	char fname[SEDLINESIZE];
	unsigned char bufile[SEDLINESIZE];

	if (name)
	{
		strcpy(fname, name);
		if (!optNoviceMode)
			tgetsnfile(fname, "Filename? ");
	}
	else
		strcpy(fname, file_name);

	if (*fname)
	{
/*		if (!nosavename)
			strcpy(file_name, fname);
*/
		strcpy(bufile, fname);
		return write_file(fname, first, last);
	}
	return 1;
}

/* Switch to a new file */
donewfile(name)
char *name;
{
	unsigned char rfile[SEDLINESIZE];

	if (name)
		strcpy(rfile, name);
	else
		tgetsfile(rfile, "New file? ");
	if (!strcmp(rfile, "#"))
		strcpy(rfile, prevfilename);

	if (*rfile)
	{
		char *cc;

		if (modflag)
		{
			if (!writecurrentfile((char *)0, mem_buf, end_buf))
				modflag = 0;
			else
				return;
		}
		cc = (char *)patchenv(rfile);
		strcpy(rfile, cc);
		Free(cc);
		newfilename = roomfor(rfile);
		strncpy(prevfilename, file_name, SEDLINESIZE);
		prevfilename[SEDLINESIZE - 1] = '\0';
		doquit('N');
	}
}

/* Quit from editor or file */
doquit(origch)
unsigned char origch;
{
	if (modflag)
	{	/* Not written out */
		sstatmes("\007Not saved yet");
		modflag = 0;
	}
	else
	{
		gotoxy(0, height - 1);
		hil(0);
		clear_eol();
		quitflag = 1;		/* Flag to main loop */
	}
}

/* Unfreeze the screen */
unfreeze(doredraw)
{
	frozen = 0;
	scr_top = curscrtop;
	if (doredraw)
		redhere();
}

gotoLine(wantline)
{
	if (!wantline)
		wantline = 1;
	if (wantline < 0)
		wantline = 1000000;	/* Look! */
	if (wantline <= top_lin)
	{	/* Search from front of file */
		cur_lin = 1;
		cur_lp = cur_ptr = mem_buf;
	}
	else
	{
		cur_lin = top_lin;
		cur_lp = cur_ptr = top_lp;
	}
	while ((wantline > cur_lin) && (cur_lp < end_buf))
	{
		cur_lp = getnext(cur_lp);
		cur_lin++;
	}
	if (cur_lp == end_buf)
	{
		cur_lp = get_prev(cur_lp);
		cur_lin--;
	}
	cur_ptr = cur_lp;
	newcurs();
	set_param(cur_lp);
	fix_cpos();
}

int
backable(c)
char c;
{
	if (c >= 'a' && c <= 'z')
		return 1;
	if (c >= 'A' && c <= 'Z')
		return 1;
	if (c == '_' || c == '$')
		return 1;
	return 0;
}

int
stoppable(c)
char c;
{
	if (c == ' ' || c == '\t' || c == '\n' || c == ';')
		return 1;
	return 0;
}

completion()
{
	unsigned char *thisPat = cur_ptr;

	if (thisPat == mem_buf)
		return;
	thisPat--;
	while (thisPat > mem_buf && backable(thisPat[-1]))
		thisPat--;

	if (thisPat != cur_ptr)
	{	/* Search backwards for the same pattern */
		unsigned char *probe = thisPat - 1;
		int len = cur_ptr - thisPat;
		while (probe >= mem_buf)
		{
			if (!strncmp(probe, thisPat, len) && !stoppable(probe[len]))
			{	/* Found it */
				probe += len;
				while (!stoppable(*probe))
					sendtome(*probe++);
				break;
			}
			probe--;
		}
	}
}
