#include <stdio.h>
#include <sys/ioctl.h>
#if defined(linux) || defined(NAV)
#include <sys/time.h>
#else
#include <time.h>
#endif

#include "../sedit.h"

int sigwinchkludge;

/* Push-ahead buffer: power of 2 */
static unsigned char getcbuf[CAPTURESIZE];
#define GETCMASK (sizeof(getcbuf) - 1)
int getchead = 0, getctail = 0;
extern int inleditor;
extern int blkset;
extern int cur_x, cur_y;
extern unsigned char *cur_ptr;
extern unsigned char *end_buf;
extern unsigned char *blkstart;	/* Start of block */
extern unsigned char *blkend;	/* End of block */
extern unsigned char *scan_xy();
extern int cur_off;

/* Push a char into input buffer */
sendtome(c)
unsigned char c;
{
	if (((getchead + 1) & GETCMASK) == getctail)
		return;
	getcbuf[getchead++] = c;
	getchead &= GETCMASK;
}

stringtome(s)
char *s;
{
	while (*s)
		sendtome(*s++);
}

ipbready()
{
	return getchead != getctail;
}

/*
 * getchar() ignoring control-D eof char
 */

timedRead(timeout)
unsigned timeout;
{
	fd_set readfds;
	struct timeval tv;
	unsigned char ret;
	int nfound;

	if (timeout == 0xffffffff)
	{
		flushop();
		for ( ; ; )
		{
			int readret;
			readret = read(0, &ret, 1);
			if (readret < 0)
			{
				if (sigwinchkludge)
				{
					sigwinchkludge = 0;
					flushop();
					continue;
				}
				else
				{
					fprintf(stderr, "read() returns %d\n", readret);
					doexit(-1);
				}
			}
			else
			{
				sigwinchkludge = 0;
				break;
			}
		}
		if (ret == 0)
			ret = ' ';		/* Control-space */
	}
	else
	{
		FD_SET(0, &readfds);
		tv.tv_sec = timeout / 1000;
		tv.tv_usec = (timeout % 1000) * 1000;
		nfound = select(1, &readfds, 0, 0, &tv);
		if (nfound == -1)
		{
			extern int errno;
			write(2, "Select() failure\n", 17);
#if 0
			fprintf(stderr, "errno=%d\n", errno);
#endif
			ret = -1;
		}
		else
		{
			if (nfound)
			{
				flushop();
				read(0, &ret, 1);
				if (ret == 0)
					ret = ' ';		/* Control-space */
			}
			else
			{
				ret = 0;
			}
		}
	}
	return (int)ret;
}

#define NFKEYS		12
#define FKEYSIZE	1000

int fkeyRecording, currentFkey, currentFkeyLen;
char fkeys[NFKEYS][FKEYSIZE];

noFkeyRawgetchar(timeout)
{
	unsigned int eofchar, ch, Stdin;

retry:
	if (getchead != getctail)
	{
		ch = getcbuf[getctail++];
		getctail &= GETCMASK;
	}
	else
	{
		int cc;

#if 0
		if (sigwinchkludge)
		{	/* Got a sigwinch while doing something: gobble the char */
			cc = timedRead(0xffffffff);
			sigwinchkludge = 0;
		}
		do
		{	/* Handle SIGWINCH while sitting in the read */
			sigwinchkludge = 0;
			cc = timedRead(timeout);
		} while (sigwinchkludge);
#else
		cc = timedRead(timeout);
#endif
		ch = cc;
	}
	if (ch == '\n')
		ch = '\r';		/* Unix is silly sometimes */

/* fprintf(stderr, "noFkeyRawGetchar() returns 0x%x\n", ch); */
	return ch;
}

fkeyRx(ch)
{
	if (fkeyRecording)
	{
		if (currentFkeyLen < (FKEYSIZE - 1))
		{
			fkeys[currentFkey][currentFkeyLen++] = ch;
			fkeys[currentFkey][currentFkeyLen] = '\0';
		}
	}
}

static void getnums(n, n0, n1, n2, n3)
int n;
int *n0, *n1, *n2, *n3;
{
	int i;
	for (i = 0; i < n; i++)
	{
		int acc = 0;
		int c;

		do
		{
			c = noFkeyRawgetchar(0xffffffff);
			if (c >= '0' && c <= '9')
				acc = acc * 10 + c - '0';
		}
		while (c >= '0' && c <= '9');

		switch (i)
		{
		case 0: *n0 = acc; break;
		case 1: *n1 = acc; break;
		case 2: *n2 = acc; break;
		case 3: *n3 = acc; break;
		}
	}
}

/* Get a char from the keyboard.  Map function keys */
_rawgetchar()
{
	static int ch1 = 0, ch2 = 0, ch3 = 0, ch4 = 0;
	int ret;

	if (ch2)
	{
		ret = ch2;
		ch2 = 0;
		return ret;
	}
	if (ch3)
	{
		ret = ch3;
		ch3 = 0;
		return ret;
	}
	if (ch3)
	{
		ret = ch3;
		ch3 = 0;
		return ret;
	}

	ch1 = noFkeyRawgetchar(0xffffffff);
	if (ch1 == 0x1b)
	{
#ifdef NAV
		ch2 = noFkeyRawgetchar(300);
#else
		ch2 = noFkeyRawgetchar(100);
		if (ch2 == 0)
			ch2 = noFkeyRawgetchar(100);
#endif
		if (ch2 == 0)
		{	/* Timeout */
			ch2 = 0;
		}
		else if (ch2 == '[')
		{
			ch3 = noFkeyRawgetchar(0xffffffff);
			ch4 = -1;
			if (ch3 == '1' || ch3 == '2')
			{
				ch4 = noFkeyRawgetchar(0xffffffff);
				if (ch4 == '~')
					goto ugh;
				if (ch4 >= '0' && ch4 <= '9')
				{
					noFkeyRawgetchar(0xffffffff);	/* Gobble '~' */
					if (fkeyRecording)
					{
						stopFkeyRecording();
						ch2 = ch3 = ch4 = 0;
						ch1 = rawgetchar();
					}
					else
					{
						gotFkey(fkeyNum(ch3, ch4));
						ch2 = ch3 = ch4 = 0;
						ch1 = rawgetchar();
					}
				}
			}
			else
			{
ugh:
				switch (ch3)
				{
				case 'A': 								/* up arrow */
					ch2 = ch3 = ch4 = 0;
					ch1 = CTL('E'); break;
				case 'B':								/* down arrow */
					ch2 = ch3 = ch4 = 0;
					ch1 = CTL('X'); break;
				case 'C':
					ch2 = ch3 = ch4 = 0;
					ch1 = CTL('D')|0x80; break;				/* forward arrow */
				case 'D':
					ch2 = ch3 = ch4 = 0;
					ch1 = CTL('S'); break;				/* back arrow */
				case '2':								/* Insert */
				case '4':								/* home */
				case '7':								/* Linux home */
				case '5':								/* page up */
				case '6':								/* page down */
				case '8':								/* End */
					if (ch4 == -1)
						ch4 = noFkeyRawgetchar(0xffffffff);
					if (ch4 == '~')
					{
						switch (ch3)
						{
						case '2':						/* Insert becomes forward delete */
							ch2 = ch3 = ch4 = 0;
							ch1 = CTL('G');
							break;
						case '4':						/* Home */
						case '7':
							ch2 = ch3 = ch4 = 0;
							ch1 = CTL('L');
							break;
						case '5':
							ch2 = ch3 = ch4 = 0;
							ch1 = CTL('R');
							break;
						case '6':
							ch2 = ch3 = ch4 = 0;
							ch1 = CTL('C');
							break;
						case '8':						/* End */
							ch2 = ch3 = ch4 = 0;
							ch1 = CTL('B');
							break;
						}
					}
					break;
				}
			}
		}
		else if (ch2 == 'P')
		{	/* ESCPx;yP : pick position */
			int x, y;

			mergeinsbuffer();
			getnums(2, &x, &y, 0, 0);
			if (inleditor)
			{
				beep();
			}
			else
			{
				unsigned char *cp = scanxy(x, y);
				cur_ptr = cp;
				if (cur_ptr >= end_buf)
					cur_ptr = end_buf - 1;
				redoptr();
				newcurs();
				statline(cur_x, cur_y);
				cur_off = cur_x;
			}
			ch1 = ch2 = ch3 = ch4 = 0;
			return _rawgetchar();
		}
		else if (ch2 == 'S')
		{	/* ESCSxs;ys;xe;yeS :select text */
			int xs, ys, xe, ye;

			mergeinsbuffer();
			getnums(4, &xs, &ys, &xe, &ye);
			if (inleditor)
			{
				beep();
			}
			else
			{
				if (blkset)
				{
					blkset = 0;
					redblock();
				}

				blkstart = scanxy(xs, ys);
				blkend = scanxy(xe, ye);

				if (blkset)
				{
					blkset = 0;
					redblock();
				}
				if (blkstart < blkend)
				{
					blkset = 1;
					redblock();
				}

				cur_ptr = blkstart;
				if (cur_ptr >= (end_buf - 2))
					cur_ptr = end_buf - 2;
				redoptr();
				newcurs();
				statline(cur_x, cur_y);
				cur_off = cur_x;
			}
			ch1 = ch2 = ch3 = ch4 = 0;
			return _rawgetchar();
		}
	}

	fkeyRx(ch1);
	return ch1;
}

startFkeyRecording(k)
{
	if (k < NFKEYS)
	{
		fkeyRecording = 1;
		currentFkey = k;
		fkeys[currentFkey][0] = '\0';
		currentFkeyLen = 0;
	}
}

stopFkeyRecording()
{
	fkeyRecording = 0;
}

char *
getFkey(k)
{
	return fkeys[k];
}

int
fkeyNum(ch2, ch3)
{
	int fkey = 0;
	switch ((ch2 - '0') * 10 + ch3 - '0')
	{
	case 11: fkey = 1; break;
	case 12: fkey = 2; break;
	case 13: fkey = 3; break;
	case 14: fkey = 4; break;
	case 15: fkey = 5; break;
	case 17: fkey = 6; break;
	case 18: fkey = 7; break;
	case 19: fkey = 8; break;
	case 20: fkey = 9; break;
	case 21: fkey = 10; break;
	case 23: fkey = 11; break;
	case 24: fkey = 12; break;
	}
	return fkey;
}

gotFkey(fkey)
{
	stringtome(getFkey(fkey - 1));
}

/* Start recording a function key. The ESC has been received */
mapFkey()
{
	int ch2, ch3, ch4;

	ch2 = noFkeyRawgetchar(0xffffffff);
	if (ch2 == '[')
	{
		ch3 = noFkeyRawgetchar(0xffffffff);
		if (ch3 == '1' || ch3 == '2')
		{
			ch4 = noFkeyRawgetchar(0xffffffff);
			if (ch4 >= '0' && ch4 <= '9')
			{
				int fkey;
				noFkeyRawgetchar(0xffffffff);		/* '~' */
				if (fkeyRecording)
					stopFkeyRecording();
				else
				{
					fkey = fkeyNum(ch3, ch4) - 1;
					if (fkey >= 0 && fkey < 12)
					{
						startFkeyRecording(fkey);
					}
				}
			}
			else
			{
				sendtome(ch2);
				sendtome(ch3);
				sendtome(ch4);
			}
		}
		else
		{
			sendtome(ch2);
			sendtome(ch3);
		}
	}
	else
	{
		sendtome(ch2);
	}
}
