#include <stdio.h>

#include <string.h>
#include "../sedit.h"

typedef struct rollf
{
	struct rollf *prev;
	struct rollf *next;
	char *name;
} rollf;

static rollf *here;

extern unsigned char *file_name;

/*
 * Constructor
 */
static rollf *
newRollf(name)
char *name;
{
	rollf *ret = zMalloc(sizeof(*ret));
	ret->name = roomfor(name);
	return ret;
}

static rollf *
rollfStart(void)
{
	rollf *rf = here;

	if (rf)
	{
		while (rf->prev)
			rf = rf->prev;
	}
	return rf;
}

void freeRollFiles()
{
	rollf *rf = rollfStart();

	while (rf)
	{
		rollf *next = rf->next;
		Free(rf->name);
		Free(rf);
		rf = next;
	}
}

/*
 * Remove 'name' from the list
 */
deleteRollFile(name)
char *name;
{
	rollf *rf = rollfStart();

	if (rf)
	{
		do
		{
			if (!strcmp(rf->name, name))
			{
				if (rf->prev)
					rf->prev->next = rf->next;
				if (rf->next)
					rf->next->prev = rf->prev;
				Free(rf->name);
				Free(rf);
				break;
			}
			rf = rf->next;
		}
		while (rf);
	}
}

/*
 * We have just switched to a new file.
 * Add the passed name to the roll Files.
 * We put it between 'here' and 'here + 1', then
 * advance 'here' to point to it.
 */
addRollFile(name)
char *name;
{
	if (here == 0)
	{
		here = newRollf(name);
/*		here->prev = here->next = here;
*/
	}
	else if (!strcmp(name, here->name))
	{	/* Nothing to do */
	}
	else
	{
		rollf *rf = here;

		deleteRollFile(name);		/* Prepare to move the file */
		rf = newRollf(name);
		rf->next = here->next;
		rf->prev = here;
		here->next = rf;
		if (rf->next)
			rf->next->prev = rf;
		here = rf;
	}
}

char *
firstRollFile(void)
{
	here = rollfStart();
	return here->name;
}

char *
roll(dirn)
int dirn;
{
	if (here == 0)
		return (char *)file_name;
	if (dirn > 0)
	{	/* Forwards */
		if (here->next)
			here = here->next;
		return here->name;
	}
	else
	{	/* Backwards */
		if (here->prev)
			here = here->prev;
		return here->name;
	}
}

void dumpRollFiles(void)
{
	rollf *rf;

	Printf("RollFiles:\n");
	for (rf = rollfStart(); rf; rf = rf->next)
	{
		Printf(	"%c %s\n",
				(rf == here) ? '*' : ' ',
				rf->name);
	}
}

int atFirstRollFile(void)
{
	return here->prev == 0;
}

int atLastRollFile(void)
{
	return here->next == 0;
}
