/*	TABS4 ex: set ts=4:
 * Screen editor functions
 */

#include "../sedit.h"
#include <string.h>
#include <stdio.h>

extern unsigned char *mem_buf;	/* Start of text */
extern unsigned char *end_buf;	/* Points to null after last line */
extern int top_lin;	/* Line number for line at top of screen */
extern int bot_lin;	/* Number of line beyond last displayed */
extern int cur_lin;	/* Current */
extern unsigned char *top_lp;	/* Points to source for top line */
extern unsigned char *bot_lp;	/* Points to the string beyond the last displayed */
extern unsigned char *cur_lp;	/* Start of current line */
extern unsigned char *cur_ptr;	/* Current working pointer */
extern int cur_off;	/* cur_ptr - cur_lp (column offset of cursor) */
extern int save_coff;	/* Flag for preserving cur_off */
extern int bot_y;	/* Display line at which no text is displayed */
extern int top_y;	/* Display line of top displayed line */
extern int cur_y;	/* Current Y cursor pos */
extern int cur_x;	/* Current X cursor pos */
extern int lin_siz;	/* Number of screen lines occupied by a string */
extern int lin_dlen;	/* Number of screen unsigned chars occupied by a string */
extern int lin_len;	/* Number of real unsigned chars in a string */
extern int width;	/* Screen width */
extern int height;	/* Screen height */
extern int scrnleft;	/* Used for preventing printing from going off the bottom */
extern int max_mem;	/* Maximum allowed buffer size */
extern int ibindex;	/* Index for insert buffer */
extern int nreps;	/* Number of repeats for search/substitute */
extern int nfound;	/* Number of successful searches/substitutes */
extern int dlscren;	/* Flag: Enable scrolling in draw_lin() */
extern int modflag;	/* Flag: file modified */
extern unsigned char *blkstart;	/* Start of block */
extern unsigned char *blkend;	/* End of block */
extern unsigned char *newcur_ptr;/* Used for keeping track of where we are */
extern unsigned char **markers;	/* Markers */
extern int blkset;	/* Flag: block is marked */
extern int hilit;	/* Flag: highlight mode is set */
extern int scr_top;	/* Current screen top */
extern int curscrtop;
extern int frozen;	/* Flag: screen frozen */
extern int tabwidth;

extern unsigned char *file_name;
extern unsigned char *statmes;	/* Top left status message */

extern int attrnow, attrwant;
int dispcontrol;
int overstrike;		/* Chars overtype current ones */
static char hilblockonly = 0;	/* Flag: only redraw highlit block */
extern int nondocmode, autotabs;
extern int lowbaud;

int is_xterm = 0;

void xterm_title(char *app_name, char *path_name)
{
	char *file_name = strrchr(path_name, '/');

	if (file_name == 0)
	{
		file_name = path_name;
	}
	else
	{
		file_name++;
		if (*file_name == 0)	/* Ended with a '/' */
		{
			file_name--;
			while ((file_name >path_name) && (file_name[-1] != '/'))
			{
				file_name--;
			}
		}
	}

	{
		char *buf = (char *)malloc(	strlen(path_name) +
									strlen(file_name) +
									30);
		char *leader;

		if (app_name == 0 || app_name[0] == 0)
		{
			leader = 0;
		}
		else
		{
			leader = (char *)malloc(strlen(app_name) + 2);
			sprintf(leader, "%s ", app_name);
		}

		sprintf(	buf,
					"\033]1;%s%s\007" "\033]2;%s%s\007",
					leader ? leader : "",
					file_name,
					leader ? leader : "",
					path_name);
		SCREENWRITE(buf, strlen(buf));
		free(buf);
		if (leader)
			free(leader);
	}
}

void set_name(char *n)
{
	if (is_xterm)
	{
		static char *last_file_name;

		if (last_file_name == 0 || 
			strcmp(last_file_name, file_name) != 0)
		{
			xterm_title("dde", file_name);

			if (last_file_name)
				free(last_file_name);
			last_file_name = strdup(file_name);
		}
	}
}

/* Return two-thirds of the current page size (for control-C & R) */
mpaglen()
{
	return (2 * (height - scr_top)) / 3;
}

unsigned char *hilitestart, *hiliteend;

/*
 * Draw the line at *lp on the screen at position d, y. Stop on \r. Return the
 * new position
 */

draw_lin(lp, d, y)
register unsigned char *lp;
register int d;
register int y;
{
	register unsigned char ch;
	register char tabo;
	char dohil;
	int d2, y2;

	scrnleft = (height - y) * width - d;
	if (term_scrolls_at_last_col())
		scrnleft--;
	dohil = (blkset || hilit);	/* Call 'sethil' if true */
	gotoxy(d, y);

	while (scrnleft || dlscren)
	{
		if (dohil && hildif(lp))
		{
			if (hilit)
				hilitestart = lp;
			else
				hiliteend = lp;
			hil(hilit);
		}
		if ((ch = (nondocmode ? *lp : MASK(*lp))) < 0x20)
		{
			if (!dispcontrol && !nondocmode)
			{
				char match = 1;
				switch (ch)
				{
				case CC_BOLD:
					if (attrnow & BOLD)
						attrwant &= ~BOLD;
					else
						attrwant |= BOLD;
					break;
				case CC_UNDERLINE:
					if (attrnow & UNDERLINE)
						attrwant &= ~UNDERLINE;
					else
						attrwant |= UNDERLINE;
					break;
				case CC_SUBSCRIPT:
					if (attrnow & SUBSCRIPT)
						attrwant &= ~SUBSCRIPT;
					else
						attrwant |= SUBSCRIPT;
					break;
				case CC_SUPERSCRIPT:
					if (attrnow & SUPERSCRIPT)
						attrwant &= ~SUPERSCRIPT;
					else
						attrwant |= SUPERSCRIPT;
					break;
				case CC_ITALICS:
					if (attrnow & ITALICS)
						attrwant &= ~ITALICS;
					else
						attrwant |= ITALICS;
					break;
				case CC_WIDE:
					if (attrnow & WIDE)
						attrwant &= ~WIDE;
					else
						attrwant |= WIDE;
					break;
				default:
					match = 0;
					break;
				case '\n':
				case '\0':
					goto docont;
				}
				if (match)
				{
					setbul();
				}
			}
			else
			{
docont:
				switch (ch)
				{
				case '\t':
					if (!nondocmode)
					{
						if (dispcontrol)
							goto uparrow;
						break;
					}
					tabo = tabwidth - (d % tabwidth);
					d += tabo;
					while (tabo--)
						lputchar(' ');
					break;
				case '\n':
				case '\0':
					tclrhil();
					if (!(d == width && !term_scrolls_at_last_col()))
						clear_eol();
					resthil();
					if (!nondocmode)
					{
						if (*lp == '\n')
						{	/* Hard newline */
							if (attrnow)
							{
								attrwant = 0;
								setbul();
							}
						}
						y2 = y + d / width;
						d2 = d % width;
						if (*lp == '\n' &&
							d2 <= (width - 1) && y2 < (height - 1))
						{
							gotoxy(width - 1, y2);
							opchar('<');
						}
					}
					goto out;
				default:
uparrow:
					lputchar('^');
					lputchar(ch | 0x40);
					d += 2;
					break;
				}
			}
		}
		else
		{
			if (ch == ' ')
				lputchar(ch);
			else
				lputchar(*lp);
			d++;
		}
		lp++;
	}
out:
	return d;
}

/*
 * If there is room on the screen, print a char
 *
 * ONLY draw_lin CAN CALL THIS FUNCTION
 */

lputchar(ch)
unsigned char ch;
{
	if (scrnleft)
	{
		opchar(ch);
		scrnleft--;
	}
	else
	{
		if (dlscren)	/* Make more room */
		{
			up_scrl(1);
			gotoxy(width - 1, height - 2);
			opchar(ch);
			scrnleft--;
		}
	}
}

/*
 * If the passed pointer is in the block, set highlight mode
 */

sethil(ptr)
unsigned char *ptr;
{
	if (hildif(ptr))
		hil(hilit);
}

hildif(ptr)
unsigned char *ptr;
{
	if (blkset && (ptr >= blkstart) && (ptr < blkend))
	{
		if (!hilit)
		{
			hilit = 1;
			return 1;
		}
	}
	else
	{
		if (hilit)
		{
			hilit = 0;
			return 1;
		}
	}
	return 0;
}

/*
 * Draw whole screen, with the line at *lp starting at display line y. Update
 * all the red tape for screen management. The line at *lp has number lin
 */

redraw(lp, lin, y)
unsigned char *lp;
int lin, y;
{
	register unsigned char *cp;

	top_lp = lp;
	top_y = y;
	top_lin = lin;
	for (;;)
	{	/* Back up through file until we go off the top */
		if ((top_y == scr_top) || (top_lp == mem_buf))
			break;
		set_param(cp = get_prev(top_lp));
		if ((top_y - lin_siz) < scr_top)
			break;
		top_y -= lin_siz;
		top_lin--;
		top_lp = cp;
	}
	top_y = scr_top;
	draw(top_lp, top_lin, scr_top);	/* Redraw the whole lot */
}

/*
 * Redraw whole screen with current line in middle
 */

redwhole()
{
	register int draw_y;

	draw_y = (height + scr_top) / 2;	/* Where to centre things */
	set_param(cur_lp);
	if ((lin_siz + draw_y) >= height)
		draw_y = height - lin_siz;
	redraw(cur_lp, cur_lin, draw_y);
}

/* Redraw screen, only updating marked block */
redblock()
{
	if ((get_prev(top_lp) > blkend) || (bot_lp < blkstart))
		return;
	hilblockonly = 1;
	redhere();
	hilblockonly = 0;
}

/*
 * Redraw the screen at the current cursor position, provided the whole line
 * fits on screen
 */

redhere()
{
	set_param(cur_lp);
	if ((cur_y + lin_siz) > height)
		cur_y = height - lin_siz;
	redraw(cur_lp, cur_lin, cur_y);
}

/*
 * Draw the screen from display row y to the bottom. Update display red tape,
 * set cur_lp if found
 */

draw(lp, lin, y)
unsigned char *lp;
int lin;
register int y;
{
	bot_lin = lin;
	bot_lp = lp;

	while (y < height)
	{
		if (bot_lp == cur_lp)
		{
			cur_y = y;	/* Found our current line */
			cur_lin = bot_lin;
		}

		if (bot_lp == end_buf)
		{
			bot_y = y;
			do
				draw_lin(".\n", 0, y++);
			while (y < height && !lowbaud);
			if (lowbaud)
			{
				clear_eos(0, y);
				y = height;
			}
		}
		else
		{
			unsigned char *nxtline = getnext(bot_lp);

			set_param(bot_lp);
			if (hilblockonly)
			{	/* Only draw if highlit block involved */
				if (bot_lp <= blkend && nxtline >= blkstart)
					draw_lin(bot_lp, 0, y);
			}
			else
				draw_lin(bot_lp, 0, y);
			if (y + lin_siz > height)	/* Partial line */
				y = height;
			else
			{
				bot_y = y += lin_siz;
				bot_lp = nxtline;
				++bot_lin;
			}
		}
	}
}

/*
 * Scan a line, set up cur_x (if found), set lin_siz = number of screen rows
 * occupied, lin_dlen = printed length of line, lin_len = real length.
 * Set up 'wantloc' to point to char at horizontal offset 'wantx', 'wanty'.
 */

int curcur;		/* Character offset into line where cursor is */
int wantx;		/* X coord we are searching for */
int wanty;		/* Y coord we are searching for */
unsigned char *wantloc;		/* Where wantx corresponds to */

set_param(lp)
register unsigned char *lp;
{
	register int dlen;	/* Displayed length */
	register int len;	/* Real length */
	register unsigned char c;
	int wantdlen;
	unsigned char *olp;		/* Original lp */

	olp = lp;
	dlen = len = 0;
	curcur = 0;
	wantdlen = wanty * width + wantx;
	for (;;)
	{
		if (lp == cur_ptr)	/* Found our X pos */
			cur_x = dlen;
		if (dlen <= cur_off)
			curcur = lp - olp;	/* Offset position */
		if (dlen <= wantdlen)
			wantloc = lp;

		if ((c = MASK(*lp++)) < 0x20)
		{
			switch(c)
			{
			case '\t':
				if (nondocmode)
				{
					dlen += tabwidth;
					dlen -= dlen % tabwidth;
				}
				else if (dispcontrol)
				{	/* doc mode, displaying as ^I */
					dlen += 2;
				}
				len++;
				break;
			case '\n':
			case '\0':
				if (term_scrolls_at_last_col())
				{
					lin_siz = dlen / width + 1;
				}
				else
				{
					lin_siz = (dlen - 1) / width + 1;
				}
				lin_dlen = dlen;
				lin_len = len;
				return;
			default:
				if (dispcontrol || nondocmode)
					dlen += 2; /* For ^X type output */
				len++;
				break;
			}
		}
		else
		{
			dlen++;
			len++;
		}
	}
}

/*
 * Preserve the vertical column by moving cursor
 */

fix_cpos()
{
	set_param(cur_lp);
	if (cur_off > lin_dlen)	/* This line too short */
		cur_ptr = cur_lp + lin_len;
	else
		cur_ptr = cur_lp + curcur;
	set_param(cur_lp);
	save_coff = 1;		/* Prevent alteration of cur_off */
}

/*
 * Scan from *lp for the next line of source. Return a pointer to it, else
 * to the end of the buffer
 */

unsigned char *getnext(unsigned char *lp)
{
	*end_buf = 0;
	lp = INDEX(lp, '\n');
	return (lp == 0) ?  end_buf : lp + 1;
}

/*
 * With lp pointing to the start of a line, scan backwards looking for the
 * start of the previous line. 
 */

unsigned char *get_prev(unsigned char *lp)
{
	lp -= 1;	/* Now points to \n */
	while (MASK(*--lp) != '\n' && lp >= mem_buf)
		;
	return (lp < mem_buf) ? mem_buf : lp + 1;
}

/*
 * Scroll up sufficiently to display *bot_lp. Reset bottom stuff, y stuff,
 * current stuff.
 */

scrl_up()
{
	register int want;

	set_param(bot_lp);
	want = lin_siz - (height - bot_y);
	if (want < 0)
		want = 0;
	up_scrl(want);
	draw(bot_lp, bot_lin, bot_y - want);
	top_y -= want;
	while (top_y < scr_top)
	{
		set_param(top_lp);
		top_y += lin_siz;
		top_lp = getnext(top_lp);
		top_lin++;
	}
}

/*
 * When called, cur_y is the y coord of the desired line, and may be negative.
 * If it is negative, scroll down until we just fit. If a sub-line is pushed
 * off, redraw bottom
 */

scrl_dn()
{
	while (cur_y < scr_top)
	{
		++cur_y;
		dn_scrl();
		if (++bot_y > height)
		{
			bot_lp = get_prev(bot_lp);
			--bot_lin;
			set_param(bot_lp);
			bot_y -= lin_siz;
		}
	}
	top_lp = cur_lp;
	top_y = scr_top;
	top_lin = cur_lin;
	draw_lin(cur_lp, 0, scr_top);
}

/*
 * The cursor has changed. If still on the screen then move it, else
 * redraw the screen
 */

newcurs()
{
	register unsigned char *tempptr;

#ifdef notdef
cur_lp = cur_ptr;
while (cur_lp[-1] != '\n' && cur_lp > mem_buf)
	cur_lp--;
#endif

	if ((cur_ptr < bot_lp) && (cur_ptr >= top_lp))
	{	/* New string is on screen. Find its location */
		tempptr = top_lp;
		cur_y = top_y;
		while (tempptr <= cur_ptr)
		{
			set_param(tempptr);
			cur_y += lin_siz;
			tempptr = getnext(tempptr);
		}
		cur_y -= lin_siz;	/* Adjust for going beyond */
	}
	else
		redwhole();
}

/*
 * Prepare the search/substitute result message
 */

prep_nfound()
{
	sprintf(statmes, "%u found", nfound);
}

/*
 * Refind the current position, set up cur_lp, cur_lin.
 */

redoptr()
{
	register unsigned char *tempptr;
	register unsigned i;

	i = cur_ptr - mem_buf;	/* Number of bytes to scan */
	cur_lin = 1;
	cur_lp = tempptr = mem_buf;
	while (i--)
	{	/* More chars to go */
		if (MASK(*tempptr++) == '\n')
		{
			cur_lin++;
			cur_lp = tempptr;
		}
	}
}

up_scrl(n)
int n;
{
	dellines(scr_top, n);
}

/* If highlight set, clear it */
tclrhil()
{
	if (hilit)	/* Select black background */
		hil(0);
}

/* Restore highlight */
resthil()
{
	hil((hilit) ? 1 : 0);	/* Select the background */
}

/* Set up attributes based on number of ^B's, ^U's etc up to cursor */
setattr()
{
	register unsigned char *cp;
	register char bold, ul, sub, sup, it, wide;

	bold = ul = sub = sup = it = wide = 0;
	cp = cur_ptr;
	while ((cp > mem_buf) && (*cp != '\n'))
		cp--;
	while (cp < cur_ptr)
	{
		if (*cp == CC_BOLD)
			bold++;
		else if (*cp == CC_UNDERLINE)
			ul++;
		else if (*cp == CC_SUBSCRIPT)
			sub++;
		else if (*cp == CC_SUPERSCRIPT)
			sup++;
		else if (*cp == CC_ITALICS)
			it++;
		else if (*cp == CC_WIDE)
			wide++;
		cp++;
	}
	attrwant = 0;
	if (bold & 1)
		attrwant |= BOLD;
	if (ul & 1)
		attrwant |= UNDERLINE;
	if (sub & 1)
		attrwant |= SUBSCRIPT;
	if (sup & 1)
		attrwant |= SUPERSCRIPT;
	if (it & 1)
		attrwant |= ITALICS;
	if (wide & 1)
		attrwant |= WIDE;
}

int
statline(x, y)
{
	int i;
	char buf[200];
	char *jmode;
	extern int justifymode, nondocmode;

	if (curscrtop)
	{
		unsigned char *sm = statmes;

		gotoxy(0, STATUS);
		hil(2);

		if (nondocmode)
			jmode = (autotabs) ? "AI" : "NI";
		else
		{
			if (justifymode == J_RAGGED)
				jmode = "RR";
			else if (justifymode == J_CENTRE)
				jmode = "CJ";
			else
				jmode = "RJ";
		}
#ifndef SHORTEN_FILENAME     /* CPW */
        {
            char short_name[SEDLINESIZE];
            int level = 1;

            strcpy(short_name, file_name);
            while (strlen(short_name) >= (width - 30))
            {
                char *fncopy = strdup(file_name);
                char *pieces[40];
                int n, n2;
                
                pieces[0] = strtok(fncopy + 1, "/") - 1;
                for (n = 1; (pieces[n] = strtok(NULL, "/")) != NULL; ++n)
                    ;
                /* always include 1st & last section of path plus filename */
                if (((n - 3) - level) <= 0)
                {
                    free(fncopy);       /* nothing else left to remove */
                    break;
                }
                strcpy(short_name, pieces[0]);
                strcat(short_name, "/...");
                for (n2 = level; n2 < n; ++n2)
                    strcat(strcat(short_name, "/"), pieces[n2]);
                free(fncopy);
                ++level;
            }

            /* same as below except "short_name" instead of "file_name" */
            sprintf(buf, "%-4s Ch=%-5u Li=%-4u Col=%-3u '%s' %2d%%%c<%s><%s><%s>",
                sm,
                cur_ptr-mem_buf+1+ibindex,
                cur_lin,
                cur_ptr - cur_lp + 1 + ibindex,
                short_name,
                (100 * (cur_ptr - mem_buf)) / (end_buf - mem_buf),
                modflag ? ' ' : '_',
                overstrike ? "OS" : "IN",
                jmode,
                nondocmode ? "ND" : "DM");
        }
#else
        sprintf(buf, "%-4s Ch=%-5u Li=%-4u Col=%-3u '%s' %2d%%%c<%s><%s><%s>",
            sm,
            cur_ptr-mem_buf+1+ibindex,
            cur_lin,
            cur_ptr - cur_lp + 1 + ibindex,
            file_name,
            (100 * (cur_ptr - mem_buf)) / (end_buf - mem_buf),
            modflag ? ' ' : '_',
            overstrike ? "OS" : "IN",
            jmode,
            nondocmode ? "ND" : "DM");
#endif

		i = strlen(buf);
		if (i >= width)		/* Adjust for esc sequences */
			i = width - 1;
		SCREENWRITE(buf, i);
		if (i < (width - 1))
			clear_eol();
		resthil();
	}
	else
	{
		if (statmes[0] == '\007')
			beep();
	}
	gotoxy(x, y);

	set_name(file_name);
}

/* Find address of a char from screen X & Y */

unsigned char *
scanxy(x, y)
{
	unsigned char *cp, *cp2;
	int ypos;
	extern unsigned char *top_lp, *getnext();
	extern int top_y;
	extern int lin_siz;
	extern int wantx, wanty;
	extern unsigned char *wantloc;

	cp = top_lp;
	ypos = top_y;

	for ( ; ; )
	{
		set_param(cp);
		if (lin_siz + ypos > y)
			break;	/* Got the line */
		ypos += lin_siz;
		cp2 = getnext(cp);
		if (cp2 == cp)
			return cp2;
		cp = cp2;
	}

	/* cp points to desired line */
	wantx = x;
	wanty = y - ypos;
	set_param(cp);
	return wantloc;
}
