/*	TABS4
 * Searching / substituting code
 */

#include <stdio.h>
#include "../sedit.h"
#include "regex.h"

extern unsigned char *mem_buf;				/* Start of text */
extern unsigned char *cur_ptr;
extern unsigned char *end_buf;
extern unsigned char *blkstart, *blkend;		/* Block pointers */
extern int nfound;

static unsigned char uppercase[256];

#define NLFUDGE		0x0e			/* ^N char */

#ifndef isdigit
#define isdigit(c) ((c)>='0' && (c)<='9')
#endif

static unsigned char *
buildupcase()
{
	static char doneit = 0;

	if (!doneit)
	{
		int i;
		for (i = 0; i < 256; i++)
		{
			if (i >= 'a' && i <= 'z')
				uppercase[i] = (i - 'a') + 'A';
			else
				uppercase[i] = i;
		}
		doneit = 1;
	}
	return uppercase;
}

/*
 * Search for the string specified by the passed structure
 * Return 0 or pointer
 */

search(sdp)
struct sr_data *sdp;
{
	extern int nodrawonsearch;
	unsigned char firstnl;
	int nreps;
	int wrap;
	char reverse;
	struct re_pattern_buffer rpb;

	rpb.buffer = 0;

	nfound = 0;
	if (!sdp->initialised)
	{
		sstatmes("\007No previous search");
		return;
	}

	nreps = sdp->ntimes;
	wrap = sdp->wrap;
	if (usuallyWrap())
		wrap = !wrap;
	reverse = sdp->reverse;

	firstnl = *(sdp->old);
	fixnl(sdp->old);
	if (*(sdp->old))
	{
		if (sdp->useregexp && !reverse)
		{	/* Regular expression */
			struct re_registers regs;
			char fastmap[(1 << 8)];
			char *msg;
			int ret;
			int nnewlines;
			register unsigned char *pp;
			unsigned char *ptr;

			rpb.allocated = 40;
			rpb.buffer = Malloc(rpb.allocated);
			rpb.fastmap = fastmap;
			rpb.translate = sdp->ignorecase ?
								(char *)buildupcase() :
								(char *)0;

			msg = re_compile_pattern(sdp->old, strlen(sdp->old), &rpb);
			if (msg != (char *)0)
			{
				sstatmes("\007Compile failed: %s ", msg);
				goto done;
			}
			re_compile_fastmap(&rpb);

			pp = (unsigned char *)sdp->old;
			nnewlines = 0;
			while (*pp)
				if (*pp++ == '\n')
					nnewlines++;

			ptr = cur_ptr + 1;	/* Start after cursor char */
			while (nreps)
			{
				long range;		/* Size of string to match */
				long nn;

				pp = ptr;
				nn = nnewlines;
				while (pp < end_buf && nn >= 0)
				{
					if (*pp == '\n')
						nn--;
					pp++;
				}
				range = pp - ptr;
				/* re_search (pbufp, string, size, startpos, range, regs) */
				ret = re_search(&rpb, ptr, end_buf - ptr - 1,
							0, range, &regs);
				if (ret < 0)
				{
					ptr = getnext(ptr);
					if (ptr >= end_buf)
						break;
				}
				else
				{	/* Got a match */
					cur_ptr = ptr + regs.start[0];
					ptr += regs.end[0] + 1;
					nreps--;
					nfound++;
				}
			}
		}
		else
		{	/* Not regular expression */
			while (nreps--)
			{
				if (locate(sdp, 1))
					nreps = 0;	/* Force an end */
				else
					nfound++;
			}
		}
	}
	if (nfound && firstnl == CTL('N') && cur_ptr < end_buf)
		cur_ptr++;		/* Go to next line if newline */
	if (nfound && !nodrawonsearch)
	{
		redoptr();
		newcurs();
	}
	prep_nfound();
done:
	if (rpb.buffer)
		Free(rpb.buffer);
	return;
}

/*
 * Regexp search/replace
 */

static int
regsubstitute(sdp)
struct sr_data *sdp;
{
	struct re_pattern_buffer rpb;
	struct re_registers regs;
	char fastmap[(1 << 8)];
	char *msg;
	int ret;
	int nnewlines;
	register unsigned char *pp;
	unsigned char *ptr;
	unsigned char *startptr, *endptr;
	int nreps, wrap;
	char useblock;

	/* Representation of the registered replacement string */
	struct newbit
	{
		char isreg;		/* Flag: is a register spec rather than a string */
		unsigned char *text;	/* Should be a union... */
		int textlen;
		int regnum;
		struct newbit *next;
	} *newstr = 0, *nb;

	rpb.buffer = 0;
	nfound = 0;
	wrap = sdp->wrap;
	if (usuallyWrap())
		wrap = !wrap;
	nreps = sdp->ntimes;
	fixnl(sdp->new);
	fixnl(sdp->old);
	if (strchr(sdp->old, '\r') || strchr(sdp->old, '\n'))
	{
		sstatmes("\007Newlines not permitted");
		goto done;
	}
	if (!*(sdp->old))
		goto done;

	rpb.allocated = 40;
	rpb.buffer = Malloc(rpb.allocated);
	rpb.fastmap = fastmap;
	rpb.translate = sdp->ignorecase ? (char *)buildupcase() : (char *)0;

	msg = re_compile_pattern(sdp->old, strlen(sdp->old), &rpb);
	if (msg != (char *)0)
	{
		sstatmes("\007Compile failed: %s ", msg);
		goto done;
	}
	re_compile_fastmap(&rpb);

	/*
	 * Prepare the replacement string structure array
	 * by parsing the user input
	 */

	pp = (unsigned char *)sdp->new;				/* Input */
	while (*pp)
	{
		nb = Malloc(sizeof(*nb));
		nb->next = (struct newbit *)0;
		if (*pp == '\\' && isdigit(pp[1]))
		{	/* A register */
			nb->isreg = 1;
			nb->regnum = pp[1] - '0';
			pp += 2;
		}
		else
		{	/* Normal text */
			unsigned char *ppstart;
			unsigned char *in, *out;

			ppstart = pp;
			while (*pp && !(*pp == '\\' && isdigit(pp[1])))
			{
				if (*pp == '\\' && pp[1] == '\\')
					pp++;		/* Escaped '\' */
				pp++;
			}
			nb->isreg = 0;
			nb->textlen = pp - ppstart;
			nb->text = Malloc(nb->textlen + 1);
			strncpy(nb->text, ppstart, nb->textlen);
			nb->text[nb->textlen] = '\0';		/* Terminate */
			/* remove double '\' */
			in = out = nb->text;
			while (*in)
			{
				if (*in == '\\' && in[1] == '\\' && isdigit(in[2]))
					in++;
				*out++ = *in++;
			}
			*out = '\0';
		}

		/* Append to list */
		if (newstr == (struct newbit *)0)
			newstr = nb;
		else
		{
			struct newbit *nb2;

			nb2 = newstr;
			while (nb2->next)
				nb2 = nb2->next;
			nb2->next = nb;
		}
	}

#ifdef REDEBUG
	fprintf(stderr, "newstr:\n");
	for (nb = newstr; nb; nb = nb->next)
	{
		if (nb->isreg)
			fprintf(stderr, "Reg: %d\n", nb->regnum);
		else
			fprintf(stderr, "Text: <%s> (%d)\n", nb->text, nb->textlen);
	}
#endif
	if (useblock = blkok())
	{
		startptr = blkstart;
		endptr = blkend;
	}
	else
	{
		startptr = sdp->fromtop ? mem_buf : cur_ptr;
		endptr = end_buf;
	}

	signalled();
	ptr = startptr;
#ifdef REDEBUG
	fprintf(stderr, "nreps=%d\n", nreps);
#endif
	while (nreps && !signalled())
	{
		long range;		/* Size of string to match */
		long nn;

#ifdef REDEBUG
		fprintf(stderr, "Starting search from %d\n", ptr - mem_buf);
#endif
		endptr = useblock ? blkend : end_buf;		/* This can change */

		/* If the expression starts with '^', move to the start of a line */
		if (sdp->old[0] == '^')
		{
			if ((ptr > mem_buf && ptr[-1] == '\n') || ptr == mem_buf)
				;
			else
			{
				while ((ptr < end_buf) && (ptr < endptr) && (*ptr++ != '\n'))
					;
			}
		}

		/* If the expression ends in a '$', don't match if we're at the end of a line now */
		if (sdp->old[strlen(sdp->old) - 1] == '$' && ptr < end_buf)
			ptr++;

#ifdef REDEBUG
		fprintf(stderr, "Advanced to %d\n", ptr - mem_buf);
#endif
		pp = ptr;

		/* If the expression ends in a '$', place end pointer at end of previous line */
		if (sdp->old[strlen(sdp->old) - 1] == '$')
		{
			while (endptr >= ptr && endptr >= mem_buf && *endptr != '\n')
				endptr--;
		}

		if (ptr >= endptr)
			break;

		/* Work out endpoint */
		while (pp < endptr)
		{
			if (*pp == '\n')
				break;
			pp++;
		}
		range = pp - ptr;
#ifdef REDEBUG
		fprintf(stderr, "Searching..\n");
#endif
		ret = re_search(&rpb, ptr, endptr - ptr, 0, range, &regs);
		if (ret < 0)
		{
			ptr = getnext(ptr);
#ifdef REDEBUG
			fprintf(stderr, "new ptr: %d\n", ptr - mem_buf);
#endif
			if (ptr >= endptr)
				break;
		}
		else
		{	/* Got a match */
			unsigned char *newtext, *pnewtext;
			int newlen;

#ifdef REDEBUG
			fprintf(stderr, "Match\n");
			{
				int i;
				for (i = 0; i < RE_NREGS; i++)
					fprintf(stderr, "regs.start[%d]=%d, regs.end[%d]=%d\n",
							i, regs.start[i], i, regs.end[i]);
			}
#endif
			/* Work out how long it's going to be */
			newlen = 0;
			for (nb = newstr; nb; nb = nb->next)
			{
				if (nb->isreg)
					newlen += regs.end[nb->regnum] -
								regs.start[nb->regnum];
				else
					newlen += nb->textlen;
			}
			newlen++;
#ifdef REDEBUG
			fprintf(stderr, "newlen=%d\n", newlen);
#endif
			/* Make it! */
			pnewtext = newtext = Malloc(newlen);
			*newtext = '\0';
			for (nb = newstr; nb; nb = nb->next)
			{
				int iii;

				if (nb->isreg)
				{
					for (iii = regs.start[nb->regnum];
							iii < regs.end[nb->regnum]; iii++)
						*pnewtext++ = ptr[iii];
				}
				else
				{
					strcpy(pnewtext, nb->text);
					while (*pnewtext)
						pnewtext++;
				}
			}
			*pnewtext = '\0';
#ifdef REDEBUG
			fprintf(stderr, "newtext: <%s>\n", newtext);
#endif
			/* Insert it::::: blkswap(to, oldlen, from, newlen) */

			blkswap(ptr + regs.start[0], regs.end[0] - regs.start[0],
						newtext, pnewtext - newtext);
			{
				int advance = regs.start[0] + (pnewtext - newtext);
				if (advance == 0)
					advance = 1;
				cur_ptr = ptr + advance;
/*				ptr += regs.end[0] ? regs.end[0] : 1;*/
				ptr = cur_ptr;
			}
			Free(newtext);
			nreps--;
			nfound++;
		}
	}

	nb = newstr;
	while (nb)
	{
		struct newbit *nb2;

		nb2 = nb->next;
		if (!nb->isreg)
			Free(nb->text);
		Free(nb);
		nb = nb2;
	}

	if (nfound)
	{
		redoptr();
		redwhole();
	}
	prep_nfound();
done:
	if (rpb.buffer)
		Free(rpb.buffer);
	return;
}

/*
 * Search/replace
 */

substitute(sdp)
struct sr_data *sdp;
{
	int oldlen, newlen;
	unsigned char *cpsave;
	int nreps, wrap;

	if (!sdp->initialised)
	{
		sstatmes("\007No previous replace");
		return;
	}

	if (sdp->useregexp)
	{
		regsubstitute(sdp);
		return;
	}

	wrap = sdp->wrap;
	if (usuallyWrap())
		wrap = !wrap;

	nreps = sdp->ntimes;
	nfound = 0;
	fixnl(sdp->new);
	fixnl(sdp->old);
	if (!*(sdp->old))
		return;

	oldlen = strlen(sdp->old);
	newlen = strlen(sdp->new);
	signalled();			/* Clear flag */
	if (blkok())
	{	/* Replace within marked block */
		cpsave = cur_ptr;
		cur_ptr = blkstart;
	}
	else
		cpsave = 0;

	while (!signalled() && nreps-- && !locate(sdp, 0))
	{
		if (cpsave)
		{
			if (cur_ptr >= blkend)
				break;
			cpsave = cur_ptr;
		}

		if (nreps == 0)
			predelete(cur_ptr, oldlen);
		blkswap(cur_ptr, oldlen, sdp->new, newlen);
		cur_ptr += newlen;
		nfound++;
	}

	if (cpsave)
		cur_ptr = cpsave;
	if (nfound)
	{
		redoptr();
		redwhole();
	}
	prep_nfound();
}

/*
 * Search from cur_ptr+off for the passed string.
 */

locate(sdp, offset)
struct sr_data *sdp;
int offset;
{
	register unsigned char *str;
	register unsigned char *loc_ptr;
	register int strsize;
	register unsigned char *in, *out;
	int ret;

	str = (unsigned char *)sdp->old;
	strsize = strlen(str);
	loc_ptr = cur_ptr + (sdp->reverse ? -offset : offset);

	if (loc_ptr >= end_buf)
		return 1;

	if (sdp->ignorecase)
	{
		unsigned char *up;
		register unsigned char *str2;
		unsigned char c;

		up = buildupcase();
		str2 = Malloc(strsize + 1);
		in = str;
		out = str2;
		do						/* Make upper case string */
			*out++ = up[*in];
		while (*in++);
		c = *str2;

		for ( ; ; )
		{
			if (signalled())
				break;
			/* Scan for first char */
			if (sdp->reverse)
			{
				while ((loc_ptr > mem_buf) && up[*loc_ptr] != c)
					loc_ptr--;
			}
			else
			{
				while (*loc_ptr && up[*loc_ptr] != c)
					loc_ptr++;
			}
			if ((loc_ptr < mem_buf) || !*loc_ptr)
			{	/* Not found */
				ret = 1;
				break;
			}

			in = str2;
			out = loc_ptr;
			while (*in)
			{
				if (up[*out] != *in)
					break;
				out++;
				in++;
			}
			if (!*in)
			{	/* Found it */
				cur_ptr = loc_ptr;
				ret = 0;
				break;
			}
			if (sdp->reverse)
				loc_ptr--;
			else
				loc_ptr++;
		}
		Free(str2);
	}
	else
	{	/* Case sensitive */
		unsigned char c;

		c = *str;

		for ( ; ; )
		{
			if (signalled())
				break;
			/* Scan for first char */
			if (sdp->reverse)
			{
				while ((loc_ptr > mem_buf) && (*loc_ptr != c))
					loc_ptr--;
			}
			else
			{
				while (*loc_ptr && *loc_ptr != c)
					loc_ptr++;
			}
			if ((loc_ptr < mem_buf) || !*loc_ptr)
			{	/* Not found */
				ret = 1;
				break;
			}

			in = str;
			out = loc_ptr;
			while (*in)
			{
				if (*out != *in)
					break;
				out++;
				in++;
			}
			if (!*in)
			{	/* Found it */
				cur_ptr = loc_ptr;
				ret = 0;
				break;
			}
			if (sdp->reverse)
				loc_ptr--;
			else
				loc_ptr++;
		}
	}
	return ret;
}

/* Scan string, replacing NLFUDGE chars with \n */
fixnl(str)
unsigned char *str;
{
	while (*str)
	{
		if (*str == NLFUDGE)
			*str = '\n';
		str++;
	}
#ifdef notdef
	register unsigned char *copy, *in, *out;

	copy = Malloc(strlen(str) + 1);
	strcpy(copy, str);
	in = copy;
	out = str;
	do
	{
		if (*in == NLFUDGE)
		{
			*out++ = '\n';
		}
		else
			*out++ = *in;
	}
	while (*in++);
	Free(copy);
#endif
}

#ifdef notdef
/* Strip \r's from a string */
rmcr(cpin)
register unsigned char *cpin;
{
	register unsigned char *cpout;

	cpout = cpin;
	while (*cpin)
	{
		if (MASK(*cpin) != '\r')
			*cpout++ = *cpin++;
		cpin++;
	}
	*cpout = '\0';
}
#endif
