/*
 * Some more functions
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>

#ifdef NAV
#	include <sys/fs/ufs_fsdir.h>
	typedef struct direct DIR;
#else
#	include <sys/dir.h>
#endif

#include <sys/stat.h>
#include "../sedit.h"
#include <string.h>
#include <sys/fcntl.h>

#ifdef __hpux
#define dirent direct
#endif

extern char *roomfor();

extern unsigned char *mem_buf;
extern unsigned char *end_buf;
extern unsigned char *cur_lp;	/* Start of current line */
extern unsigned char *blkstart;		/* Start of block */
extern unsigned char *blkend;		/* End of block */
extern int blkset;
extern unsigned char *cur_ptr;
extern unsigned char **markers;		/* Markers */
extern unsigned char *hilitestart, *hiliteend;
extern int tabwidth;
extern unsigned char *file_name;
int beingEdited;
int zapBeingEdited;

static char stateFileName[] = ".ddestate";

char *
convertToStateFileName(name)
char *name;
{
	char *home = getenv("HOME");
	char *ret;
	if (home == 0)
		abort();
	ret = Malloc(strlen(home) + 1 + strlen(stateFileName) + 1);
	sprintf(ret, "%s/%s", home, stateFileName);
	return ret;
}

char *basePart(name)
char *name;
{
	char *ret;

	ret = name;
	while (strchr(ret, '/'))
		ret = strchr(ret, '/') + 1;
	return ret;
}

/*
 * Locate this file's entry in its state file, return the fd with the
 * file pointer at the right place.  Return the file's fd.
 * If no entry was found, set *noEntry.  Return -1 on error.
 */

int
findFileState(name, noEntry)
char *name;
int *noEntry;
{
	char *statefile = convertToStateFileName(name);
	int fd = open(statefile, O_RDWR|O_CREAT, 0666);
	struct oldFile of;
	char *bp = basePart(name);

	Free(statefile);
	*noEntry = 0;
	if (fd < 0)
		return fd;
	while (read(fd, &of, sizeof(of)) == sizeof(of))
	{
		if (!strncmp(of.name, bp, sizeof(of.name)))
		{
			lseek(fd, -sizeof(of), SEEK_CUR);
			return fd;
		}
	}
	*noEntry = 1;
	return fd;
}

saveFileState(name)
char *name;
{
	struct oldFile of;
	int i;
	int fd, noEntry;

	fd = findFileState(name, &noEntry);
	if (fd >= 0)
	{
		int pos;

		memset(&of, 0, sizeof(of));
		pos = lseek(fd, 0, SEEK_CUR);
		read(fd, &of, sizeof(of));
		lseek(fd, pos, SEEK_SET);

		strncpy(of.name, basePart(name), sizeof(of.name));
		for (i = 0; i < 10; i++)
			of.markers[i] = markers[i] ? (markers[i] - mem_buf) : 0;
		of.blkstart = blkstart ? (blkstart - mem_buf) : 0;
		of.blkend = blkend ? (blkend - mem_buf) : 0;
		of.blkset = blkset;
		of.tabwidth = tabwidth;
		of.cur_ptr = cur_ptr - mem_buf;
		of.hilitestart = hilitestart ? (hilitestart - mem_buf) : 0;
		of.hiliteend = hiliteend ? (hiliteend - mem_buf) : 0;
		of.beingEdited--;
		time(&(of.useTime));
		if (of.beingEdited < 0)
			of.beingEdited = 0;
		if (zapBeingEdited)
			of.beingEdited = 1;
		write(fd, &of, sizeof(of));
		close(fd);
	}

	/* Add new entry to the switch menu */
	initswitchmenu();
}

/*
 * Look for an old file, return offset, or 0 if not there.
 */

int
findOldFile(name)
char *name;
{
	int fd, noEntry;
	struct oldFile of;

	memset(&of, 0, sizeof(of));
	fd = findFileState(name, &noEntry);
	if (fd >= 0)
	{
		if (noEntry == 0)
		{
			read(fd, &of, sizeof(of));
			close(fd);
			return of.cur_ptr;
		}
		else
		{
			close(fd);
			return 0;
		}
	}
	return 0;		
}

unsigned char *
qualify(offset)
long offset;
{
	unsigned char *ret;

	ret = mem_buf + offset;
	if (ret < mem_buf || ret >= end_buf)
		ret = (unsigned char *)0;	/* No pointer */
	return ret;
}

/*
 * Set up the pointers
 */

setpointers(name)
char *name;
{
	struct oldFile of;
	int i, fd, noEntry;
	int gotit = 0;

	memset(&of, 0, sizeof(of));
	fd = findFileState(name, &noEntry);
	if (fd >= 0)
	{
		if (!noEntry)
		{
			if (read(fd, &of, sizeof(of)) == sizeof(of))
			{
				for (i = 0; i < 10; i++)
					markers[i] = qualify(of.markers[i]);
				blkstart = qualify(of.blkstart);
				blkend = qualify(of.blkend);
				tabwidth = of.tabwidth;
				blkset = of.blkset;
				cur_ptr = qualify(of.cur_ptr);
				hilitestart = qualify(of.hilitestart);
				hiliteend = qualify(of.hiliteend);
				beingEdited = of.beingEdited;
				of.beingEdited++;
				lseek(fd, -sizeof(of), SEEK_CUR);
				write(fd, &of, sizeof(of));
				gotit = 1;
			}
		}
		else
		{	/* File has no entry */
			strncpy(of.name, basePart(name), sizeof(of.name));
			of.beingEdited = 1;
			write(fd, &of, sizeof(of));
		}
		close(fd);
	}
	if (!gotit)
	{
		for (i = 0; i < 10; i++)
			markers[i] = 0;
		blkstart = 0;
		blkend = 0;
		blkset = 0;
		cur_ptr = mem_buf;
		hilitestart = 0;
		hiliteend = 0;
	}
	return;
}

/*
 * Scan the current directory, return array of pointers to
 * all regular file's names
 */

char **
scandot(pnentries)
int *pnentries;
{
	DIR *dir;
	struct dirent *direct;
	int pass, nentries;
	char **names;
	char buf[MAXNAMLEN+5];

	dir = opendir(".");
	if (dir == (DIR *)0)
		return (char **)-1;
	for (pass = 0; pass < 2; pass++)
	{
		if (pass)
			rewinddir(dir);
		nentries = 0;
		while (direct = readdir(dir))
		{
#ifdef __hpux
			strncpy(buf, direct->d_name, direct->d_namlen);
			buf[direct->d_namlen] = '\0';
#else
			strcpy(buf, direct->d_name);
#endif
			if (pass == 0)
				nentries++;
			else
			{
				if (!boring(buf))
					names[nentries++] = roomfor(buf);
			}
		}
		if (pass == 0)
		{
			names = Malloc(nentries * sizeof(*names));
		}
	}
	closedir(dir);
	*pnentries = nentries;
	return names;
}

/* Return true if passed filename ends in one of these */

boring(name)
char *name;
{
	static char *missem[] =
		{ ".bak", ".obj", ".o", ".out", ".xrel", ".temp", ".exec" };
	register int i, sln;
	struct stat statbuf;

	sln = strlen(name);
	for (i = 0; i < sizeof(missem)/sizeof(*missem); i++)
	{
		if (strlen(missem[i]) >= sln)
			continue;
		if (!strucmp(missem[i], name + sln - strlen(missem[i])))
			return 1;		/* Match */
	}
	if (stat(name, &statbuf) < 0)
		return 1;
	if ((statbuf.st_mode & S_IFMT) != S_IFREG)
		return 1;
	return 0;
}

/* Case-insensitive string comparison: return zero if equal */
strucmp(s1, s2)
register char *s1, *s2;
{
	while (*s1)
	{
		if (toupper(*s1) != toupper(*s2))
			return 1;
		s1++;
		s2++;
	}
	return *s2;
}

/*
 * ^QP support functions
 */

#define NQPS	20

char *qpstack[NQPS];
int qpsp;		/* Index into stack */

qpstackinit()
{
	qpsp = 0;
}

addqpstack(p)
char *p;
{
	qpstack[qpsp++] = p;
	if (qpsp == NQPS)
	{
		/* Exemplary code, what? */
		movmem(&qpstack[1], &qpstack[0],
				sizeof(qpstack) - sizeof(*qpstack));
		qpsp--;
	}
}

/* Get top from stack */
char *
popqpsp()
{
	char *ret;
	if (qpsp)
		ret = qpstack[--qpsp];
	else
		ret = (char *)0;
	return ret;
}

/* Do ^QP relocations */
doqpreloc(source, dest, distance)
register char *source, *dest;
{
	register int i;
	register char *ptr;

	for (i = 0; i < qpsp; i++)
	{
		if (ptr = qpstack[i])
		{
			if (ptr >= source)
				ptr += distance;	/* Within moved block */
			else
			{
				if ((distance < 0) && (ptr >= dest))
					ptr = dest;	/* Marker in deleted block */
			}
			qpstack[i] = ptr;	/* New value */
		}
	}
}

/*
 * Find matching parenthesis
 */

unsigned char *
_parenmatch(char *incstr, char *decstr, int direction)
{
	register unsigned char *ptr;
	unsigned char ch;
	int count;
	extern unsigned char *end_buf, *mem_buf;
	int inclen, declen;

	ptr = cur_ptr;
	ch = *ptr;

	if (incstr == 0 || decstr == 0)
	{
		switch (ch)
		{
		case '(':
			incstr = "(";
			decstr = ")";
			direction = FORWARD;
			break;
		case '[':
			incstr = "[";
			decstr = "]";
			direction = FORWARD;
			break;
		case '{':
			incstr = "{";
			decstr = "}";
			direction = FORWARD;
			break;
		case ')':
			incstr = ")";
			decstr = "(";
			direction = BACKWARD;
			break;
		case ']':
			incstr = "]";
			decstr = "[";
			direction = BACKWARD;
			break;
		case '}':
			incstr = "}";
			decstr = "{";
			direction = BACKWARD;
			break;
		case '#':
			if (!strncmp((char *)cur_ptr, "#if", 3))
			{
				incstr = "#if";
				decstr = "#endif";
				direction = FORWARD;
			}
			else if (!strncmp((char *)cur_ptr, "#endif", 6))
			{
				incstr = "#endif";
				decstr = "#if";
				direction = BACKWARD;
			}
			else
			{
				return ptr;
			}
			break;
		default:
			return ptr;
		}
	}

	count = 1;
	ptr += direction;

	if (ptr <= mem_buf || ptr >= end_buf)
		return cur_ptr;

	inclen = strlen(incstr);
	declen = strlen(decstr);

	while (count && (ptr > mem_buf) && (ptr < end_buf))
	{
		if (!strncmp((char *)ptr, incstr, inclen))
			count++;
		else if (!strncmp((char *)ptr, decstr, declen))
			count--;
		ptr += direction;
	}
	if (count == 0)
		return ptr - direction;
	else
	{
		sstatmes("\007Off by %d", count);
		return cur_ptr;
	}
}

void
parenmatch(char *incstr, char *decstr, int direction)
{
	unsigned char *tempptr = _parenmatch(incstr, decstr, direction);
	if (tempptr != cur_ptr)
	{
		cur_ptr = tempptr;
		redoptr();
		newcurs();
		set_param(cur_lp);
	}
}

/*
 * Passed a string, see if it has wild cards, and if so,
 * scan current directory for a match, replacing original
 */

exwild(str)
char *str;
{
}

#ifndef Linux
toupper(c)
{
	if (c >= 'a' && c <= 'z')
		return c + 'A' - 'a';
	return c;
}
#endif

void movstr(void *from, void *to, int len)
{
	char *_from = (char *)from;
	char *_to = (char *)to;

	while (*_from && len--)
		*_to++ = *_from++;
	*_to = '\0';
}

/*
 * Return the filename part of the currently edited file
 */
char *currentFileName(void)
{
	char *ret;
	char *p;

	p = strrchr((char *)file_name, '/');
	if (p)
		ret = p + 1;
	else
		ret = (char *)file_name;
	return ret;
}

/*
 * Return all of the current pathname
 */

char *currentPathName(void)
{
	return (char *)file_name;
}

char *nodotslash(char *_p)
{
	while (_p[0] == '.' && _p[1] == '/')
		strcpy(_p, _p + 2);
	return _p;
}

/*
 * Search upwards from the passed directory, looking for
 * the passed file
 */

char *searchup(char *_dirName, char *_fileName)
{
	char *ret = 0;
	struct stat slashstat;

	if (stat("/", &slashstat) == 0)
	{
		struct stat thisdir;
		int gotit = 0;
		int hitslash = 0;
		int problem = 0;
		_dirName = roomfor(_dirName);
		do
		{
			ret = Malloc(strlen(_dirName) + strlen(_fileName) + 4);
			sprintf(ret, "%s/%s", _dirName, _fileName);
			if (access(ret, R_OK) == 0)
			{
				gotit = 1;
			}
			else
			{
				if (stat(_dirName, &thisdir) != 0)
				{
					problem = 1;
					Free(_dirName);
					_dirName = 0;
					Free(ret);
					ret = 0;
				}
				else
				{
					if (thisdir.st_ino == slashstat.st_ino &&
						thisdir.st_dev == slashstat.st_dev)
					{
						hitslash = 1;
						Free(_dirName);
						_dirName = 0;
						Free(ret);
						ret = 0;
					}
					else
					{
						char *newdirname = Malloc(strlen(_dirName) + 4);
						sprintf(newdirname, "%s/..", _dirName);
						Free(_dirName);
						_dirName = newdirname;
					}
				}
			}
		}
		while (gotit == 0 && hitslash == 0 && problem == 0);
	}
	return ret;
}
