/* TABS4
 * Stuff for handling tags
 */

#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>

#include <sys/stat.h>
#include <sys/fcntl.h>
#include "../sedit.h"

extern unsigned char *mem_buf, *end_buf;
extern unsigned char *cur_ptr;
extern unsigned char *file_name;		/* Current filename */
extern int modflag;
extern unsigned char *stringtolookfor;
extern int look_for_line;
extern int line_to_look_for;

typedef struct tagFile
{
	struct tagFile *next;
	char *name;					/* The name of the tags file */
	char *dir;					/* The name of its directory */
	int offset;					/* The offset of it's contents in allTags */
	ino_t st_ino;
	dev_t st_dev;
} tagFile;

tagFile *tagFiles = 0;

gototag(tagname)
char *tagname;
{
	register unsigned char *cp;
	unsigned char *name, *searchstr;
	char *tagnamecopy = 0;
	int freeit = 0;
	int is_line_number;

	if (tagname != 0 && tagname != (char *)1)
	{
		tagnamecopy = tagname;
		freeit = 0;
	}

	if (tagname == 0)
	{	/* Extract from under cursor */
		unsigned char *startpt = cur_ptr;

		cp = cur_ptr;
		if (!isidch(*cp))
			return;

		/* new Mod */
		while (startpt > mem_buf && isidch(startpt[-1]))
			startpt--;

		while (isidch(*cp))
			cp++;
		tagnamecopy = Malloc(cp - startpt + 2);
		freeit = 1;
		movmem(startpt, tagnamecopy, cp - startpt);
		tagnamecopy[cp - startpt] = '\0';
	}
	else if (tagname == (char *)1)
	{
		tagnamecopy = tagname;
	}

	if ((long)lookuptag(tagnamecopy, &name, &searchstr, &is_line_number) <= 0)
	{
		sstatmes("\007No such tag");
		return;
	}
	/* Tag OK. Is it in this file? */
	if (pathcmp(name, file_name) != 0)
	{	/* Different file */
		if (modflag && writecurrentfile(0, mem_buf, end_buf))
			return;
		modflag = 0;
		donewfile(name);
		if (is_line_number)
		{
			stringtolookfor = 0;
			look_for_line = 1;
			line_to_look_for = atoi(searchstr);
		}
		else
		{
			stringtolookfor = searchstr;
			look_for_line = 0;
		}
	}
	else if (is_line_number)
	{
		sendtome(CTL('q'));
		sendtome(CTL('g'));
		stringtome(searchstr);
		sendtome('\r');
	}
	else
	{
		extern int nodrawonsearch;
		struct sr_data sdp;

		strncpy(sdp.old, searchstr, SEDLINESIZE);
		sdp.old[SEDLINESIZE - 1] = '\0';
		sdp.useregexp = 0;
		sdp.reverse = 0;
		sdp.ignorecase = 0;
		sdp.fromtop = 0;
		sdp.wrap = 0;
		sdp.ntimes = 1;
		sdp.initialised = 1;

		cur_ptr = mem_buf;
		nodrawonsearch = 1;
		search(&sdp);
		nodrawonsearch = 0;
		redoptr();
		redwhole();
	}
	Free(name);
	if (freeit)
		Free(tagnamecopy);
}

isidch(c)
char c;
{
	return	(c >= 'a' && c <= 'z') ||
		(c >= 'A' && c <= 'Z') ||
		(c >= '0' && c <= '9') ||
		(c == '_') ||
		(c == '$') ||
		(c == ':') ||
		(c == '~') ||
		(c == '&');
}
	
/*
 * Lookup a passed tag. Return filename, search string
 * return null if not found
 */

static char deftagfilename[] = "tags";
static int nnames = 0;
static char **tagfilenames = 0;
static char gotpaths = 0;
static char *allTags = 0;
static int allTagsSize = 0;

char *lookuptag(char *str, unsigned char **pname, unsigned char **psearch,
			int *pis_number)
{
	static char *lasttag = 0;
	static char *lasticp = 0;
	char buf2[1000];
	int ifd, ret;
	register char *icp, *ocp;
	char *name, *search;
	char *tagfilename;
	int nameno;
	int finalerr = 0;
	int is_line_number = 0;

	/* Isolate the pathnames */
	if (!gotpaths)
	{
		tagFile *tf;

		if ((tagfilename = getenv("TAGS")) == 0)
			tagfilename = deftagfilename;
		/* User has a tag file environment setting */
		nnames = sliceargs(tagfilename, &tagfilenames, 0, 1);
		gotpaths = 1;

		for (nameno = 0; nameno < nnames; nameno++)
			addTagFile(tagfilenames[nameno]);

		for (tf = tagFiles; tf != 0; tf = tf->next)
		{
			struct stat statbuf;
			char *whereToRead;

			tagfilename = tf->name;
			tf->offset = allTagsSize;

			if (stat(tagfilename, &statbuf) < 0)
				continue;

			if (allTags == 0)
			{
				allTagsSize = statbuf.st_size;
				allTags = Malloc(allTagsSize + 1);
				whereToRead = allTags;
			}
			else
			{
				allTags = Realloc(	allTags,
									allTagsSize + statbuf.st_size + 1);
				whereToRead = allTags + allTagsSize;
				allTagsSize += statbuf.st_size;
			}

			if ((ifd = open(tagfilename, O_RDONLY)) < 0)
			{
				finalerr = ifd;
				continue;
			}

			if ((ret = read(ifd, whereToRead, statbuf.st_size)) !=
							statbuf.st_size)
			{
				finalerr = ret;
				continue;
			}

			close(ifd);
			whereToRead[statbuf.st_size] = '\0';
		}
	}

	name = (char *)0;

	if (str == (char *)1)
	{
		if (lasttag == 0)
		{
			return (char *)0;
		}
		str = lasttag;
		icp = lasticp;
	}
	else
	{
		if (lasttag)
			Free(lasttag);
		lasttag = roomfor(str);
		icp = allTags;
	}

	while (!name && icp && *icp)
	{
		register char *xp, sxp;

		xp = icp;
		while (*xp && (*xp != '\r') && (*xp != '\n'))
			xp++;
		sxp = *xp;
		ocp = buf2;
		while (	*icp &&
				(icp < xp) &&
				(*icp != ' ') &&
				(*icp != '\t') &&
				((ocp - buf2) < sizeof(buf2)))
		{
			*ocp++ = *icp++;
		}
		if (ocp - buf2 >= sizeof(buf2))
			continue;
		if (!*icp)
			continue;
		*ocp = '\0';				/* Isolate first field */
		if (!strcmp(buf2, str))
		{	/* Match */
			while (*icp && (icp < xp) && ((*icp == ' ') || (*icp == '\t')))
				icp++;				/* Go to file name */
			if (!*icp)
				continue;
			/* Extract filename */
			name = Malloc(xp - icp + 1);
			ocp = name;
			while (*icp && (icp < xp) && (*icp != ' ') && (*icp != '\t'))
			{
				if (*icp == '\\')
					icp++;
				*ocp++ = *icp++;
			}
			*ocp = '\0';
			if (!*icp)
				continue;
			/* Extract search string */
			while (*icp && (icp < xp) && ((*icp == ' ') || (*icp == '\t')))
				icp++;
			if (!*icp)
				continue;

			search = Malloc(strlen(icp) + 2);
			ocp = search;

			if (*icp == '/' && icp[1] == '^')
			{
				icp += 2;

				/* Start the search pattern */
				*ocp++ = CTL('N');
			}
			else if (isdigit(*icp))
			{	/* Line number */
				is_line_number = 1;
			}

			while (*icp && (icp < xp) && (*icp != '\r') && (*icp != '\n'))
			{
				if (*icp == '\\')	/* Skip escapes in the search string */
					icp++;
				else
					*ocp++ = *icp++;
			}
			*ocp = '\0';
			if (ocp[-1] == '/')
				ocp--;
			if (ocp[-1] == '$')
				ocp--;
			*ocp = '\0';

			/* Locate the tagFile entry for this string */
			{
				int offset = xp - allTags;
				tagFile *tf = tagFiles;
				while (tf)
				{
					if ((tf->next == 0) ||(tf->next->offset > offset))
						break;
					tf = tf->next;
				}
				if (tf)
				{
					char *newname = Malloc(strlen(tf->dir) + strlen(name) + 4);
					sprintf(newname, "%s/%s", tf->dir, name);
					Free(name);
					name = newname;
				}
			}
			*pname = name;
			*psearch = search;
			*pis_number = is_line_number;

			if (0)
			{
				fprintf(stderr,
					"pname='%s', psearch='%s', pnumber=%d\n",
					*pname,
					*psearch,
					*pis_number);
			}

			lasticp = icp;
			return name;
		}
		icp = xp;
		while (*icp == '\n' || *icp == '\r')
			icp++;
	}

	if (finalerr < 0)
		statsyserr(finalerr);

	return (char *)0;
}

void cleanUpTags(void)
{
	if (tagfilenames)
	{
		cleanUpSlicedArgs(tagfilenames, nnames);
	}
	if (allTags)
	{
		Free(allTags);
	}

	{
		tagFile *tf = tagFiles;
		while (tf)
		{
			tagFile *tf2 = tf->next;
			Free(tf->name);
			Free(tf->dir);
			Free(tf);
			tf = tf2;
		}
	}
}

void addTagFile(const char *_path)
{
	struct stat statbuf;
	if (stat(_path, &statbuf) == 0)
	{
		tagFile *tf = tagFiles;
		while (tf)
		{
			if (	tf->st_ino == statbuf.st_ino &&
					tf->st_dev == statbuf.st_dev)
			{
				break;
			}
			tf = tf->next;
		}
		if (tf == 0)
		{
			tf = (tagFile *)Malloc(sizeof(*tf));
			tf->name = nodotslash(roomfor(_path));
			tf->st_ino = statbuf.st_ino;
			tf->st_dev = statbuf.st_dev;
	
			tf->dir = nodotslash(roomfor(_path));
			if (strrchr(tf->dir, '/'))
			{
				*(strrchr(tf->dir, '/')) = '\0';
			}
			else
			{
				strcpy(tf->dir, ".");
			}

			tf->next = 0;
			if (tagFiles == 0)
			{
				tagFiles = tf;
			}
			else
			{
				tagFile *tf2 = tagFiles;
				while (tf2->next)
					tf2 = tf2->next;
				tf2->next = tf;
			}
		}
	}
}

