/*	TABS4
 * terminal dependent stuff
 */

#include <stdio.h>
#include <sys/ioctl.h>
#include "../sedit.h"

#define XHIL

/* The height and width of the screen in chars. */
extern int height, width;

/* 'curscrtop' is 0 if status line disabled, else 1 */
extern int curscrtop;

/* 'dispcontrol' toggles the display of control chars (^OD) */
extern int dispcontrol;

/* 'overstrike' is true if overtyping chars (^\ toggles) */
extern int overstrike;

/* 'attrnow' describes the current output attribute setting. 'attrwant' is
 * the new, desired setting
 */
extern int attrnow, attrwant;

/* Left and right margin settings, overridden by command line switches */
extern int rmargin, lmargin;

/* 'justifymode' can be one of J_RIGHT, J_RAGGED and J_CENTRE */
extern int justifymode;

extern char *getenv();

extern char *tgetstr();

int ismgr = 0;
extern int inleditor;

extern int is_xterm;

/*
 * 'cur_x' and 'cur_y' are the current cursor position. May be useful
 * for fudging things like clear to end of line if the terminal can't do it.
 * Note that cur_x can be greater than 'width', so the real Y cursor position
 * is at (cur_y + cur_x / width) and the real X position is at
 * (cur_x % width).
 */
extern int cur_x, cur_y;

char *BC;		/* Backspace */
char *UP;		/* Cursor up */
char *DL;		/* Delete line */
char *CE;		/* Clear to EOL */
char *CL;		/* Clear screen */
char *CD;		/* Clear to end of screen */
char *AL;		/* Insert line */
char *SO;		/* Standout */
char *SE;		/* Standend */
char *CM;		/* Cursor motion */
char *SR;		/* Scroll down */
char *SF;		/* Scroll up */
char *CS;		/* Scroll region set */
char *MD;       /* Bold */
char *ME;       /* Attributes off */
int cur_hil;

char usealdl;		/* If true, use AL & DL */

/*
 * Initialise the terminal for editing. Just the very first thing done.
 */

char *
Tgetstr(id, area)
char *id, **area;
{
	char *ret;
	ret = tgetstr(id, area);
	if (ret == (char *)0)
	{
		fprintf(stderr, "Cannot extract '%s' capability from termcap\n", id);
		doexit(1);
	}
	return ret;
}

int term_scrolls_at_last_col()
{
	if (is_xterm)
		return 0;
	return 1;
}

terminit()
{
	char *termname;
	char tbuf[1024];
	static char obuf[1024];
	char *area;
	extern int errno;

	termname = (char *)getenv("TERM");
	if (termname == (char *)0)
	{
		fprintf(stderr, "Environment variable $TERM not set\n");
		doexit(1);
	}

	if (!strncmp(termname, "xterm", 5) ||
			!strncmp(termname, "rxvt", 4))
	{
		is_xterm = 1;
		xterm_pointer_mode(1);
	}

	if (tgetent(tbuf, termname) != 1)
	{
		fprintf(stderr, "Cannot get terminal entry for your terminal '%s'\n",
				termname);
		doexit(1);
	}
	area = obuf;

	BC = tgetstr("bc", &area);
	UP = tgetstr("up", &area);		/* tgoto() uses these */
	CE = Tgetstr("ce", &area);
	CL = Tgetstr("cl", &area);
	CD = tgetstr("cd", &area);		/* optional */
	SO = Tgetstr("so", &area);
	SE = Tgetstr("se", &area);
	CM = Tgetstr("cm", &area);
#ifdef XHIL /* CPW */
    MD = Tgetstr("md", &area);
    ME = Tgetstr("me", &area);
    if (!(MD && *MD && ME && *ME))
    {
	        MD = SO;
	        ME = SE;
	}
    cur_hil = 0;
#endif
	AL = tgetstr("al", &area);
	DL = tgetstr("dl", &area);

	/* Set up 'width' and 'height' */
	width = tgetnum("co");
	height = tgetnum("li");
	getwinsize(&height, &width);

	if (AL == 0)
	{
		AL = Malloc(100);
		sprintf(AL, "\0337\033[%%d;%dr\0338\033M\033[;r\0338", height);
	}
	if (DL == 0)
	{
		DL = Malloc(100);
		sprintf(DL, "\0337\033[%%d;%dr\033[%d;0H\n\033[;r\0338", height, height);
	}

	if (!strcmp(termname, "xterm"))
	{
		int i;
		for (i = 0; i < height - 1; i++)
			putchar('\n');
	}

	if (AL && DL && *AL && *DL)
		usealdl = 1;
	else
	{
		SR = Tgetstr("sr", &area);
		SF = Tgetstr("sf", &area);
		CS = Tgetstr("cs", &area);
	}

	if (width <= 0 || height <= 0)
	{
		fprintf(stderr, "Cannot determine screen dimensions from termcap\n");
		doexit(1);
	}

	/* Set up the default margin settings */
#if 0
	rmargin = width - 5;
	lmargin = 0;
#endif

	/* Select the default justification mode */
	justifymode = J_RIGHT;

	/* Display control chars as ^X */
	dispcontrol = 1;

	/* Start up in insert mode */
	overstrike = 0;

	/* Assume every attribute is on, so setbul() will turn them all off */
	attrwant = 0;
	attrnow = BOLD|UNDERLINE|SUBSCRIPT|SUPERSCRIPT|ITALICS|WIDE;
	setbul();
}

/* Terminate the editing session. Called just before quitting */
termterm()
{}

/* tputs output function */
outc(c)
{
	opchar(c);
}

/*
 * Delete a line
 */
dellines(linenum, nlines)
register int linenum, nlines;
{
	if (linenum < height)
	{
		if (DL && *DL)
		{
			while (nlines--)
			{
				char buf[100];

				gotoxy(0, linenum);
				sprintf(buf, DL, nlines);
/*				sprintf(buf, DL, linenum + 1);*/
				SCREENPUTS(buf);
			}
		}
		else
		{	/* vt100 */
			char *cp;
			cp = (char *)tgoto(CS, linenum, height - 1);
			SCREENPUTS(cp);
			/* Now scroll up */
			tputs(SF, 1, outc);
			/* Restore scroll region */
			cp = (char *)tgoto(CS, 0, height - 1);
			SCREENPUTS(cp);
		}
	}
}

/*
 * Move the cursor to the appropriate position.
 */

gotoxy(x, y)
register int x, y;
{
	char *cp;

	if (x >= width)
	{
		y += x / width;
		x = x % width;
	}
	cp = (char *)tgoto(CM, x, y);
	SCREENPUTS(cp);
}

void clear_eol(void)
{
	SCREENPUTS(CE);
}

clr_scrn()
{
	gotoxy(0, 0);
	SCREENPUTS(CL);
	gotoxy(0, 0);
}

void clear_eos(int x, int y)
{
	int y1;

	if (CD && *CD)
	{
		SCREENWRITE(CD);
	}
	else
	{
		y1 = y;
		gotoxy(x, y1);
		clear_eol();
		while (++y1 < height)
		{
			gotoxy(0, y1);
			clear_eol();
		}
		gotoxy(x, y);
	}
}

/*
 * Scroll the whole screen down a row
 */

dn_scrl()
{
	extern int scr_top;

	if (AL && *AL)
	{
		char buf[100];

		gotoxy(0, scr_top);
		sprintf(buf, AL, scr_top + 1);
		SCREENPUTS(buf);
	}
	else
	{	/* vt100 */
		 tputs(SR, 1, outc);
	}
}

inslin(row)
{
	char *cp;

	if (AL && *AL)
	{
		char buf[100];

		sprintf(buf, AL, row + 1);
		SCREENPUTS(buf);
	}
	else
	{	/* vt100 */
		cp = (char *)tgoto(CS, row, height - 1);
		SCREENPUTS(cp);
		/* Now scroll down */
		dn_scrl();
		/* Restore scroll region */
		cp = (char *)tgoto(CS, 0, height - 1);
		SCREENPUTS(cp);
	}
}

prcrlf()
{
	opchar('\n');
}

void hil(val)
{
    if (SO && *SO && SE && *SE && ME && *ME && MD && *MD)
    {
        if (val != cur_hil)
        {
            if (cur_hil == 1)
                SCREENPUTS(ME);
            else if (cur_hil == 2)
                SCREENPUTS(SE);
            if (val == 1)
                SCREENPUTS(MD);
            else if (val == 2)
                SCREENPUTS(SO);
            cur_hil = val;
        }
	}
	else if (SO && *SO && SE && *SE)
	{
        if (val)
            SCREENPUTS(SO);
        else
            SCREENPUTS(SE);
    }
}

#if 0
hil(val)
{
	if (SO && *SO && SE && *SE)
	{
		if (val)
			SCREENPUTS(SO);
		else
			SCREENPUTS(SE);
	}
}
#endif

setbul()
{
	if (dispcontrol && (attrnow))
	{
		hil(0);
		attrnow = 0;
	}
	else
	{
		hil(1);
		attrnow = attrwant;
	}
}

rawgetchar()
{
	flushop();
	return _rawgetchar();
}

beep()
{
	opchar(7);
}

rawputchar(c)
{
	opchar(c);
}

SCREENPUTS(s)
char *s;
{
	SCREENWRITE(s, strlen(s));
}

Printf(str, p1, p2, p3, p4, p5, p6, p7, p8, p9)
char *str;
{
	char buf[1000];

	sprintf(buf, str, p1, p2, p3, p4, p5, p6, p7, p8, p9);
	SCREENWRITE(buf, strlen(buf));
}

/* Stubs for MGR version of the code */
dooz()
{}

eschit()
{
/*	terminit();*/
	getwinsize(&height, &width);
	redwhole();
}

savestate()
{}

reststate()
{}

redrawmenubar()
{}

initswitchmenu()
{}


char obuf[16384];
int obi = 0;

SCREENWRITE(addr, n)
char *addr;
{
	while (n--)
	{
		obuf[obi++] = *addr++;
		if (obi == NELEM(obuf))
			flushop();
	}
}

void opchar(char c)
{
	SCREENWRITE(&c, 1);
}

flushop()
{
	if (obi)
	{
		/*
		fwrite(obuf, 1, obi, stdout);
		*/
		write(fileno(stdout), obuf, obi);
		fflush(stdout);
		obi = 0;
	}
}

void xterm_pointer_mode(yes)
int yes;
{
	Printf("\033]%d;\007", yes ? 8 : 9);
}
