/*	TABS8
 * The MGR part of DrDoc
 */

#include <stdio.h>
#include "mgrstate.h"
#include <term.h>		/* NODEP */
#include <signal.h>
#include "../sedit.h"

#ifdef applix1616
#include <syscalls.h>
#endif

#define MAXMENUSIZE	20
#define dprintf if(debug)fprintf
#define NENTRIES(array)		(sizeof(array)/sizeof(*array))

/* A rectangular screen region */
struct object
{
	int x_min, y_min, x_max, y_max;
	int id;		/* Unique identifier */
	int flags;
	int menuno;	/* Menu number (if it is a menu) */
	int which;	/* What object it is */
};
#define OBJECT	struct object		/* Hitech typedef yuk */

/* Bits in object.flags above */
#define OBJ_ACTIVE	1		/* Currently activated */
#define OBJ_PDMENU	2		/* It is a menu */

/* Settings for object.which */
#define WH_PDMENU	1		/* The pulldown menu */
#define WH_UPARROW	2		/* The scroll up arrow */
#define WH_DOWNARROW	3		/* The scroll down arrow */
#define WH_SCROLLBAR	4		/* The scrollbar */
#define WH_SCROLLBOX	5		/* The little box */

/* Size of scrollup/down arrows */
#define ARROWWIDE	24
#define ARROWHIGH	16

static char uparrowname[] = "scrollup";
static char downarrowname[] = "scrolldown";
static char boxname[] = "scrollbox";
static char iconname[] = "editor";

OBJECT *objects[10];
OBJECT *scrollboxobj;			/* Object descriptor for scroll box */
OBJECT *scrollbarobj;			/* Object descriptor for scroll bar */
OBJECT *menuobject = 0;			/* Descriptor for current menu title line */

extern unsigned char *mem_buf;	/* Start of text */
extern unsigned char *end_buf;	/* Points to null after last line */
extern unsigned char *cur_ptr;	/* Current working pointer */
extern unsigned char *blkstart, *blkend, *file_name;
extern char *roomfor();
extern int cur_lin;	/* Current */
extern unsigned char *cur_lp;	/* Start of current line */
extern int cur_y;	/* Current Y cursor pos */
extern int modflag;	/* Flag: file modified */

int ipstate;				/* For rawgetchar() */
int firsttime;				/* For scroll box */

static int xmax, ymax;
static int font_high, font_wide;
static int win_x, win_y, win_wide, win_high;
static int border;
static int debug;
static char menu_selected;
static char menubar[500];
static char input[200];

OBJECT *mkobject();
OBJECT *objfind();
unsigned char *scanxy();

struct pulldown
{
	char *title;
	OBJECT *obj;		/* Pointer to its object struct */
	int nentries;
	struct menu_entry menu[MAXMENUSIZE];
} pulldowns[] =
{
	{
		"File",
		(OBJECT *)0,
		7,
		{
			{"Write & exit         ^KX","\013X\r"},
			{"Write & resume       ^KD","\013D\r"},
			{"Merge another file   ^KR", "\013R"},
			{"Switch files         ^KA", "\013A"},
			{"Toggle status line   ^KS", "\013S"},
			{"Shell escape         ^KI", "\013I"},
			{"Quit                 ^Kq","\013q"}
		}
	},
#ifdef notdef
	{
		"Positioning",
		(OBJECT *)0,
		4,
		{
			{"Start of file    ^QR","\021R"},
			{"Page up          ^R","\022"},
			{"Page down        ^C","\003"},
			{"End of file      ^QC","\021C"}
		}
	},
#endif
	{
		"Blocks",
		(OBJECT *)0,
		7,
		{
			{"Move block       ^KV", "\013V"},
			{"Copy block       ^KC", "\013C"},
			{"Delete block     ^KY", "\013Y"},
			{"Write block      ^KW", "\013W"},
			{"Reformat block   ^KO", "\013O"},
			{"Hide block       ^KH", "\013H"},
			{"Yank block       ^KP", "\013P"}
		}
	},
	{
		"Formatting",
		(OBJECT *)0,
		7,
		{
			{"Reformat paragraph    ^OO", "\0139\004\017E\017O\0219"},
			{"Centre current line   ^OC", "\017C"},
			{"Set left margin       ^OL", "\017L"},
			{"Set right margin      ^OR", "\017R"},
			{"Right justify", "\036JR\r"},
			{"Centre justify", "\036JC\r"},
			{"Ragged right", "\036Jr\r"}
		}
	},
	{
		"Help",
		(OBJECT *)0,
		4,
		{
			{"General help              ^QH", "\021H"},
			{"Cursor commands (^Q)      ^Q?", "\021?"},
			{"File commands (^K)        ^K?", "\013?"},
			{"On-screen commands (^O)   ^O?", "\017?"}
		}
	},
	{
		"Special",
		(OBJECT *)0,
		3,
		{
			{"Close ", "\036SK\r" },
			{"Cut   ", "\036SC\r" },
			{"Paste ", "\036SP\r" }
		}
	},
	{
		"Switch",
		(OBJECT *)0,
		1,
		{
			{ "New file", "\036SN\r" },
		}
	}
};

struct pulldown *switchpd;
int switchmenuentryno;
int filenamemenuno;

clean()
{
	Printf("\nSignalled");
	quit(0);
}

#define BBSIZE	4000		/* stdio output buffer size */

menuinit()
{
	char *bbuf;

	debug = getenv("DEBUG");

	m_setup(0);

	bbuf = Malloc(BBSIZE);
	if (bbuf != (char *)0)
		setvbuf(m_termout, bbuf, _IOFBF, BBSIZE);

	m_push(P_FLAGS|P_MENU|P_EVENT|P_TEXT);
#ifdef notdef
	/* Block input so K/B does not get tangled with mgr responses */
	while (SGETCHAR())
		sendtome(GETCHAR());
#endif
	m_setmode(M_NOINPUT);
	/* Use absolute pixel coords */
	m_setmode(M_ABS);

	signal(SIGINT, clean);
	readicons();
	setupscreen(1);

	/* Request info on middle button press */
	m_setevent(BUTTON_2, "\036[%p]\r");
	/* Grab blocks with button 1 */
	m_setevent(BUTTON_1, "\036<%t>\r");
	m_setevent(RESHAPE, "\036R\r");
	m_setevent(REDRAW, "\036R\r");
	m_clearmode(M_NOINPUT);
}

char *
mym_gets(p)
char *p;
{
	int ii;

	ii = 0;
	for (;;)
	{
		p[ii] = rawgetchar();
		if (p[ii] == '\r' || p[ii] == '\n')
			break;
		ii++;
	}
	p[ii] = '\0';
	return p;
}

/* A \036 has been received: process the event */
mgrevent()
{
	register OBJECT *obj;
	int mx, my;
	char *index();
	int ii;

	mergeinsbuffer();		/* Do any pending insertions */
	mym_gets(input);
#ifdef notdef
	if (ch = index(input, '\r'))
		*ch = '\0';
	if (ch = index(input, '\n'))
		*ch = '\0';
#endif
#ifdef DEBUG
	dprintf(stderr, "Got input line (%s)\r\n", input);
#endif
	switch (input[0])
	{
	case '[':	/* Press middle button */
		sscanf(input + 1, "%d %d]", &mx, &my);
		obj = objfind(mx, my);
		if (obj)
		{
			if (obj->flags & OBJ_PDMENU)
			{
				m_func(B_XOR);
				m_bitwrite(obj->x_min - border, obj->y_min - border,
					obj->x_max - obj->x_min, obj->y_max - obj->y_min);
				menuobject = obj;
				get_size(&win_x, &win_y, 0, 0);
				m_movemouse(win_x + obj->x_min, win_y + obj->y_max);
				m_selectmenu(obj->menuno);
				menu_selected = 1;
			}
			switch (obj->which)
			{
			case WH_UPARROW:
#ifdef notdef
				ipstate = IPS_SENDCW;
				m_setevent(BUTTON_2U, "\036]\r");
#endif
				for (ii = 0; ii < 10; ii++)
					sendtome(CTL('W'));
				break;
			case WH_DOWNARROW:
#ifdef notdef
				ipstate = IPS_SENDCZ;
				m_setevent(BUTTON_2U, "\036]\r");
#endif
				for (ii = 0; ii < 10; ii++)
					sendtome(CTL('Z'));
				break;
			case WH_SCROLLBAR:		/* Hit scrollbar */
				{
					unsigned char *cp;
					int nchars;

					nchars = end_buf - mem_buf;
					nchars *= my - obj->y_min;
					nchars /= obj->y_max - obj->y_min;
					cp = mem_buf + nchars;
					if (cp < mem_buf)
						cp = mem_buf;
					if (cp >= end_buf)
						cp = end_buf - 1;
					cur_ptr = cp;
					redoptr();
					redhere();
				}
				break;
			}
		}
		else if (mx > 0 && mx < win_wide && my > 0 && my < win_high)
		{	/* A hit on the text area */
			unsigned char *cp;

			mx = (mx - border) / font_wide;
			my = (my - border) / font_high - 1;
			cp = scanxy(mx, my);
			cur_ptr = cp;
			if (cur_ptr >= (end_buf - 2))
				cur_ptr = end_buf - 2;
			redoptr();
			newcurs();
		}
		break;
	case ']':	/* Release middle button */
		menusoff();
		ipstate = 0;
		m_clearevent(BUTTON_2U);
		break;
	case '<':	/* Scanned block */
		{
			int sx, sy, ex, ey;
			extern int blkset;

			sscanf(input + 1, "%d %d %d %d",
				&sx, &sy, &ex, &ey);
			ex += sx;
			ey += sy;
			if (blkset)
			{
				blkset = 0;
				redblock();
			}
			blkstart = scanxy(sx, sy);
			blkend = scanxy(ex, ey);
			if (blkstart < blkend)
			{
				blkset = 1;
				redblock();
			}
		}
		break;
	case 'J':	/* Justification mode */
		{
			extern int justifymode;
			switch (input[1])
			{
			case 'R':
				justifymode = J_RIGHT;
				break;
			case 'C':
				justifymode = J_CENTRE;
				break;
			case 'r':
				justifymode = J_RAGGED;
				break;
			}
		}
		break;
	case 'N':	/* Switch to another file! Write out current one first */
		if (modflag && writecurrentfile((char *)0, mem_buf, end_buf))
			return;
		modflag = 0;
		donewfile(input + 1);
		break;
	case 'Q':
		mym_gets(input);	/* Gobble button release */
		quit(0);
	case 'R':	/* Reshape/redraw */
		m_textreset();
		setupscreen(0);
		redrawmenubar();
		editinit(0);
		redwhole();
		break;
	case 'S':		/* Special function */
		switch (input[1])
		{
		case 'C':	/* Cut */
			if (blkok())
			{
				unsigned char savech;
				savech = *blkend;
				*blkend = '\0';		/* Terminator */
				m_snarf(blkstart);	/* Send the text to mgr */
				*blkend = savech;
			}
			break;
		case 'K':	/* Close & iconify */
			closewind();
			break;
		case 'P':	/* Paste */
			m_setmode(M_NOINPUT);
			m_put();
			/* Efficiently merge from stdin */
			read_fd(STDIN, "stdin", 1, 1, 0);
			m_clearmode(M_NOINPUT);
			draw(cur_lp, cur_lin, cur_y);
			break;
		case 'N':	/* New file requested: load menu with files from disk */
			flushop();
			m_clearevent(BUTTON_2);
			uploadfilenames();
			/* Filename menu will appear on middle button press */
			m_selectmenu(filenamemenuno);
			flushop();
			menu_selected = 1;
			{
				char namebuf[100];
				tgets(namebuf, "New file? ");
				menusoff();
				m_setevent(BUTTON_2, "\036[%p]\r");
				if (*namebuf)
				{
					if (modflag && writecurrentfile((char *)0,
							mem_buf, end_buf))
						return;
					modflag = 0;
					donewfile(namebuf);
				}
			}
			break;
		}
	}
}

menucleanup()
{
	clearscrollbar();
	Printf("\n");
	m_pop();
}

clearscrollbar()
{
	m_textreset();
	m_func(B_CLEAR);
	m_bitwrite(0, 0, win_wide, font_high);
	m_bitwrite(win_wide - ARROWWIDE, 0, win_wide, win_high);
}

menusoff()
{
	if (menu_selected)
		m_nomenu();
	menu_selected = 0;
	if (menuobject)
	{
		m_func(B_XOR);
		m_bitwrite(menuobject->x_min - border,
			menuobject->y_min - border,	/* Dunno why border */
			menuobject->x_max - menuobject->x_min,
			menuobject->y_max - menuobject->y_min);
		menuobject = 0;
	}
}

/* Look for an object at (x, y). If found, return pointer to the struct */

OBJECT *
objfind(x, y)
register int x, y;
{
	register int id;
	register OBJECT *obj;

	for (id = 0; id < NENTRIES(objects); id++)
	{
		obj = objects[id];
		if (obj)
		{
			if (x > obj->x_min && x < obj->x_max &&
				y > obj->y_min && y < obj->y_max)
				return obj;
		}
	}
	return 0;
}

OBJECT *
mkobject(x_min, y_min, x_max, y_max, flags, which)
{
	register OBJECT *obj;
	int id;

	for (id = 0; id < NENTRIES(objects); id++)
	{
		if (!objects[id])
			break;
	}
	if (id == NENTRIES(objects))
	{
		fprintf(stderr, "Out of objects\n");
		doexit(-1);
	}
	obj = Malloc(sizeof(*obj));
	obj->x_min = x_min;
	obj->y_min = y_min;
	obj->x_max = x_max;
	obj->y_max = y_max;
	obj->flags = flags;
	obj->which = which;
	obj->id = id;
	objects[id] = obj;
	return obj;
}

/*
 * Close window, display icon
 */

closewind()
{
	static int iconx, icony, editoriconloaded = 0, firsttime = 1;
	static int small_windx = -1, small_windy;
	int x, y, border, temp;
	int windxsize, windysize, iconxpos;
	char line[80];

	if (editoriconloaded || (m_bitfile(4, iconname, &iconx, &icony) > 0))
	{
		editoriconloaded = 1;
		/* save current window context */
		m_push(P_CURSOR | P_EVENT | P_MENU | P_POSITION | P_TEXT | P_WINDOW);
		m_setmode(M_ABS | M_NOWRAP);
		m_setcursor(5);
		get_size(&x, &y, 0, 0);		/* fetch window coordinates */
		get_param( (char *)0, (int *)0, (int *)0, &border );
		windxsize = font_wide * (strlen(file_name));
		if (windxsize < iconx)
			windxsize = iconx;
		windxsize += 2 * border;
		windysize = icony + 2 + 2 * border + font_high;
		/* Make sure window is big enough for mgr */
		temp = 8 + font_wide * 5;
		if (windxsize < temp)
			windxsize = temp;
		temp = 8 + font_high;
		if (windysize < temp)
			windysize = temp;
		if (small_windx == -1)
		{
			small_windx = x;
			small_windy = y;
		}

		if (small_windx + windxsize >= xmax)
			small_windx = xmax - windxsize - 5;

		m_shapewindow(small_windx, small_windy, windxsize, windysize);
		m_setmode(M_ACTIVATE);		/* window is now the active window */
		m_setevent(BUTTON_1, "A\r");	/* Open up on middle or right button */
		m_setevent(BUTTON_2, "A\r");
		m_setevent(MOVE, "B\r");
		m_setevent(RESHAPE, "B\r");
		m_setevent(REDRAW, "R\r");
		/* Display the 'editor' icon */
		m_func(B_COPY);
		iconxpos = (windxsize - iconx) / 2;
		if (iconxpos < 0)
			iconxpos = 0;
		m_setmode(M_NOWRAP);
		m_clear();
		m_bitcopyto(iconxpos, 0, iconx, icony, 0, 0, 0, 4);
		m_moveprint(0, icony + font_high + 2, file_name);

		if (firsttime)
			firsttime = 0;
		else
		{
			m_clearmode(M_ACTIVATE);	/* bury the window */
			m_setevent(ACTIVATE, "A\r"); /* Edit next time we're clicked */
		}

		while (mym_gets(line))			/* read a line from MGR */
		{
			switch (*line)
			{
			case 'A':			/* Middle/right button */
				m_pop();		/* restore window context */
				m_setcursor(0);
				return;
			case 'B':			/* Moved or reshaped: bury it */
				get_size(&small_windx, &small_windy,
					(int *)0, (int *)0);
				m_clearmode(M_ACTIVATE);
				m_setevent(ACTIVATE, "A\r"); /* Edit next time we're clicked */
				break;
			case 'R':			/* Redraw */
				m_clear();
				m_bitcopyto(iconxpos, 0, iconx, icony, 0, 0, 0, 4);
				m_moveprint(0, icony + font_high + 2, file_name);
				break;
			}
		}
	}
}

destroyobject(obj)
register OBJECT *obj;
{
	if (obj)
	{
		objects[obj->id] = (OBJECT *)0;
		Free(obj);
	}
}

quit(code)
{
	m_pop();
	Printf("\n");
	doexit(code);
}

/* User hit the ESC key */
eschit()
{
	setupscreen(0);
	redrawmenubar();
}

/* Init the screen.
 * If flag == 1, reload menus
 */
setupscreen(reload)
{
	int entry;
	int menubarindex;
	register OBJECT *obj;
	int retry;

	m_clear();
	m_textreset();

	menuobject = (OBJECT *)0;
	for (entry = 0; entry < NENTRIES(objects); entry++)
		destroyobject(objects[entry]);
	for (retry = 0; retry < 5; retry++)
	{	/* MGR responses can get tangled up with type-ahead */
		get_font(&font_wide, &font_high);
		get_size(&win_x, &win_y, &win_wide, &win_high);
		get_param(0, &xmax, &ymax, &border);
		if (font_wide && font_high && win_wide && win_high && xmax && ymax && border)
			break;
	}

	if (retry == 5)
	{
		fprintf(stderr, "\n\nPANIC: Can't get stuff from MGR\n\n");
		doexit(1);
	}

	/* Setup the menu string: menu numbers are allocated here */
	menubar[menubarindex = 0] = '\0';
	for (entry = 0; entry < NENTRIES(pulldowns); entry++)
	{
		int len;
		register struct pulldown *pd;

		pd = &pulldowns[entry];
		/* Load up the menus */
		len = strlen(pd->title);
		if (!strcmp(pd->title, "Switch"))
		{
			switchpd = pd;
			switchmenuentryno = entry + 1;
		}
		if (reload == 1)
			menu_load(entry + 1, pd->nentries, pd->menu);
		obj = mkobject(menubarindex * font_wide + border,
			border,
			(menubarindex+len+2) * font_wide - 1 + border,
			font_high - 1 + border,
			OBJ_ACTIVE | OBJ_PDMENU, WH_PDMENU);
		obj->menuno = entry + 1;
		strcat(menubar, " ");
		strcat(menubar, pd->title);
		strcat(menubar, " ");
		menubarindex = strlen(menubar);
	}

	filenamemenuno = ++entry;		/* Allocate menu number for starting loaded menu */

	/* Create the objects for the up and down arrows, scroll bar */
	m_func(B_COPY);
	obj = mkobject(win_wide - ARROWWIDE,0,win_wide - 1,ARROWHIGH - 1,
			OBJ_ACTIVE, WH_UPARROW);
	m_bitcopyto(obj->x_min, obj->y_min, ARROWWIDE, ARROWHIGH, 0, 0, 0, 1);
	obj = mkobject(win_wide - ARROWWIDE, win_high - ARROWHIGH, win_wide - 1,
		win_high - 1, OBJ_ACTIVE, WH_DOWNARROW);
	m_bitcopyto(obj->x_min, obj->y_min, ARROWWIDE, ARROWHIGH, 0, 0, 0, 2);

	/* Get the scroll box BEFORE scroll bar so hits on it have higher priority */
	obj = mkobject(win_wide - ARROWWIDE, ARROWHIGH, win_wide - 1,
			ARROWHIGH * 2, OBJ_ACTIVE, WH_SCROLLBOX);
	scrollboxobj = obj;

	/* The scroll bar */
	obj = mkobject(win_wide - ARROWWIDE, ARROWHIGH, win_wide - 1,
		win_high - ARROWHIGH, OBJ_ACTIVE, WH_SCROLLBAR);
	scrollbarobj = obj;
	m_func(B_SET);
	m_bitwrite(obj->x_min, obj->y_min + 1,
		obj->x_max - obj->x_min, obj->y_max - obj->y_min - 2);

	/* Define the text region */
	m_textregion(0, font_high, win_wide - ARROWWIDE, win_high - font_high);

	firsttime = 1;
}

/*
 * Redraw the menubar line, including filename
 */

redrawmenubar()
{
	int entry;

	m_textreset();

	/* Setup the menu string */
	*menubar = '\0';
	for (entry = 0; entry < NENTRIES(pulldowns); entry++)
	{
		register struct pulldown *pd;

		pd = &pulldowns[entry];
		strcat(menubar, " ");
		strcat(menubar, pd->title);
		strcat(menubar, " ");
	}
	strcat(menubar, " '");
	strcat(menubar, file_name);
	strcat(menubar, "'");
	while (strlen(menubar) < (win_wide - ARROWWIDE) / font_wide)
		strcat(menubar, " ");	/* Yerch */

	menubar[(win_wide - ARROWWIDE) / font_wide] = '\0';

	/* Draw the pull-down menu */
	m_standout();
	m_moveprint(0, font_high, menubar);
	m_standend();

	/* Define the text region */
	m_textregion(0, font_high, win_wide - ARROWWIDE, win_high - font_high);
}

/* Set up the switch menu */
initswitchmenu()
{
	extern struct oldfile *oldfilehead;
	register struct oldfile *ofp;
	register struct menu_entry *mep;
	int nentries, entry;

	ofp = oldfilehead;
	nentries = 0;
	while (ofp)
	{
		nentries++;
		ofp = ofp->next;
	}
	/* Allocate the menu */
	if (!nentries)
		return;
	mep = Malloc((nentries + 1) * sizeof(*mep));
	ofp = oldfilehead;
	entry = 0;
	while (ofp)
	{
		mep[entry].value = roomfor(ofp->name);
		mep[entry].action = Malloc(strlen(ofp->name) + 5);
		sprintf(mep[entry].action, "\036N%s\r", ofp->name);
		entry++;
		ofp = ofp->next;
	}
	/* Copy last entries using struct assignment */
	mep[entry++] = switchpd->menu[0];
	/* Send it to mgr */
	menu_load(switchmenuentryno, nentries + 1, mep);
	/* All used! */
	for (entry = 0; entry < nentries; entry++)
	{
		Free(mep[entry].value);
		Free(mep[entry].action);
	}
	Free(mep);
}

/*
 * Upload a list of the files in . to an mgr menu
 */

#define MAXPERMENU 20		/* Max number of entries per menu */

uploadfilenames()
{
	extern char **scandot();
	register char **dotnames;
	register struct menu_entry *mep;
	int nentries, entry, pageno;

	dotnames = scandot(&nentries);
	if (((long)dotnames <= 0) || !nentries)
		return;
	mep = Malloc(nentries * sizeof(*mep));
	for (entry = 0; entry < nentries; entry++)
	{
		mep[entry].value = roomfor(dotnames[entry]);
		mep[entry].action = Malloc(strlen(dotnames[entry]) + 5);
		sprintf(mep[entry].action, "%s\r", dotnames[entry]);
	}
	for (pageno = 0; pageno < nentries; pageno += MAXPERMENU)
	{
		int submenuno;
		int subnentries;

		submenuno = pageno / MAXPERMENU;
		subnentries = nentries - pageno;
		if (subnentries > MAXPERMENU)
			subnentries = MAXPERMENU;
		menu_load(filenamemenuno + submenuno, subnentries, mep + pageno);
		if (pageno)
		{	/* Link to previous menu */
			m_pagemenu(filenamemenuno + submenuno - 1,
				filenamemenuno + submenuno);
		}
	}
	for (entry = 0; entry < nentries; entry++)
	{
		Free(mep[entry].value);
		Free(mep[entry].action);
		Free(dotnames[entry]);
	}
	Free(mep);
	Free(dotnames);
}

/*
 * Draw the scroll box: called every time the editor needs
 * input: a good place to housekeep
 */
posscrollbox()
{
	register int box_row;
	register OBJECT *obj = scrollboxobj;
	extern int getchead, getctail;

	menusoff();

	if (getchead != getctail)
		return;		/* More characters to go */
#ifdef notdef
	if (ipstate || rawsgetchar())
		return;		/* Don't draw it yet: slows things up */
#endif

	box_row = ((scrollbarobj->y_max - scrollbarobj->y_min - ARROWHIGH - 1) *
			(cur_ptr - mem_buf)) /
			(end_buf - mem_buf);
	box_row += ARROWHIGH + 1;

	if (box_row != obj->y_min || firsttime)
	{	/* Redraw it */
		if (!firsttime)
		{
			m_func(B_SET);
			m_bitwrite(obj->x_min, obj->y_min,
				obj->x_max - obj->x_min - 1,
				obj->y_max - obj->y_min - 1);
		}
		firsttime = 0;
		obj->y_min = box_row;
		obj->y_max = box_row + ARROWHIGH;
		m_func(B_COPY);
		m_bitcopyto(obj->x_min, obj->y_min, ARROWWIDE, ARROWHIGH, 0, 0, 0, 3);
	}
}

readicons()
{
	int dummy1, dummy2;

	m_bitfile(1, uparrowname, &dummy1, &dummy2);
	m_bitfile(2, downarrowname, &dummy1, &dummy2);
	m_bitfile(3, boxname, &dummy1, &dummy2);
}

/* Find address of a char from screen X & Y */

unsigned char *
scanxy(x, y)
{
	unsigned char *cp, *cp2;
	int ypos;
	extern unsigned char *top_lp, *getnext();
	extern int top_y;
	extern int lin_siz;
	extern int wantx, wanty;
	extern unsigned char *wantloc;

	cp = top_lp;
	ypos = top_y;

	for ( ; ; )
	{
		set_param(cp);
		if (lin_siz + ypos > y)
			break;	/* Got the line */
		ypos += lin_siz;
		cp2 = getnext(cp);
		if (cp2 == cp)
			return cp2;
		cp = cp2;
	}

	/* cp points to desired line */
	wantx = x;
	wanty = y - ypos;
	set_param(cp);
	return wantloc;
}

savestate()
{
	clearscrollbar();
	m_push(P_DEFAULT);
	printf("\n");
	m_cleareos();
}

reststate()
{
	m_pop();
	setupscreen(0);
	redrawmenubar();
}

dooz()
{
#ifdef notdef
extern unsigned char *mem_buf;	/* Start of text */
extern unsigned char *end_buf;	/* Points to null after last line */
extern int top_lin;	/* Line number for line at top of screen */
extern int bot_lin;	/* Number of line beyond last displayed */
extern int cur_lin;	/* Current */
extern unsigned char *top_lp;	/* Points to source for top line */
extern unsigned char *bot_lp;	/* Points to the string beyond the last displayed */
extern unsigned char *cur_lp;	/* Start of current line */
extern unsigned char *cur_ptr;	/* Current working pointer */
extern int bot_y;	/* Display line at which no text is displayed */
extern int top_y;	/* Display line of top displayed line */
extern int cur_y;	/* Current Y cursor pos */
extern int cur_x;	/* Current X cursor pos */
extern int lin_siz;	/* Number of screen lines occupied by a string */
extern int lin_dlen;	/* Number of screen unsigned chars occupied by a string */
extern int width;	/* Screen width */
extern int height;	/* Screen height */
extern int nreps;	/* Number of repeats for search/substitute */
extern int nfound;	/* Number of successful searches/substitutes */
extern int modflag;	/* Flag: file modified */
extern unsigned char *blkstart;	/* Start of block */
extern unsigned char *blkend;	/* End of block */
extern unsigned char *newcur_ptr;/* Used for keeping track of where we are */
extern unsigned char **markers;	/* Markers */
extern int blkset;	/* Flag: block is marked */
extern int scr_top;	/* Current screen top */
extern int curscrtop;

	fprintf(stderr, "top_lin=%d, bot_lin=%d, cur_lin=%d\n", top_lin, bot_lin, cur_lin);
	fprintf(stderr, "top_lp=0x%x, bot_lp=0x%x, cur_lp=0x%x, cur_ptr=0x%x\n",
			top_lp, bot_lp, cur_lp, cur_ptr);
	fprintf(stderr, "bot_y=%d, top_y=%d, cur_y=%d, cur_x=%d\n",
			bot_y, top_y, cur_y, cur_x);
	fprintf(stderr, "scr_top-%d, curscrtop=%d\n", scr_top, curscrtop);
#endif
}
