/*
 * Header file for screen editor
 */

#include <sys/types.h>
#include <sys/time.h>

#include <curses.h>
#include <term.h>
#include <string.h>

#ifndef FORMATTER

#define MEM_MAX		30000	/* Maximum amount of memory which the editor attempts to use */
#define IBSIZE		50	/* Insert buffer size */

#define	STATUS	0		/* Status line */
#define TGETPOS 5		/* X position for getting input line */
#define SEDLINESIZE	256	/* Max line we can input */

#define CLREOL	"\033T"
#define DELLINE	"\033R"
#define CLRSCRN	"\033*"
#define INSLIN	"\033E"
#define CLREOS	"\033Y"
#define MVCURS1	'\033'
#define MVCURS2	'='

#define MASK(x) (x & 0x7f)
#define INDEX maskindex
#define STRNCMP maskstrncmp
unsigned char	*maskindex();

#define CTL(c) ((c) & 0x1f)
#define ALT(c) ((c) | 0x80)

/* Masks for attributes */
#define SUBSCRIPT	1
#define SUPERSCRIPT	2
#define BOLD		4
#define UNDERLINE	8
#define WIDE		0x10		/* No on-screen support */
#define ITALICS		0x40

/* Justification modes */
#define J_RIGHT		0
#define J_RAGGED	1
#define J_CENTRE	2
#define J_MAX		3		/* Number of modes */

#endif

#ifndef NELEM
#define NELEM(x)	(sizeof(x)/sizeof(*x))
#endif

/* Imbedded control codes for attributes: pretty WS-like */
#define CC_BOLD		(CTL('B'))
#define CC_UNDERLINE	(CTL('S'))
#define CC_ITALICS	(CTL('Y'))
#define CC_SUBSCRIPT	(CTL('V'))
#define CC_SUPERSCRIPT	(CTL('T'))
#define CC_WIDE		(CTL('W'))
#define CC_USER1	(CTL('E'))
#define CC_USER2	(CTL('Q'))
#define CC_USER3	(CTL('R'))

/* A structure representing previously edited files */
struct oldFile
{
	char name[32];
	long markers[10];
	long blkstart;
	long blkend;
	int blkset;
	int cur_ptr;
	int hilitestart;
	int hiliteend;
	int tabwidth;
	int beingEdited;
	time_t useTime;
	int spare[7];
};

#define CAPTURESIZE	256

struct sr_data
{
	char new[SEDLINESIZE];
	char old[SEDLINESIZE];
	int useregexp;
	int reverse;
	int ignorecase;
	int fromtop;
	int wrap;
	int ntimes;
	int initialised;
};

int findOldFile();
char *roll();
char *roomfor();
void parenmatch(char *incstr, char *decstr, int direction);
void movstr(void *from, void *to, int len);
void freeDDEPaths();
void leditEnd();
void freeRollFiles();
void cleanUpSlicedArgs(char **subStrings, int nstrings);
int sliceargs(char *str, char ***strs, int canexpand, int colon);
void cleanUpTags(void);
void opchar(char c);
char *currentFileName(void);
char *currentPathName(void);
void addTagFile(const char *_path);
char *_patchenv(char *in);
char *nodotslash(char *_p);
char *searchup(char *_dirName, char *_fileName);
void xterm_pointer_mode();
unsigned char *scanxy();
char *patchenv(char *name);
int read_fd(long fd, unsigned char *name, int mode, int poll, int isnewfile);
void clear_eol(void);
void clear_eos(int x, int y);
unsigned char *getnext(unsigned char *lp);
char *lookuptag(char *str, unsigned char **pname, unsigned char **psearch,
				int *pis_number);
char *nledit(char *lbuf, int length);

unsigned char *getnext(unsigned char *lp);
unsigned char *get_prev(unsigned char *lp);
unsigned char *fwd_fnw(unsigned char *ptr);
unsigned char *rev_fnw(unsigned char *ptr);

unsigned char *Index();
extern int splatr();

extern int unlink_input;

#define FORWARD 1
#define BACKWARD -1

#define Malloc(n) malloc(n)
#define Realloc(a, n) realloc(a, n)
#define Free(n) free(n)

static inline void *zMalloc(int n)
{
	void *ret = malloc(n);
	if (ret)
		memset(ret, 0, n);
	return ret;
}
