/*	TABS8
 * terminal dependent stuff for mgr
 */

#include <stdio.h>
#include "../sedit.h"
#include "mgrstate.h"

#ifdef applix1616
#include <syscalls.h>
#endif

#undef WIDE				/* Used by mgr */
#include <term.h>	/* NODEP */	/* MGR include file */

/* The O/S version number, as returned from GETROMVER system call */
extern int romver;

/* The height and width of the screen in chars. */
extern int height, width;

/* 'curscrtop' is 0 if status line disabled, else 1 */
extern int curscrtop;

/* 'dispcontrol' toggles the display of control chars (^OD) */
extern int dispcontrol;

/* 'overstrike' is true if overtyping chars (^\ toggles) */
extern int overstrike;

/* 'attrnow' describes the current output attribute setting. 'attrwant' is
 * the new, desired setting
 */
extern int attrnow, attrwant;

/* Left and right margin settings, overridden by command line switches */
extern int rmargin, lmargin;

/* 'justifymode' can be one of J_RIGHT, J_RAGGED and J_CENTRE */
extern int justifymode;

int ismgr = 1;

/*
 * 'cur_x' and 'cur_y' are the current cursor position. May be useful
 * for fudging things like clear to end of line if the terminal can't do it.
 * Note that cur_x can be greater than 'width', so the real Y cursor position
 * is at (cur_y + cur_x / width) and the real X position is at
 * (cur_x % width).
 */
extern int cur_x, cur_y;

/*
 * Initialise the terminal for editing. Just the very first thing done.
 */

terminit()
{
	/* Set up 'width' and 'height' */
	/* Print out a char to select the current window */
	menuinit();				/* Init scroll bar, etc */
	m_ttyset();
	editinit(1);
	/* Select the default justification mode */
	justifymode = J_RIGHT;

	/* Display control chars as ^X */
	dispcontrol = 1;

	/* Start up in insert mode */
	overstrike = 0;

	/* Disable status line */
	curscrtop = 0;
}

editinit(redomargins)
{
	get_colrow(&width, &height);		/* Get window size */

	if (redomargins)
	{
		/* Set up the default margin settings */
		rmargin = width - 5;
		lmargin = 0;
	}

	/* Assume every attribute is on, so setbul() will turn them all off */
	attrwant = 0;
	attrnow = BOLD|UNDERLINE|SUBSCRIPT|SUPERSCRIPT|ITALICS;
	setbul();
}

/* Terminate the editing session. Called just before quitting */
termterm()
{
	menucleanup();
	m_cleareos();
	sleep(1);
	m_ttyreset();
}

/*
 * Delete a line
 */
dellines(linenum, nlines)
register int linenum, nlines;
{
	if (linenum < height)
	{
		while (nlines--)
		{
			gotoxy(0, linenum);
			m_deleteline();
		}
	}
}

/*
 * Move the cursor to the appropriate position.
 */

gotoxy(x, y)
register int x, y;
{
	if (x >= width)
	{
		y += x / width;
		x = x % width;
	}
	m_move(x, y);
}

clr_eol()
{
	m_cleareol();
}

clr_eos(x, y)
{
	m_cleareos();
}

clr_scrn()
{
	gotoxy(0, 0);
	m_clear();
	gotoxy(0, 0);
}

dn_scrl()
{
	extern int scr_top;
	gotoxy(0, scr_top);
	m_addline();		/* Insert a line at top of screen */
}

inslin(row)
{
	m_addline();
}

prcrlf()
{
	Printf("\n");
}

/* Set a background colour */
hil(val)
{
	if (val)
		m_standout();
	else
		m_standend();
}

setbul()
{
	if (!dispcontrol && attrwant && !attrnow)
	{
		m_standout();
	}
	else if (!attrwant && attrnow)
	{
		m_standend();
	}
	attrnow = attrwant;
}

/* Push-ahead buffer: power of 2 */
static unsigned char getcbuf[CAPTURESIZE];
#define GETCMASK (sizeof(getcbuf) - 1)
int getchead = 0, getctail = 0;

#ifdef applix1616
rawsgetchar()
{
	return ((getchead != getctail) || SGETCHAR());
}
#endif

/* Push a char into input buffer */
sendtome(c)
unsigned char c;
{
	if (((getchead + 1) & GETCMASK) == getctail)
		return;
	getcbuf[getchead++] = c;
	getchead &= GETCMASK;
}

stringtome(s)
char *s;
{
	while (*s)
		sendtome(*s++);
}

ipbready()
{
	return getchead != getctail;
}

flushop()
{
	fflush(m_termout);
	fflush(stdout);
}

/*
 * getchar() ignoring control-D eof char
 */

rawgetchar()
{
	unsigned int ch;
	extern int ipstate;

	posscrollbox();
	flushop();
#ifdef notdef
	if (ipstate && !rawsgetchar())
	{
		switch (ipstate)
		{
		case IPS_SENDCW:
			return CTL('W');
		case IPS_SENDCZ:
			return CTL('Z');
		}
	}
#endif
	ipstate = 0;
	if (getchead != getctail)
	{
		ch = getcbuf[getctail++];
		getctail &= GETCMASK;
	}
	else
	{
		unsigned char cc;
		read(0, &cc, 1);
		ch = cc;
	}

	if (ch == '\n')
		ch = '\r';		/* Unix is silly sometimes */

	if (ch == 0)
		ch = ' ';		/* CTL-space */

	if (ch == 036)	/* MGR event */
	{
		mgrevent();
		return 0;		/* Null char to main loop */
	}

#ifdef notdef
	/* If it is an EOF, return the current EOF character. */
	if (ch == BEC_RPASTEOF)
	{
		Stdin = SET_SIP(-1);
		if (Stdin >= NIODRIVERS)
		{
			SET_SIP((romver > 0x41) ? OPEN("tty:", 1) : 0);
			goto retry;
		}
		eofchar = OPTION(OPT_EOFCHAR|OPT_READ, 0);
		if (eofchar < 256)
			ch = eofchar;
	}
#endif
	return ch;
}

beep()
{
	opchar(7);
}

rawputchar(c)
{
	m_putchar(c);
}

#if 0
PUTCHAR(c)
{
	m_putchar(c);
}
#endif

SCREENWRITE(addr, n)
char *addr;
{
	fwrite(addr, 1, n, m_termout);
	/*fflush(m_termout);*/
}

Printf(str, p1, p2, p3, p4, p5, p6, p7, p8, p9)
char *str;
{
	char buf[1000];

	sprintf(buf, str, p1, p2, p3, p4, p5, p6, p7, p8, p9);
	SCREENWRITE(buf, strlen(buf));
}
